Package: pxmake
Title: Make PX-Files in R
Version: 0.20.1
Authors@R: c(
             person("Johan", "Ejstrud", 
                    email = "johan@ejstrud.com",
                    role = c("cre", "aut")
                   ),
             person("Lars Pedersen",
                    email = "larp@stat.gl",
                    role = c("aut")
                   ),
             person("Statistics Greenland", 
                    role = "cph", 
                    comment = "https://stat.gl/"
                    )
            )
Description: Create PX-files from scratch or read and modify existing ones. 
    Includes a function for every PX keyword, making metadata manipulation
    simple and human-readable.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: arrow, dplyr, furrr, magrittr, methods, openxlsx, purrr,
        readxl, rlang, stats, stringi, stringr, tibble, tidyr, tools,
        utils, vctrs
Suggests: knitr, rmarkdown, withr, testthat (>= 3.0.0), curl
Config/testthat/edition: 3
Config/testthat/parallel: true
Config/testthat/start-first: 60-metamake-bexltall
VignetteBuilder: knitr
Depends: R (>= 4.1.0)
LazyData: true
URL: https://github.com/StatisticsGreenland/pxmake,
        https://statisticsgreenland.github.io/pxmake/
BugReports: https://github.com/StatisticsGreenland/pxmake/issues
NeedsCompilation: no
Packaged: 2026-04-17 16:44:35 UTC; johan
Author: Johan Ejstrud [cre, aut],
  Lars Pedersen [aut],
  Statistics Greenland [cph] (https://stat.gl/)
Maintainer: Johan Ejstrud <johan@ejstrud.com>
Repository: CRAN
Date/Publication: 2026-04-18 16:40:09 UTC
Built: R 4.5.2; ; 2026-04-20 23:40:11 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(px_add_totals,px)
S3method(px_aggregallowed,px)
S3method(px_autopen,px)
S3method(px_axis_version,px)
S3method(px_baseperiod,px)
S3method(px_cellnote,px)
S3method(px_cellnotex,px)
S3method(px_cfprices,px)
S3method(px_charset,px)
S3method(px_codepage,px)
S3method(px_confidential,px)
S3method(px_contact,px)
S3method(px_contents,px)
S3method(px_contvariable,px)
S3method(px_copyright,px)
S3method(px_creation_date,px)
S3method(px_data,px)
S3method(px_database,px)
S3method(px_datasymbol1,px)
S3method(px_datasymbol2,px)
S3method(px_datasymbol3,px)
S3method(px_datasymbol4,px)
S3method(px_datasymbol5,px)
S3method(px_datasymbol6,px)
S3method(px_datasymbolnil,px)
S3method(px_decimals,px)
S3method(px_description,px)
S3method(px_descriptiondefault,px)
S3method(px_domain,px)
S3method(px_elimination,px)
S3method(px_figures,px)
S3method(px_heading,px)
S3method(px_infofile,px)
S3method(px_language,px)
S3method(px_languages,px)
S3method(px_last_updated,px)
S3method(px_link,px)
S3method(px_map,px)
S3method(px_matrix,px)
S3method(px_next_update,px)
S3method(px_note,px)
S3method(px_notex,px)
S3method(px_official_statistics,px)
S3method(px_order,px)
S3method(px_precision,px)
S3method(px_refperiod,px)
S3method(px_showdecimals,px)
S3method(px_source,px)
S3method(px_stockfa,px)
S3method(px_stub,px)
S3method(px_subject_area,px)
S3method(px_subject_code,px)
S3method(px_tableid,px)
S3method(px_timeval,px)
S3method(px_title,px)
S3method(px_units,px)
S3method(px_update_frequency,px)
S3method(px_valuenote,px)
S3method(px_valuenotex,px)
S3method(px_values,px)
S3method(px_variable_label,px)
S3method(px_variable_type,px)
export(px)
export(px_add_totals)
export(px_aggregallowed)
export(px_autopen)
export(px_axis_version)
export(px_baseperiod)
export(px_cellnote)
export(px_cellnotex)
export(px_cfprices)
export(px_charset)
export(px_classification)
export(px_codepage)
export(px_confidential)
export(px_contact)
export(px_contents)
export(px_contvariable)
export(px_copyright)
export(px_creation_date)
export(px_data)
export(px_database)
export(px_datasymbol1)
export(px_datasymbol2)
export(px_datasymbol3)
export(px_datasymbol4)
export(px_datasymbol5)
export(px_datasymbol6)
export(px_datasymbolnil)
export(px_decimals)
export(px_description)
export(px_descriptiondefault)
export(px_domain)
export(px_elimination)
export(px_figures)
export(px_heading)
export(px_infofile)
export(px_language)
export(px_languages)
export(px_last_updated)
export(px_link)
export(px_map)
export(px_matrix)
export(px_micro)
export(px_next_update)
export(px_note)
export(px_notex)
export(px_official_statistics)
export(px_order)
export(px_precision)
export(px_refperiod)
export(px_save)
export(px_save_classification)
export(px_showdecimals)
export(px_source)
export(px_stockfa)
export(px_stub)
export(px_subject_area)
export(px_subject_code)
export(px_tableid)
export(px_timeval)
export(px_title)
export(px_units)
export(px_update_frequency)
export(px_validate)
export(px_valuenote)
export(px_valuenotex)
export(px_values)
export(px_variable_label)
export(px_variable_type)
importFrom(dplyr,across)
importFrom(dplyr,all_of)
importFrom(dplyr,any_of)
importFrom(dplyr,desc)
importFrom(dplyr,ends_with)
importFrom(dplyr,everything)
importFrom(dplyr,if_any)
importFrom(dplyr,last_col)
importFrom(dplyr,one_of)
importFrom(dplyr,starts_with)
importFrom(dplyr,where)
importFrom(magrittr,"%>%")
importFrom(methods,as)
importFrom(rlang,":=")
importFrom(stats,na.omit)
importFrom(stats,setNames)
importFrom(utils,head)
importFrom(utils,tail)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# pxmake <a href="https://statisticsgreenland.github.io/pxmake/"><img src="man/figures/logo.png" align="right" height="139" alt="pxmake website" /></a>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/pxmake)](https://cran.r-project.org/package=pxmake)
[![Tests](https://github.com/StatisticsGreenland/pxmake/actions/workflows/tests.yml/badge.svg)](https://github.com/StatisticsGreenland/pxmake/actions/workflows/tests.yml)
[![Codecov test
coverage](https://codecov.io/gh/StatisticsGreenland/pxmake/branch/main/graph/badge.svg)](https://app.codecov.io/gh/StatisticsGreenland/pxmake?branch=main)
[![Mentioned in Awesome Official
Statistics](https://awesome.re/mentioned-badge.svg)](https://github.com/SNStatComp/awesome-official-statistics-software)
[![Montly
downloads](https://cranlogs.r-pkg.org/badges/last-month/pxmake)](https://cran.r-project.org/package=pxmake)
<!-- badges: end -->

## Overview

‘pxmake’ is an R package for creating and modifying PX-files.

With pxmake you can:

- Import a PX-file, modify it, and save it as a new PX-file.
- Modify all metadata keywords in a PX-file.
- Do complex modifications to a PX-file, like adding total levels to a
  variable.
- Save a PX-file as an Excel workbook.

## Installation

``` r
# Install the latest release from CRAN
install.packages('pxmake')

# Or the development version from GitHub
# install.packages('pak')
pak::pak("StatisticsGreenland/pxmake")
```

## How to use

*(Find complete documentation on [pxmake
webpage](https://statisticsgreenland.github.io/pxmake/).)*

Use `px()` to import an existing PX-file into R.

``` r
library(pxmake)

# Import PX-file
x <- px(input = "example.px")
```

Once imported, use one of pxmake’s *modifying functions*.

In general, modifying functions are named after the keyword they modify.
It’s possible to chain multiple modifying functions together in
tidyverse style by using the pipe operator `%>%`.

``` r
library(magrittr) # import pipe

# Create px object from data frame
x <- px(data.frame(
  year = as.character(rep(2021:2023, each = 3)),
  group = c("a", "b", "c"),
  value = runif(9)
))

head(x$data, 4)
#> # A tibble: 4 × 3
#>   year  group value
#>   <chr> <chr> <dbl>
#> 1 2021  a     0.266
#> 2 2021  b     0.372
#> 3 2021  c     0.573
#> 4 2022  a     0.908

x %>%
  px_timeval("year") %>% # Set year as TIMEVAL
  px_heading("year") %>% # Set year as HEADING
  px_stub("group") %>% # Set group as STUB
  px_decimals("2") %>% # Set DECIMALS to 2
  px_save("example.px") # Save as PX-file
```

### Modifying functions

Currently the following 59 keywords have a modifying function in pxmake:

    #>  Keyword             Function name         
    #>  AGGREGALLOWED       px_aggregallowed      
    #>  AUTOPEN             px_autopen            
    #>  AXIS-VERSION        px_axis_version       
    #>  BASEPERIOD          px_baseperiod         
    #>  CELLNOTE            px_cellnote           
    #>  CELLNOTEX           px_cellnotex          
    #>  CFPRICES            px_cfprices           
    #>  CHARSET             px_charset            
    #>  CODEPAGE            px_codepage           
    #>  CONFIDENTIAL        px_confidential       
    #>  CONTACT             px_contact            
    #>  CONTENTS            px_contents           
    #>  CONTVARIABLE        px_contvariable       
    #>  COPYRIGHT           px_copyright          
    #>  CREATION-DATE       px_creation_date      
    #>  DATA                px_data               
    #>  DATABASE            px_database           
    #>  DATASYMBOL1         px_datasymbol1        
    #>  DATASYMBOL2         px_datasymbol2        
    #>  DATASYMBOL3         px_datasymbol3        
    #>  DATASYMBOL4         px_datasymbol4        
    #>  DATASYMBOL5         px_datasymbol5        
    #>  DATASYMBOL6         px_datasymbol6        
    #>  DATASYMBOLNIL       px_datasymbolnil      
    #>  DECIMALS            px_decimals           
    #>  DESCRIPTION         px_description        
    #>  DESCRIPTIONDEFAULT  px_descriptiondefault 
    #>  DOMAIN              px_domain             
    #>  ELIMINATION         px_elimination        
    #>  HEADING             px_heading            
    #>  INFOFILE            px_infofile           
    #>  LANGUAGE            px_language           
    #>  LANGUAGES           px_languages          
    #>  LAST-UPDATED        px_last_updated       
    #>  LINK                px_link               
    #>  MAP                 px_map                
    #>  MATRIX              px_matrix             
    #>  NEXT-UPDATE         px_next_update        
    #>  NOTE                px_note               
    #>  NOTEX               px_notex              
    #>  OFFICIAL-STATISTICS px_official_statistics
    #>  PRECISION           px_precision          
    #>  REFPERIOD           px_refperiod          
    #>  SHOWDECIMALS        px_showdecimals       
    #>  SOURCE              px_source             
    #>  STOCKFA             px_stockfa            
    #>  STUB                px_stub               
    #>  SUBJECT-AREA        px_subject_area       
    #>  SUBJECT-CODE        px_subject_code       
    #>  TABLEID             px_tableid            
    #>  TIMEVAL             px_timeval            
    #>  TITLE               px_title              
    #>  UNITS               px_units              
    #>  UPDATE-FREQUENCY    px_update_frequency   
    #>  VALUENOTE           px_valuenote          
    #>  VALUENOTEX          px_valuenotex         
    #>  VALUES              px_values             
    #>  VARIABLE-LABEL      px_variable_label     
    #>  VARIABLE-TYPE       px_variable_type

In addition to the above, the following 3 modifying functions are
available:

    #>  Function name
    #>  px_add_totals
    #>  px_figures   
    #>  px_order

See the help page for each modifying function for more information.

<details>

<summary>

Keywords without modifying functions
</summary>

These 26 keywords currently doesn’t have a modifying function, but can
be implemented.

    #>  Keyword             Function name          Priority Complexity
    #>  ATTRIBUTE-ID        px_attribute_id                           
    #>  ATTRIBUTE-TEXT      px_attribute_text                         
    #>  ATTRIBUTES          px_attributes                             
    #>  DATANOTE            px_datanote                               
    #>  DATANOTECELL        px_datanotecell                           
    #>  DATANOTESUM         px_datanotesum                            
    #>  DATASYMBOLSUM       px_datasymbolsum                          
    #>  DAYADJ              px_dayadj                                 
    #>  DEFAULT-GRAPH       px_default_graph                          
    #>  DIRECTORY-PATH      px_directory_path                         
    #>  DOUBLECOLUMN        px_doublecolumn                           
    #>  FIRST-PUBLISHED     px_first_published                        
    #>  HIERARCHIES         px_hierarchies                            
    #>  HIERARCHYLEVELS     px_hierarchylevels                        
    #>  HIERARCHYLEVELSOPEN px_hierarchylevelsopen                    
    #>  HIERARCHYNAMES      px_hierarchynames                         
    #>  INFO                px_info                                   
    #>  KEYS                px_keys                                   
    #>  META-ID             px_meta_id                                
    #>  PARTITIONED         px_partitioned                            
    #>  PRESTEXT            px_prestext                               
    #>  PX-SERVER           px_px_server                              
    #>  ROUNDING            px_rounding                               
    #>  SEASADJ             px_seasadj                                
    #>  SURVEY              px_survey                                 
    #>  SYNONYMS            px_synonyms

Finally these 2 keywords will not have a modifying function, because
they are automatically determined by the data.

    #>  Keyword     
    #>  CODES       
    #>  VARIABLECODE

</details>

## Need help?

If you have any questions or need help, feel free to [open an
issue](https://github.com/StatisticsGreenland/pxmake/issues/new) on
GitHub, or contact [Lars and
Johan](https://github.com/StatisticsGreenland/pxmake/graphs/contributors)
via email.

## For developers

### PX specification

pxmake is based on the [PX-file format specification on Statistics
Swedens
homepage](https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf).

### PxJob

Some tests cases uses
[PxJob](https://stat.fi/en/services/statistical-data-services/statistical-databases/px-suite).
Install [pxjob64Win](https://github.com/StatisticsGreenland/pxjob64Win)
to be able ro run these tests. This only works on Windows.

### How to create a new release

1.  Run `usethis::use_release_issue()` and follow checklist.
2.  Sync fork on [PxTools/pxmake](https://github.com/PxTools/pxmake).

# Various helper functions. See test cases in test-00-helper_function.R for
# examples of function input and outputs.

#' Add quotes around string
#'
#' @param str String to quote
#'
#' @returns String
#' @keywords internal
str_quote <- function(str) {
  stringr::str_c('"', str, '"')
}

#' Add language, other than the main langue, to keyword
#'
#' @inheritParams add_sub_key_to_keyword
#' @param main_language Main language of PX-file
#' @param language Lange to add to keyword
#'
#' @returns String
#' @keywords internal
add_language_to_keyword <- function(keyword, main_language, language) {
  ifelse(language == main_language | is.na(language),
    keyword,
    stringr::str_glue("{keyword}[{language}]") |> as.character()
  )
}

#' Add a sub key word (see px-specification for details on sub keys)
#'
#' @param keyword String, name of keyword
#' @param name String, name of sub key
#'
#' @returns String
#' @keywords internal
add_sub_key_to_keyword <- function(keyword, name) {
  ifelse(is.na(name),
    keyword,
    stringr::str_glue('{keyword}("{name}")')
  )
}

#' Add cell name to keywords that support it (see px-specification)
#'
#' @inheritParams add_sub_key_to_keyword
#' @param name String, cell name to add to keyword
#'
#' @returns String
#' @keywords internal
add_cell_to_keyword <- function(keyword, name) {
  keyword_has_subkey <- stringr::str_sub(keyword, -1) == ")"

  ifelse(is.na(name),
    keyword,
    ifelse(keyword_has_subkey,
      stringr::str_glue('{stringr::str_sub(keyword, 1, -2)},"{name}")'),
      add_sub_key_to_keyword(keyword, name)
    )
  )
}

#' Add quotes around unless in some very specific cases required by
#' the px format
#'
#' @inheritParams str_quote
#'
#' @returns String
#' @keywords internal
quote_unless_yes_no <- function(str) {
  str_is_quoted <- function(str) {
    stringr::str_length(str) >= 2 &
      stringr::str_sub(str, 1, 1) == '"' &
      stringr::str_sub(str, -1, -1) == '"'
  }

  ifelse(
    str %in% c("YES", "NO") | str_is_quoted(str) |
      stringr::str_starts(str, "TLIST\\("),
    str,
    str_quote(str)
  )
}

#' Get a sorted list of distinct values in list
#'
#' @param lst List to sort
#'
#' @returns List
#' @keywords internal
lst_distinct_and_arrange <- function(lst) {
  if (length(lst) == 0) {
    return(lst)
  }

  tmp <- lapply(lapply(lst, unique), sort)
  tmp[order(names(tmp))]
}

#' Put two named lists together, remove duplicates and sort
#'
#' @param lst1 List to merge
#' @param lst2 List to merge
#'
#' @returns List
#' @keywords internal
merge_named_lists <- function(lst1, lst2) {
  if (length(lst1) == 0) {
    lst1 <- lst2
  } else if (length(lst2) == 0) {
    lst2 <- lst1
  }

  keys <- sort(unique(c(names(lst1), names(lst2))))

  add_missing_keys <- function(lst, keys) {
    keys_missing_in_list <- setdiff(keys, names(lst))
    lst[keys_missing_in_list] <- NA_character_
    lst
  }

  lst1_sorted <-
    lst1 |>
    add_missing_keys(keys) |>
    lst_distinct_and_arrange()

  lst2_sorted <-
    lst2 |>
    add_missing_keys(keys) |>
    lst_distinct_and_arrange()

  if (identical(lst1_sorted, lst2_sorted)) {
    temp <- lst1_sorted
  } else {
    temp <- setNames(
      mapply(c, lst1_sorted, lst2_sorted, SIMPLIFY = FALSE), keys
    )
  }

  lst_distinct_and_arrange(temp)
}

#' Get time scale code from values (see TIMEVAL in px-specification)
#'
#' @param values Values form PX-file
#'
#' @returns A character vector
#' @keywords internal
get_timeval_type_from_values <- function(values) {
  time_type <-
    values |>
    na.omit() |>
    stringr::str_replace_all("[:digit:]", "") |>
    paste(collapse = "") |>
    stringr::str_sub(1, 1)

  if (time_type == "") {
    time_type <- "A"
  }

  time_type
}

#' Format time values for PX-file
#'
#' @param values Time values
#'
#' @returns A character vector
#' @keywords internal
format_time_values <- function(values) {
  paste0(
    "TLIST(",
    get_timeval_type_from_values(values),
    "1),",
    values |>
      stringr::str_replace_all("[:alpha:]", "") |>
      str_quote() |>
      stringr::str_c(collapse = ",")
  )
}

#' Get time values from TIMEVAL string
#'
#' @param str String with time values
#'
#' @returns A character vector
#' @keywords internal
get_values_from_time_format <- function(str) {
  short_syntax <- any(stringr::str_detect(str, "-"))

  tmp <-
    str |>
    stringr::str_split(",") |>
    unlist()

  tlist <- tmp |> head(1)
  type <- stringr::str_sub(tlist, 7, 7)
  value_part <- tmp |> tail(-1)

  if (short_syntax) {
    times <-
      value_part |>
      stringr::str_replace_all("[^0-9-]", "") |>
      stringr::str_split("-", simplify = TRUE)

    interval_start <- times[1]
    interval_end <- times[2]

    if (type == "A") {
      values <- as.character(seq(interval_start, interval_end))
    } else {
      indicies_per_year <- c(H = 2, Q = 4, M = 12)

      year_start <- stringr::str_sub(interval_start, 1, 4)
      year_end <- stringr::str_sub(interval_end, 1, 4)

      values <-
        tidyr::crossing(
          year = seq(year_start, year_end),
          index = seq(1, indicies_per_year[type])
        ) |>
        dplyr::mutate(
          value = paste0(.data$year, .data$index), .keep = "none"
        ) |>
        dplyr::filter(as.numeric(.data$value) < interval_end) |>
        dplyr::pull(1)
    }
  } else {
    values <-
      value_part |>
      stringr::str_replace_all('"', "")
  }

  if (type == "A") {
    values
  } else {
    paste0(stringr::str_sub(values, 1, 4), type, stringr::str_sub(values, 5))
  }
}

#' Zips list
#'
#' Combine two list, by zipping them together in the order \code{v1[1]},
#' \code{v2[1]}, \code{v1[2]}, \code{v2[2]}, \code{v1[3]}, ...
#'
#' @param v1 First vector Elements ends up on uneven indexes.
#' @param v2 Second vector. Elements ends up on even indexes.
#'
#' @returns List
#' @keywords internal
zip_vectors <- function(v1, v2) {
  if (length(v1) != length(v2)) {
    stop("v1 and v2 must have same length.")
  }

  base::matrix(c(v1, v2), ncol = 2) |>
    t() |>
    as.list() |>
    unlist()
}

#' Split long strings at commas
#'
#' Long strings are split so they are no longer than 256 characters and end at
#' a comma.
#'
#' @param str String
#' @param max_line_length Integer longest allowed line length
#'
#' @returns A character vector
#' @keywords internal
break_long_lines <- function(str, max_line_length = 256) {
  if (is.null(str)) {
    ""
  } else if (is.na(str)) {
    ""
  } else if (nchar(str) > max_line_length) {
    comma_split <-
      str |>
      stringr::str_locate_all('","') |>
      as.data.frame() |>
      dplyr::filter(.data$start < max_line_length) |>
      dplyr::slice_tail(n = 1) |>
      dplyr::pull(.data$start)

    if (identical(comma_split, integer(0))) {
      # no comma_split character; split at specific point
      line_start <- paste0(stringr::str_sub(str, 1, max_line_length - 2), '"')
      line_end <- paste0('"', stringr::str_sub(str, max_line_length - 1))
    } else {
      line_start <- stringr::str_sub(str, 1, comma_split + 1)
      line_end <- stringr::str_sub(str, comma_split + 2, -1)
    }

    c(line_start, break_long_lines(line_end, max_line_length))
  } else {
    str
  }
}

#' Convert a variable to a list
#'
#' @param df Data frame
#' @param var Character. Variable to convert to list
#'
#' @returns A data frame
#' @keywords internal
wrap_varaible_in_list <- function(df, var) {
  df |>
    dplyr::rowwise() |>
    dplyr::mutate(across(all_of(var), ~ list(.x))) |>
    dplyr::ungroup()
}

#' Read lines with specific encoding
#'
#' @param path Path to file
#' @param encoding Encoding to use
#'
#' @returns A character vector
#' @keywords internal
read_lines_with_encoding <- function(path, encoding) {
  file_connection <- file(path, encoding = encoding)
  lines <- readLines(con = file_connection, warn = FALSE)
  close(file_connection)

  lines
}

#' Get PX-file content as lines
#'
#' Read file with correct encoding.
#'
#' @param px_path Path to a PX-file
#'
#' @returns A character vector
#' @keywords internal
read_px_file <- function(px_path) {
  read_lines_with_encoding(
    path = px_path,
    encoding = get_encoding_from_px_file(px_path)
  )
}

#' Read lines from file with guessed encoding
#'
#' Wrapper around readLines to guess encoding before reading.
#'
#' @param path Path to file
#'
#' @returns A character vector
#' @keywords internal
read_lines_guess_encoding <- function(path) {
  read_lines_with_encoding(
    path = path,
    encoding = guess_file_encoding(path)
  )
}

#' Default encoding to read and save PX-file in
#'
#' @returns Character
#' @keywords internal
get_default_encoding <- function() {
  "latin1"
}

#' Get encoding listed in PX-file
#'
#' Encoding is listed in CODEPAGE.
#'
#' @inheritParams read_px_file
#'
#' @returns Character
#' @keywords internal
get_encoding_from_px_file <- function(px_path) {
  con <- file(px_path, open = "r")
  on.exit(close(con))

  while (length(line <- readLines(con, n = 1, warn = FALSE)) > 0) {
    if (startsWith(line, "CODEPAGE=")) {
      return(sub('^CODEPAGE="(.+)";?$', "\\1", line))
    }
    if (startsWith(line, "DATA=")) break
  }

  get_default_encoding()
}

#' Guess encoding of file
#'
#' @param path Path to file
#' @param prefer Named list of encodings, used to give higher confidence to
#' specific encodings.
#'
#' @returns Character
#' @keywords internal
guess_file_encoding <- function(
  path,
  prefer = list(
    "UTF-8" = .2,
    "ISO-8859-1" = .2
  )
) {
  prefer_df <-
    prefer |>
    tibble::enframe(name = "Encoding", value = "Confidence") |>
    tidyr::unnest("Confidence") |>
    dplyr::mutate(across("Encoding", as.character))

  best_guess <-
    readBin(path, what = "raw", n = file.info(path)$size) |>
    stringi::stri_enc_detect() |>
    (\(x) x[[1]])() |>
    dplyr::bind_rows(prefer_df) |>
    dplyr::group_by(Encoding) |>
    dplyr::summarise(Confidence = sum(.data$Confidence)) |>
    dplyr::arrange(desc(.data$Confidence)) |>
    dplyr::pull("Encoding") |>
    (\(x) x[[1]])()


  best_guess
}

#' Check if a path has a specific extension (function factory)
#'
#' @param extension String, file name extension
#'
#' @returns Logic
#' @keywords internal
is_path_extension <- function(extension) {
  function(path) {
    if (is.character(path)) {
      identical(TRUE, tolower(tools::file_ext(path)) == tolower(extension))
    } else {
      FALSE
    }
  }
}

is_rds_file <- is_path_extension("rds")
is_parquet_file <- is_path_extension("parquet")
is_xlsx_file <- is_path_extension("xlsx")
is_px_file <- is_path_extension("px")
is_r_file <- is_path_extension("R")

#' Create and return path to temporary directory
#'
#' @returns Character
#' @keywords internal
temp_dir <- function() {
  path <- tempfile()
  dir.create(path)
  path
}

#' Change all variables to character
#'
#' @param df Data frame
#'
#' @returns A data frame
#' @keywords internal
mutate_all_vars_to_character <- function(df) {
  dplyr::mutate(df, dplyr::across(everything(), as.character))
}

#' Create temporary file
#'
#' Get a temporary file path with a specific extension (function factory)
#'
#' @param extension String, file name extension
#'
#' @returns Path to temporary file
#' @keywords internal
temp_file_with_extension <- function(extension) {
  function() {
    tempfile(fileext = extension)
  }
}

temp_px_file <- temp_file_with_extension(".px")
temp_rds_file <- temp_file_with_extension(".rds")
temp_xlsx_file <- temp_file_with_extension(".xlsx")
temp_r_file <- temp_file_with_extension(".R")
temp_parquet_file <- temp_file_with_extension(".parquet")

#' Align data frames
#'
#' Align df_a to have the same columns and column types as df_b.
#'
#' @param df_a Data frame to align
#' @param df_b Data frame to align to
#'
#' @returns A data frame
#' @keywords internal
align_data_frames <- function(df_a, df_b) {
  names_a <- names(df_a)
  names_b <- names(df_b)

  # Add columns from b not in a to a
  for (names in setdiff(names_b, names_a)) {
    if (nrow(df_a) == 0) {
      df_a[[names]] <- as.character()
    } else {
      df_a[[names]] <- NA
    }
  }

  # Reorder
  df_a <- dplyr::relocate(df_a, all_of(names_b))

  # Align types
  for (name in names_b) {
    df_a[[name]] <- as(df_a[[name]], class(df_b[[name]]))
  }

  df_a
}

#' Drop rows with only NA values
#'
#' @param df Data frame
#'
#' @returns A data frame
#' @keywords internal
drop_blank_rows <- function(df) {
  dplyr::filter(df, if_any(everything(), ~ !is.na(.)))
}

#' Create a tibble with dummy values
#'
#' @param dummy_value Value to set in all columns
#' @keywords internal
create_dummy_tibbles <- function(dummy_value) {
  function(columns) {
    tibble::tibble(
      !!!setNames(rep(list(dummy_value), length(columns)), columns)
    )
  }
}

na_tibble <- create_dummy_tibbles(NA)
character0_tibble <- create_dummy_tibbles(character(0))
asterisk_tibble <- create_dummy_tibbles("*")

#' Download parquet or PX-file and return path
#'
#' Assumes the file is .px, or parquet.
#'
#' @param url
#'
#' @returns Character
#' @keywords internal
download_px_or_parquet_and_return_path <- function(url) {
  is_parquet <- stringr::str_detect(url, "parquet")

  tmp_file_path <- ifelse(is_parquet,
    temp_parquet_file(),
    temp_px_file()
  )

  curl::curl_download(url, tmp_file_path, quiet = TRUE)

  tmp_file_path
}

test_that("strings are quoted", {
  input <- c("hi", ' " ', "single'quote")
  expect <- c('"hi"', '" " "', '"single\'quote"')

  expect_equal(str_quote(input), expect)
})

test_that("keywords has language", {
  expect_equal(
    add_language_to_keyword(
      keyword = rep("STUB", 4),
      main_language = "en",
      language = c("en", "da", "kl", NA)
    ),
    c("STUB", "STUB[da]", "STUB[kl]", "STUB")
  )
})

test_that("keyword has sub-key", {
  expect_equal(
    add_sub_key_to_keyword(
      keyword = c(
        "ELIMINATION",
        "ELIMINATION[da]",
        "ELIMINATION[kl]"
      ),
      name = c(
        "place of birth",
        "fødested",
        NA
      )
    ),
    c(
      'ELIMINATION("place of birth")',
      'ELIMINATION[da]("fødested")',
      "ELIMINATION[kl]"
    )
  )
})

test_that("keyword has cell name", {
  expect_equal(
    add_cell_to_keyword(
      keyword = c(
        'PRECISION("measure")',
        'PRECISION[da]("measure")',
        "UNIT"
      ),
      name = c("Life expectancy", NA, "Alive")
    ),
    c(
      'PRECISION("measure","Life expectancy")',
      'PRECISION[da]("measure")',
      'UNIT("Alive")'
    )
  )
})

test_that("strings are quoted", {
  input <- c(
    "A", " b ", '"', 0, 3.14, "YES", "NO", '"is quoted"', 'TLIST(A1,"2015";'
  )
  expect <- c(
    '"A"',
    '" b "',
    '"""',
    '"0"',
    '"3.14"',
    "YES",
    "NO",
    '"is quoted"',
    'TLIST(A1,"2015";'
  )

  expect_equal(quote_unless_yes_no(input), expect)
})

test_that("list values are unique and sorted", {
  lst <- list(y = c(2, 1, 3, 3, 3), x = "a", z = c(rep("b", 4), "a"))

  expect <- list(x = "a", y = c(1, 2, 3), z = c("a", "b"))

  expect_equal(lst_distinct_and_arrange(lst), expect)
  expect_equal(lst_distinct_and_arrange(list()), list())
})

test_that("lists are merged and sorted", {
  lst1 <- list(x = 1:2, y = "a")
  lst2 <- list(x = 2:3, y = c("b", "c"))

  expect <- list(x = c(1, 2, 3), y = c("a", "b", "c"))

  expect_equal(merge_named_lists(lst1, lst2), expect)

  expect_equal(merge_named_lists(list(), list()), list())

  lst4 <- list(x = c("a", "b", "c"), y = c("d", "e", "f"))
  expect_equal(merge_named_lists(lst4, lst4), lst4)

  lst5 <- list(y = c("f", "e", "d"), x = c("c", "b", "a"))
  expect_equal(merge_named_lists(lst4, lst5), lst4)

  expect_equal(merge_named_lists(list(), lst4), lst4)
  expect_equal(merge_named_lists(lst4, list()), lst4)

  lst6 <- list(a = "b")
  lst6$a <- NULL

  expect_equal(merge_named_lists(lst4, lst6), lst4)

  lst7 <- list(a = "b")
  lst8 <- list(a = "c")
  expect_equal(merge_named_lists(lst7, lst8), list(a = c("b", "c")))

  lst9 <- list(b = "1")
  lst10 <- list(a = "x", b = "1")
  expect_equal(merge_named_lists(lst9, lst10), list(a = "x", b = "1"))
})

test_that("Time values are classified", {
  test_equal <- function(values, expect) {
    expect_equal(get_timeval_type_from_values(values), expect)
  }

  test_equal(c("1995", "1996", "1997"), "A")
  test_equal(c("1995H1", "1995H2", "1996H1"), "H")
  test_equal(c("1995Q1", "1995Q2", "1995Q3"), "Q")
  test_equal(c("1995M01", "1995M02", "1995M03"), "M")
  test_equal(c("1995W01", "1995W02", "1005W03"), "W")

  test_equal(c("", NA, NULL, "2001Q1"), "Q")
})

test_that("Can get time values from string", {
  timeval1 <- as.character(2020:2024)
  timeval2 <- c("2021Q1", "2021Q2", "2021Q3", "2021Q4", "2022Q1")

  expect_values_are_preserved <- function(str) {
    expect_equal(str, get_values_from_time_format(format_time_values(str)))
  }

  expect_values_are_preserved(timeval1)
  expect_values_are_preserved(timeval2)
})

test_that("TIMEVAL short syntax is expanded", {
  timeval1 <- 'TLIST(A1, "1994"-"1996")'
  expect1 <- as.character(1994:1996)

  expect_equal(
    get_values_from_time_format(timeval1),
    expect1
  )

  timeval2 <- 'TLIST(Q1, "19951"-"19962")'
  expect2 <- c("1995Q1", "1995Q2", "1995Q3", "1995Q4", "1996Q1")

  expect_equal(
    get_values_from_time_format(timeval2),
    expect2
  )
})

test_that("Vectors are zipped", {
  expect_equal(
    zip_vectors(c(1, 3, 5), c(2, 4, 6)),
    c(1, 2, 3, 4, 5, 6)
  )
})

test_that("long lines are split into <=256 character bits", {
  expect_equal(break_long_lines(""), "")
  expect_equal(break_long_lines(NA), "")
  expect_equal(break_long_lines(NULL), "")

  input1 <- '"a", "b", "c"'
  expect1 <- '"a", "b", "c"'
  expect_equal(break_long_lines(input1), expect1)

  str_px_values_vector <- function(str) {
    str_quote(paste0(str, collapse = '","'))
  }

  input2 <- str_px_values_vector(1:104)
  expect2 <- c(
    paste0(str_px_values_vector(1:53), ","),
    paste0(str_px_values_vector(54:103), ","),
    str_px_values_vector(104)
  )
  expect_equal(break_long_lines(input2), expect2)

  input3 <- '"string which is a single value"'
  expect3 <- c(
    '"string which is a"',
    '" single value"'
  )
  expect_equal(break_long_lines(input3, max_line_length = 20), expect3)

  input4 <- paste0(
    '"a long string","composed of multiple values",',
    '"some of which are longer than the line limit of 50 characters",',
    '"and here are","a few", "short","values"'
  )

  expect4 <- c(
    '"a long string","composed of multiple values",',
    '"some of which are longer than the line limit of"',
    '" 50 characters","and here are","a few", "short",',
    '"values"'
  )
  expect_equal(break_long_lines(input4, max_line_length = 50), expect4)
})

test_that("Variables are converted to lists", {
  input <- tidyr::tibble(
    x = letters[1:3],
    y = c(1, 2, NA)
  )

  expect <- tidyr::tibble(
    x = letters[1:3],
    y = c(list(1), list(2), list(NA_real_))
  )

  expect_equal(wrap_varaible_in_list(input, "y"), expect)
})

test_that("File extensions work", {
  expect_true(is_xlsx_file("test.xlsx"))
  expect_true(is_rds_file("test.rds"))
  expect_true(is_parquet_file("test.parquet"))
  expect_true(is_px_file("test.PX"))
  expect_false(is_xlsx_file("test.px"))
  expect_false(is_xlsx_file(NULL))
  expect_false(is_xlsx_file(TRUE))
  expect_false(is_xlsx_file(data.frame()))
})

test_that("file encoding is correct", {
  get_file_encoding_for_table <- function(table_name) {
    get_encoding_from_px_file(get_px_file_path(table_name))
  }

  expect_equal(get_file_encoding_for_table("TUX01"), "iso-8859-15")
  expect_equal(
    get_file_encoding_for_table("BEXSTA_windows_1252"), "Windows-1252"
  )

  # no encoding listed; latin1 is default
  px_file <- temp_px_file()

  px(input = get_metadata_path("BEXSTA"), data = get_data_path("BEXSTA")) |>
    px_save(path = px_file)

  expect_equal(get_encoding_from_px_file(px_file), "latin1")
})

test_that("Data frames are aligned", {
  a <- data.frame(
    a = as.character(),
    b = as.numeric(),
    d = as.character()
  )

  b <- data.frame(
    a = as.numeric(),
    b = as.character(),
    c = as.character()
  )

  expect_equal(
    align_data_frames(a, b),
    data.frame(
      a = as.numeric(),
      b = as.character(),
      c = as.character(),
      d = as.character()
    )
  )

  c <- data.frame(
    a = c("1", "2"),
    b = c(3, 4),
    d = c("5", "6")
  )

  d <- data.frame(
    a = c(7),
    b = c("8"),
    c = c("9")
  )

  expect_equal(
    align_data_frames(c, d),
    data.frame(
      a = c(1, 2),
      b = c("3", "4"),
      c = NA_character_,
      d = c("5", "6")
    )
  )
})

test_that("Dummy tibbles are create", {
  colnames1 <- c("A")

  expect_equal(
    na_tibble(columns = colnames1),
    dplyr::tibble("A" = NA)
  )

  colnames2 <- c("A", "B")

  expect_equal(
    na_tibble(columns = colnames2),
    dplyr::tibble(
      "A" = NA,
      "B" = NA
    )
  )

  expect_equal(
    character0_tibble(columns = colnames1),
    dplyr::tibble("A" = character(0))
  )

  colnames2 <- c("A", "B")

  expect_equal(
    character0_tibble(columns = colnames2),
    dplyr::tibble(
      "A" = character(0),
      "B" = character(0)
    )
  )

  expect_equal(
    na_tibble(columns = c()),
    dplyr::tibble()
  )
})

test_that("PX-file is valid", {
  # Create 4 PX-files that all tests are run on
  bexsta <- create_px_file("BEXSTA")
  bexltall <- create_px_file("BEXLTALL")
  fotest <- create_px_file("FOTEST")
  no_timeval_or_codes <- create_px_file("no_timeval_or_codes")

  test_that("px lines matches regexp", {
    keywords <- px_keywords |> dplyr::pull(keyword)

    valid_lines <-
      c(
        paste0("^", keywords, "[=\\[\\(]"), # keyword followed by [ ( or =
        '^".+[",;]$', # continued lines (when previous is longer than 256)
        '^[e[:digit:][:space:]"-.]+$', # data lines
        "^;$" # last line of file
      )

    regex <- paste0(valid_lines, collapse = "|")

    expect_no_invalid_lines <- function(path) {
      px_lines <- readLines(path)

      invalid_lines <-
        px_lines[stringr::str_detect(px_lines, regex, negate = TRUE)]

      expect_equal(invalid_lines, character(0))
    }

    expect_no_invalid_lines(bexsta)
    expect_no_invalid_lines(bexltall)
    expect_no_invalid_lines(fotest)
    expect_no_invalid_lines(no_timeval_or_codes)
  })

  test_that("Codes are defined for all values", {
    expect_code_and_values_match <- function(path) {
      px_lines <- read_px_file(path)

      value_lines <- px_lines[stringr::str_detect(px_lines, "^VALUES")]
      code_lines <- px_lines[stringr::str_detect(px_lines, "^CODES")]

      # Equal number of variables
      expect_equal(length(value_lines), length(code_lines))

      # Equal number of values for each variable
      expect_equal(
        stringr::str_count(value_lines, '","'),
        stringr::str_count(code_lines, '","')
      )
    }

    expect_code_and_values_match(bexsta)
    expect_code_and_values_match(bexltall)
    expect_code_and_values_match(fotest)
    expect_code_and_values_match(no_timeval_or_codes)
  })

  test_that("header lines doesn't exceed 256 characters", {
    expect_no_too_long_lines <- function(path) {
      px_lines <- read_px_file(path)

      data_line_index <- stringr::str_which(px_lines, "^DATA=$")

      long_lines <-
        tibble::tibble(line = px_lines[1:data_line_index]) |>
        dplyr::mutate(
          text = stringr::str_extract(line, "(?<==).+"),
          length = nchar(text)
        ) |>
        dplyr::filter(length > 256)

      expect_equal(long_lines, dplyr::filter(long_lines, FALSE))
    }

    expect_no_too_long_lines(bexsta)
    expect_no_too_long_lines(bexltall)
    expect_no_too_long_lines(fotest)
    expect_no_too_long_lines(no_timeval_or_codes)
  })

  test_that("Value YES and NO are never quoted", {
    expect_no_lines_with_quoted_yes_no <- function(path) {
      px_lines <- readLines(path)

      invalid_lines <- px_lines[stringr::str_detect(px_lines, '=\"(YES|NO)\"')]

      expect_equal(invalid_lines, character(0))
    }

    expect_no_lines_with_quoted_yes_no(bexsta)
    expect_no_lines_with_quoted_yes_no(bexltall)
    expect_no_lines_with_quoted_yes_no(fotest)
    expect_no_lines_with_quoted_yes_no(no_timeval_or_codes)
  })

  test_that("timevals are added", {
    expect_that_px_filehas_timeval <- function(path) {
      px_lines <- readLines(path)

      timeval_lines <- px_lines[stringr::str_detect(px_lines, "^TIMEVAL")]

      expect_true(length(timeval_lines) > 0)
    }

    expect_that_px_filehas_timeval(bexsta)
    expect_that_px_filehas_timeval(bexltall)
    expect_that_px_filehas_timeval(fotest)
  })

  test_that("axis-version is quoted", {
    expect_that_axis_version_is_quoted <- function(path) {
      px_lines <- readLines(path)

      n_quotes <-
        stringr::str_count(
          px_lines[stringr::str_detect(px_lines, "^AXIS-VERSION")],
          '"'
        )

      expect_equal(n_quotes, 2)
    }

    expect_that_axis_version_is_quoted(bexsta)
    expect_that_axis_version_is_quoted(bexltall)
    expect_that_axis_version_is_quoted(fotest)
  })

  test_that("VARIABLECODE has variable", {
    expect_that_variablecode_has_varaible <- function(path) {
      px_lines <- readLines(path)

      n_parentheses <-
        stringr::str_count(
          px_lines[stringr::str_detect(px_lines, "^VARIABLECODE")],
          "\\("
        )

      expect_true(all(n_parentheses > 0))
    }

    expect_that_variablecode_has_varaible(bexsta)
  })

  expect_true(TRUE) # needed to run
})

#' Get data cube used in PX-file format
#'
#' @inheritParams sort_metadata_df
#' @inheritParams format_data_df
#'
#' @returns A data frame
#' @keywords internal
get_data_cube <- function(metadata_df, data_df) {
  metadata_df <- add_main_language(metadata_df)

  metadata_df_main_language <-
    metadata_df |>
    dplyr::filter(.data$main_language)

  labels <-
    metadata_df_main_language |>
    dplyr::filter(.data$keyword == "VARIABLECODE") |>
    tidyr::unnest("value") |>
    dplyr::select("label" = "variable", "variable" = "value")

  stub_and_heading_df <-
    metadata_df_main_language |>
    dplyr::filter(.data$keyword %in% c("STUB", "HEADING")) |>
    tidyr::unnest("value") |>
    dplyr::select("keyword", "label" = "value") |>
    dplyr::left_join(labels, by = "label") |>
    dplyr::mutate(variable = ifelse(is.na(.data$variable),
      .data$label,
      .data$variable
    ))

  stub_vars <-
    stub_and_heading_df |>
    dplyr::filter(.data$keyword == "STUB") |>
    dplyr::pull(.data$variable) |>
    intersect(names(data_df))

  heading_vars <-
    stub_and_heading_df |>
    dplyr::filter(.data$keyword == "HEADING") |>
    dplyr::pull(.data$variable) |>
    intersect(names(data_df))

  head_stub_variable_names <- c(heading_vars, stub_vars)

  figures_var <- setdiff(names(data_df), head_stub_variable_names)

  cells <-
    metadata_df |>
    dplyr::filter(.data$keyword == "CODES", .data$main_language) |>
    dplyr::select("keyword", "variable", "value") |>
    tidyr::unnest("value") |>
    dplyr::rename("label" = "variable") |>
    dplyr::group_by("label") |>
    dplyr::mutate(sortorder = dplyr::row_number()) |>
    dplyr::ungroup() |>
    dplyr::left_join(labels, by = "label") |>
    dplyr::filter(.data$variable %in% names(data_df)) |>
    dplyr::select("variable", "sortorder", "code" = "value")

  # Complete data by adding all combinations of variable values in data and
  # cells
  data_values <-
    data_df |>
    mutate_all_vars_to_character() |>
    dplyr::select(dplyr::all_of(head_stub_variable_names)) |>
    lst_distinct_and_arrange()

  cells_values <-
    split(cells$code, cells$variable) |>
    lst_distinct_and_arrange()

  data_and_cells_values <- merge_named_lists(data_values, cells_values)

  heading_code_vars <- paste0("code_", heading_vars, recycle0 = TRUE)
  heading_sortorder_vars <- paste0("sortorder_", heading_vars, recycle0 = TRUE)
  stub_code_vars <- paste0("code_", stub_vars, recycle0 = TRUE)
  stub_sortorder_vars <- paste0("sortorder_", stub_vars, recycle0 = TRUE)

  data_completed <-
    data_df |>
    mutate_all_vars_to_character() |>
    tidyr::complete(!!!data_and_cells_values)

  # Add sortorder variables one-by-one
  for (v in head_stub_variable_names) {
    # Add sortorder using base R syntax because
    # it uses less memory than dplyr::left_join
    cells_v <- cells[cells$variable == v, ]
    sortorder_index <- match(data_completed[[v]], cells_v$code)
    data_completed[[paste0("sortorder_", v)]] <-
      cells_v$sortorder[sortorder_index]
  }

  data_cube <-
    data_completed |>
    dplyr::rename_with(
      \(x) paste0("code_", x),
      all_of(head_stub_variable_names)
    ) |>
    # Sort by sortorder for first heading var, codes for first heading var,
    # sortorder for second heading var, etc.
    dplyr::arrange(
      dplyr::across(zip_vectors(heading_sortorder_vars, heading_code_vars))
    ) |>
    dplyr::select(-all_of(heading_sortorder_vars)) |>
    (\(.) {
      if (length(heading_code_vars) > 0) {
        tidyr::pivot_wider(.,
          names_from = all_of(heading_code_vars),
          values_from = all_of(figures_var)
        )
      } else {
        .
      }
    })() |>
    dplyr::arrange(
      dplyr::across(zip_vectors(stub_sortorder_vars, stub_code_vars))
    ) |>
    dplyr::select(-ends_with(paste0("_", stub_vars)))

  data_cube
}

#' Get lines for PX-file from px object
#'
#' @param x A px object
#'
#' @returns A character vector
#' @keywords internal
format_px_object_as_lines <- function(x) {
  metadata_df <- get_metadata_df_from_px(x)

  time_variables <- # nolint: object_usage_linter
    metadata_df |>
    dplyr::filter(.data$keyword == "TIMEVAL") |>
    dplyr::pull(.data$variable)

  metadata_lines <-
    metadata_df |>
    dplyr::left_join(
      dplyr::select(pxmake::px_keywords, "keyword", "quote_value"),
      by = "keyword"
    ) |>
    dplyr::rowwise() |>
    dplyr::mutate(keyword = .data$keyword |>
      add_language_to_keyword(
        get_main_language(metadata_df),
        .data$language
      ) |>
      add_sub_key_to_keyword(.data$variable) |>
      add_cell_to_keyword(.data$cell)) |>
    dplyr::mutate(value = .data$value |>
      tidyr::replace_na("") |>
      paste(collapse = '","') |>
      (\(.) {
        ifelse(.data$quote_value,
          quote_unless_yes_no(.),
          .
        )
      })() |>
      break_long_lines(max_line_length = 256) |>
      list()) |>
    dplyr::ungroup() |>
    tidyr::unnest("value") |>
    dplyr::mutate(
      repeated_keyword = !is.na(dplyr::lag(.data$keyword)) &
        .data$keyword == dplyr::lag(.data$keyword),
      last = is.na(dplyr::lead(.data$repeated_keyword)) |
        !dplyr::lead(.data$repeated_keyword)
    ) |>
    dplyr::mutate(line = stringr::str_c(
      ifelse(.data$repeated_keyword,
        "",
        paste0(.data$keyword, "=")
      ),
      .data$value,
      ifelse(.data$last, ";", "")
    )) |>
    dplyr::pull(.data$line)

  data_lines <-
    get_data_cube(metadata_df = metadata_df, data_df = x$data) |>
    mutate_all_vars_to_character() |>
    dplyr::mutate(
      dplyr::across(everything(), ~ tidyr::replace_na(.x, '"-"'))
    ) |>
    tidyr::unite("tmp", sep = " ") |>
    dplyr::pull(.data$tmp)

  c(metadata_lines, "DATA=", data_lines, ";")
}

#' Save px object to PX-file
#'
#' @param x A px object
#' @param path Path to save PX-file at
#'
#' @returns Nothing
#' @keywords internal
save_px_as_px_file <- function(x, path) {
  px_lines <- format_px_object_as_lines(x)

  encoding_str <-
    x$table1 |>
    dplyr::filter(.data$keyword == "CODEPAGE") |>
    dplyr::pull("value")

  if (length(encoding_str) == 0) {
    encoding_str <- get_default_encoding()
  }

  file_connection <- file(path, encoding = encoding_str)
  writeLines(px_lines, file_connection)
  close(file_connection)
}

#' Create a px object form a PX-file
#'
#' @param path Path to a PX-file
#'
#' @returns A px object
#' @keywords internal
px_from_px_file <- function(path) {
  px_lines <- read_px_file(path)

  data_line_index <- stringr::str_which(px_lines, "^DATA=$")

  error_if_not_exactly_one_data_line(data_line_index)

  metadata_lines <- px_lines[c(1:data_line_index)]
  data_lines <- px_lines[c((data_line_index + 1):length(px_lines))]

  metadata_df <-
    metadata_lines |>
    get_metadata_df_from_px_lines() |>
    add_main_language()

  variable_label <- get_variable_label(metadata_df)

  stub_heading_variables <-
    variable_label |>
    dplyr::filter(
      .data$main_language, .data$keyword %in% c("HEADING", "STUB")
    ) |>
    dplyr::arrange(desc(.data$keyword), .data$index) |>
    dplyr::pull("variable-code")

  name_relation <-
    variable_label |>
    dplyr::distinct(
      .data$`variable-code`,
      .data$language,
      .data$`variable-label`,
      .data$main_language
    )

  metadata <-
    metadata_df |>
    dplyr::rename("variable-label" = "variable") |>
    dplyr::rowwise() |>
    dplyr::mutate(`variable-label` = ifelse(.data$keyword %in% ("CONTVARIABLE"),
      unlist(.data$value),
      .data$`variable-label`
    )) |>
    dplyr::ungroup() |>
    dplyr::left_join(dplyr::select(name_relation, -"main_language"),
      by = c("language", "variable-label")
    ) |>
    dplyr::left_join(pxmake::px_keywords, by = "keyword")

  # languages
  languages <-
    metadata |>
    dplyr::filter(.data$keyword == "LANGUAGES") |>
    tidyr::unnest("value") |>
    dplyr::select("language" = "value") |>
    align_data_frames(get_base_languages())

  table_sheet_data <-
    metadata |>
    dplyr::filter(.data$table_meta) |>
    # Exclude variable specific NOTE
    dplyr::filter(
      !(.data$keyword == "NOTE" & !is.na(.data$`variable-label`))
    ) |>
    tidyr::unnest("value")

  # table1
  table1 <-
    table_sheet_data |>
    dplyr::filter(!.data$language_dependent) |>
    dplyr::select("keyword", "value") |>
    align_data_frames(get_base_table1()) |>
    sort_table1()

  # table2
  table2 <-
    table_sheet_data |>
    dplyr::filter(.data$language_dependent) |>
    dplyr::filter(
      .data$keyword != "CONTVARIABLE"
    ) |> # CONTVAR is controlled in variables1
    dplyr::select("keyword", "code" = "cell", "language", "value") |>
    align_data_frames(get_base_table2()) |>
    sort_table2(languages = languages$language)

  # variable1
  figures_variable <-
    variable_label |>
    dplyr::filter(is.na(.data$keyword)) |>
    dplyr::distinct(.data$`variable-code`) |>
    dplyr::pull("variable-code")

  if (identical(figures_variable, character(0))) {
    figures_variable <- "figures_"

    name_relation <-
      name_relation |>
      dplyr::bind_rows(tidyr::crossing(
        "variable-code" = figures_variable,
        "variable-label" = figures_variable,
        language = unique(name_relation$language)
      ))
  }

  time_var <-
    metadata |>
    dplyr::filter(.data$keyword == "TIMEVAL", .data$main_language) |>
    dplyr::pull("variable-code")

  timeval <-
    metadata |>
    dplyr::filter(.data$keyword == "TIMEVAL", .data$main_language) |>
    tidyr::unnest("value") |>
    dplyr::pull("variable-code")

  timeval_df <-
    name_relation |>
    dplyr::filter(.data$main_language) |>
    dplyr::select("variable-code") |>
    dplyr::mutate(timeval = ifelse(.data$`variable-code` %in% timeval,
      TRUE,
      FALSE
    ))

  contvariable <-
    metadata |>
    dplyr::filter(.data$keyword == "CONTVARIABLE", .data$main_language) |>
    tidyr::unnest("value") |>
    dplyr::pull("variable-code")

  contvariable_df <-
    name_relation |>
    dplyr::filter(.data$main_language) |>
    dplyr::select("variable-code") |>
    dplyr::mutate(contvariable = ifelse(.data$`variable-code` %in% contvariable,
      TRUE,
      FALSE
    ))

  variable_type_df <-
    metadata |>
    dplyr::filter(.data$main_language, .data$keyword %in% c("VARIABLE-TYPE")) |>
    tidyr::unnest("value") |>
    dplyr::filter(!toupper(.data$value) %in% c("TIMEVAL", "CONTVARIABLE")) |>
    dplyr::select("variable-code", "variable-type" = "value")

  # variables2
  variables2 <-
    metadata |>
    dplyr::filter(
      .data$keyword %in% c("NOTE", "DOMAIN", "MAP"),
      !is.na(.data$`variable-code`)
    ) |>
    dplyr::mutate(keyword = tolower(.data$keyword)) |>
    dplyr::select(
      "keyword", "language", "variable-code",
      "variable-label", "value"
    ) |>
    tidyr::unnest("value") |>
    tidyr::pivot_wider(
      names_from = "keyword",
      values_from = "value"
    ) |>
    # Add variables without NOTE, DOMAIN or MAP to get all variable-labels
    (\(current_df) {
      dplyr::bind_rows(
        current_df,
        dplyr::anti_join(
          dplyr::select(name_relation, -.data$main_language), # nolint: object_usage_linter
          current_df,
          by = c("variable-code", "language", "variable-label")
        )
      )
    })() |>
    align_data_frames(get_base_variables2()) |>
    sort_variables2(
      data_table_names = stub_heading_variables,
      languages = languages$language
    )

  # cells1, cells2
  codes <-
    metadata |>
    dplyr::filter(.data$main_language, .data$keyword %in% c("CODES")) |>
    tidyr::unnest("value") |>
    dplyr::rename("code" = "value") |>
    dplyr::group_by(.data$`variable-code`) |>
    dplyr::mutate(order = dplyr::row_number()) |>
    dplyr::ungroup() |>
    dplyr::select("variable-code", "code", "order")

  precision <-
    metadata |>
    dplyr::filter(.data$keyword == "PRECISION") |>
    tidyr::unnest("value") |>
    dplyr::rename(
      "value" = "cell",
      "precision" = "value"
    ) |>
    dplyr::select("variable-code", "value", "precision")

  values <-
    metadata |>
    dplyr::filter(.data$keyword %in% c("VALUES")) |>
    tidyr::unnest("value") |>
    dplyr::group_by(.data$`variable-code`, .data$language) |>
    dplyr::mutate(order = dplyr::row_number()) |>
    dplyr::mutate(`variable-code` = ifelse(is.na(.data$`variable-code`),
      .data$`variable-label`,
      .data$`variable-code`
    )) |>
    dplyr::select(
      "variable-code",
      "value",
      "language",
      "main_language",
      "order"
    )

  valuenote <-
    metadata |>
    dplyr::filter(.data$keyword %in% c("VALUENOTE")) |>
    tidyr::unnest("value") |>
    dplyr::select("variable-code", "language",
      "value" = "cell",
      "valuenote" = "value",
      "language"
    )

  codes_and_values <-
    codes |>
    # Use main language values as codes, if codes are missing
    dplyr::full_join(values,
      by = c("variable-code", "order"),
      multiple = "all"
    ) |>
    dplyr::full_join(valuenote, by = c("variable-code", "value", "language")) |>
    dplyr::group_by(.data$`variable-code`, .data$order) |>
    dplyr::mutate(code = ifelse(is.na(.data$code),
      .data$value[.data$main_language],
      .data$code
    )) |>
    dplyr::ungroup()

  cells <-
    codes_and_values |>
    dplyr::select(-"main_language") |>
    dplyr::left_join(precision, by = c("variable-code", "value"))

  cells1 <-
    cells |>
    dplyr::distinct(
      .data$`variable-code`, .data$code, .data$order, .data$precision
    ) |>
    align_data_frames(get_base_cells1()) |>
    sort_cells1(data_table_names = stub_heading_variables)

  cells2 <-
    cells |>
    dplyr::select("variable-code", "code", "language", "value", "valuenote") |>
    align_data_frames(get_base_cells2()) |>
    sort_cells2(
      data_table_names = stub_heading_variables,
      languages = languages$language
    )

  # variables 1
  # elimination
  elimination_df <-
    metadata |>
    dplyr::filter(.data$main_language, .data$keyword %in% c("ELIMINATION")) |>
    tidyr::unnest("value") |>
    dplyr::left_join(dplyr::filter(codes_and_values, .data$main_language),
      by = c("variable-code", "value")
    ) |>
    dplyr::mutate(code = ifelse(is.na(.data$code),
      .data$value,
      .data$code
    )) |>
    dplyr::select("variable-code", "elimination" = "code")

  variables1 <-
    variable_label |>
    dplyr::distinct(.data$`variable-code`,
      "pivot" = .data$keyword,
      "order" = .data$index
    ) |>
    dplyr::left_join(variable_type_df, by = "variable-code") |>
    dplyr::left_join(contvariable_df, by = "variable-code") |>
    dplyr::left_join(timeval_df, by = "variable-code") |>
    dplyr::left_join(elimination_df, by = "variable-code") |>
    dplyr::filter(!.data$`variable-code` %in% figures_variable) |>
    dplyr::bind_rows(dplyr::tibble(
      `variable-code` = figures_variable,
      pivot = "FIGURES",
      contvariable = FALSE,
      timeval = FALSE
    )) |>
    align_data_frames(get_base_variables1()) |>
    sort_variables1()

  # acrosscells
  acrosscells_variables <- intersect(
    unique(metadata$keyword),
    c("CELLNOTE", "CELLNOTEX")
  )

  acrosscells <-
    metadata |>
    dplyr::filter(.data$keyword %in% acrosscells_variables) |>
    dplyr::mutate(keyword = tolower(.data$keyword)) |>
    tidyr::pivot_wider(names_from = "keyword", values_from = "value") |>
    dplyr::mutate(
      `variable-label` = stringr::str_remove_all(.data$`variable-label`, '"')
    ) |>
    tidyr::separate_wider_delim(
      cols = "variable-label",
      delim = ",",
      names = stub_heading_variables
    ) |>
    dplyr::select(all_of(c(
      stub_heading_variables,
      "language",
      tolower(acrosscells_variables)
    ))) |>
    align_data_frames(get_base_acrosscells(stub_heading_variables)) |>
    dplyr::mutate(
      across(tolower(acrosscells_variables), ~ dplyr::na_if(.x, "NULL"))
    )

  # data
  # Order: stub1, stub2, ..., heading1, heading2, ...
  expand_order <-
    variable_label |>
    dplyr::filter(.data$main_language) |>
    dplyr::mutate(keyword_order = dplyr::case_when(
      .data$keyword == "STUB" ~ 1,
      .data$keyword == "HEADING" ~ 2,
      TRUE ~ NA
    )) |>
    dplyr::arrange(.data$keyword_order, .data$index) |>
    dplyr::mutate(expand_order = dplyr::row_number()) |>
    dplyr::select("variable-code", "expand_order")

  if (identical(time_var, character(0))) {
    time_values_df <- tidyr::tibble()
  } else {
    time_values <-
      metadata |>
      dplyr::filter(.data$keyword %in% c("TIMEVAL"), .data$main_language) |>
      tidyr::unnest("value") |>
      dplyr::pull("value") |>
      get_values_from_time_format()

    time_values_df <- dplyr::tibble(
      "variable-code" = time_var,
      "code" = time_values
    )
  }

  stub_and_heading_values <-
    codes_and_values |>
    dplyr::filter(
      .data$main_language,
      .data$`variable-code` %in% stub_heading_variables,
      # Exclude time variable as it is already in time_values_df
      !.data$`variable-code` %in% time_var
    ) |>
    dplyr::bind_rows(time_values_df) |>
    dplyr::distinct(.data$`variable-code`, .data$code) |>
    dplyr::group_by(.data$`variable-code`) |>
    dplyr::summarise(code = list(.data$code)) |>
    dplyr::left_join(expand_order, by = "variable-code") |>
    dplyr::arrange(.data$expand_order) |>
    dplyr::select("variable-code", "code") |>
    tibble::deframe()

  figures <-
    data_lines |>
    stringr::str_replace_all(";", "") |>
    stringr::str_split(" ") |>
    unlist() |>
    stringr::str_subset("^$", negate = TRUE) |>
    tibble::enframe(name = NULL, value = figures_variable) |>
    dplyr::mutate(across(everything(), ~ suppressWarnings(as.numeric(.x))))

  data_df <-
    do.call(tidyr::expand_grid, stub_and_heading_values) |>
    dplyr::bind_cols(figures) |>
    dplyr::as_tibble()

  new_px(
    languages = languages,
    table1 = table1,
    table2 = table2,
    variables1 = variables1,
    variables2 = variables2,
    cells1 = cells1,
    cells2 = cells2,
    acrosscells = acrosscells,
    data = data_df
  )
}

# Create test data sets
#
# Different data sources are stored in 'data-raw' and formatted data sets are in
# the folder 'data'.
library(tidyverse)
library(testthat)
library(readxl)
library(pxweb)
library(haven)
library(arrow)

bexsta <-
  pxweb_get(
    url = "https://bank.stat.gl/api/v1/en/Greenland/BE/BE01/BE0120/BEXSTA.px",
    query = list(
      "place of birth" = c("T", "N", "S"),
      "gender" = "*",
      "time" = as.character(2018:2022)
    )
  ) |>
  as.data.frame(
    column.name.type = "code",
    variable.value.type = "code"
  ) |>
  rename(persons = last_col()) |>
  arrange_all()

saveRDS(bexsta, file = test_path("fixtures", "data", "BEXSTA.rds"))
write_parquet(bexsta, test_path("fixtures", "data", "BEXSTA_parquet.parquet"))

bexsta |>
  filter(`place of birth` != "T", gender != "T") |>
  write_rds(test_path("fixtures", "data", "BEXSTA_WITHOUT_TOTALS.rds"))

test_path("fixtures", "data-raw", "BEXLTALL_RAW.rds") |>
  read_rds() |>
  rename(
    `place of birth` = pob,
    nop = nop.code,
    age = age.code,
    sex = sex.code,
    calcbase = calcbase.code,
    measure = measure.code
  ) |>
  filter(is.finite(value)) |>
  group_by(age) |>
  saveRDS(file = test_path("fixtures", "data", "BEXLTALL.rds"))

mandatory_keywords <- function() {
  pxmake::px_keywords |>
    dplyr::filter(.data$mandatory) |>
    dplyr::pull("keyword")
}

keywords_indexed_by_contvariable <- function() {
  pxmake::px_keywords |>
    dplyr::filter(.data$indexed_by_contvariable) |>
    dplyr::pull("keyword")
}

language_dependant_keywords <- function() {
  pxmake::px_keywords |>
    dplyr::filter(.data$language_dependent) |>
    dplyr::pull("keyword")
}

default_value <- function(keyword) {
  pxmake::px_keywords |>
    dplyr::filter(.data$keyword == !!keyword) |>
    dplyr::pull("default_value")
}

#' @keywords internal
#' @aliases pxmake-package.R
"_PACKAGE"

## usethis namespace: start
#' @importFrom dplyr across
#' @importFrom dplyr all_of
#' @importFrom dplyr any_of
#' @importFrom dplyr desc
#' @importFrom dplyr ends_with
#' @importFrom dplyr everything
#' @importFrom dplyr if_any
#' @importFrom dplyr last_col
#' @importFrom dplyr one_of
#' @importFrom dplyr starts_with
#' @importFrom dplyr where
#' @importFrom magrittr %>%
#' @importFrom methods as
#' @importFrom rlang :=
#' @importFrom stats na.omit
#' @importFrom stats setNames
#' @importFrom utils head
#' @importFrom utils tail
## usethis namespace: end
NULL

excel_sheet_exists <- function(sheet, excel_path) {
  sheet %in% readxl::excel_sheets(excel_path)
}

#' Get specific sheet from Excel workbook
#'
#' @param sheet String. Sheet to read.
#' @param add_automatically Logical. If TRUE, return an empty data frame if the
#' sheet does not exist.
#'
#' @returns A data frame.
#' @keywords internal
get_excel_sheet <- function(sheet, add_automatically = FALSE) {
  function(excel_path) {
    if (add_automatically) {
      if (!excel_sheet_exists(sheet, excel_path)) {
        return(data.frame())
      }
    } else {
      error_if_excel_sheet_does_not_exist(sheet, excel_path)
    }

    readxl::read_xlsx(excel_path, sheet) |>
      drop_blank_rows()
  }
}

get_table_sheet <- get_excel_sheet("Table")
get_table2_sheet <- get_excel_sheet("Table2")
get_variables_sheet <- get_excel_sheet("Variables")
get_cells_sheet <- get_excel_sheet("Cells")
get_acrosscells_sheet <- get_excel_sheet(
  "Acrosscells",
  add_automatically = TRUE
)
get_data_sheet <- get_excel_sheet("Data", add_automatically = FALSE)

#' Get figures variable from Excel workbook
#'
#' @inheritParams px_from_excel
#'
#' @returns Character, name of figures variable.
#' @keywords internal
get_figures_variable_from_excel <- function(excel_path) {
  figures_variable <-
    excel_path |>
    get_variables_sheet() |>
    dplyr::filter(toupper(.data$pivot) == "FIGURES") |>
    dplyr::distinct(.data$`variable-code`) |>
    dplyr::pull(.data$`variable-code`)

  error_if_not_exactly_one_figures_variable(figures_variable)

  figures_variable
}

#' Create a px object from an Excel workbook
#'
#' @param excel_path Path to Excel metadata workbook.
#' @param data A data frame if data isen't stored in the Excel workbook.
#'
#' @returns A px object.
#' @keywords internal
px_from_excel <- function(excel_path, data = NULL) {
  # languages, table1
  table_sheet <-
    get_table_sheet(excel_path) |>
    dplyr::filter(!is.na(.data$keyword))

  languages <-
    table_sheet |>
    dplyr::filter(.data$keyword %in% c("LANGUAGES")) |>
    dplyr::mutate(
      value = stringr::str_replace_all(.data$value, " ", "") |>
        # remove quotes to be backwards compatible
        stringr::str_replace_all('"', "") |>
        stringr::str_split(pattern = ",")
    ) |>
    tidyr::unnest("value") |>
    tidyr::drop_na("value") |>
    dplyr::select("language" = "value") |>
    align_data_frames(get_base_languages())

  table1 <-
    table_sheet |>
    align_data_frames(get_base_table1()) |>
    dplyr::filter(!.data$keyword %in% c("LANGUAGES")) |>
    sort_table1()

  # table2
  table2 <-
    excel_path |>
    get_table2_sheet() |>
    dplyr::filter(!is.na(.data$keyword)) |>
    tidyr::pivot_longer(
      cols = ends_with("_value"),
      names_to = c("language"),
      names_pattern = "^([[:alpha:]]+)_.*$"
    ) |>
    align_data_frames(get_base_table2()) |>
    sort_table2(languages = languages$language)

  # variables1, variables2
  variables_sheet <-
    excel_path |>
    get_variables_sheet()

  variables1 <-
    variables_sheet |>
    align_data_frames(get_base_variables1()) |>
    sort_variables1() |>
    dplyr::select(get_base_variables1() |> names())

  # data_df, variables2, cells1, cells2
  if (is.null(data)) {
    if (excel_sheet_exists("Data", excel_path)) {
      data <- get_data_sheet(excel_path)
    } else {
      # Empty dummy data set
      data <-
        variables_sheet |>
        dplyr::pull("variable-code") |>
        rlang::rep_named(list(as.character())) |>
        dplyr::as_tibble()
    }
  }

  data_df <- format_data_df(data,
    figures_variable = get_figures_variable_from_excel(excel_path)
  )

  variables2 <-
    variables_sheet |>
    dplyr::select(-all_of(intersect(
      setdiff(
        get_base_variables1() |> names(),
        "variable-code"
      ),
      names(variables_sheet)
    ))) |>
    tidyr::pivot_longer(
      cols = -c("variable-code"),
      names_to = c("language", "keyword"),
      names_pattern = "^([[:alpha:]]+)_(.*)$"
    ) |>
    tidyr::pivot_wider(names_from = "keyword") |>
    align_data_frames(get_base_variables2()) |>
    dplyr::mutate(`variable-label` = ifelse(is.na(.data$`variable-label`),
      .data$`variable-code`,
      .data$`variable-label`
    )) |>
    sort_variables2(
      data_table_names = names(data_df),
      languages = languages$language
    )

  cells_sheet <-
    excel_path |>
    get_cells_sheet()

  cells1 <-
    cells_sheet |>
    align_data_frames(get_base_cells1()) |>
    dplyr::select("variable-code", "code", "order" = "sortorder", "precision")

  cells2 <-
    cells_sheet |>
    dplyr::select(-"sortorder", -"precision") |>
    tidyr::pivot_longer(
      cols = ends_with(c("_code-label", "_valuenote")),
      names_to = c("language", "keyword"),
      names_pattern = "^([[:alpha:]]+)_(.*)$"
    ) |>
    tidyr::pivot_wider(names_from = "keyword") |>
    dplyr::rename("value" = "code-label") |>
    align_data_frames(get_base_cells2())

  # acrosscells
  stub_heading_variables <-
    variables1 |>
    dplyr::filter(toupper(.data$pivot) %in% c("STUB", "HEADING")) |>
    dplyr::pull("variable-code")

  acrosscells <-
    excel_path |>
    get_acrosscells_sheet() |>
    (\(.) {
      if (ncol(.) != 0) {
        . |>
          tidyr::pivot_longer(
            cols = ends_with(c("cellnote", "cellnotex")),
            names_to = c("language", "keyword"),
            names_pattern = "^([[:alpha:]]+)_(.*)$"
          ) |>
          tidyr::pivot_wider(names_from = "keyword")
      } else {
        .
      }
    })() |>
    align_data_frames(get_base_acrosscells(stub_heading_variables))


  new_px(
    languages = languages,
    table1 = table1,
    table2 = table2,
    variables1 = variables1,
    variables2 = variables2,
    cells1 = cells1,
    cells2 = cells2,
    acrosscells = acrosscells,
    data = data_df
  )
}

#' Add a data frame as a sheet to an Excel workbook
#'
#' @param wb An Excel workbook
#' @param df A data frame
#' @param sheet_name Name of the sheet
#'
#' @returns Nothing
#' @keywords internal
add_excel_sheet <- function(wb, df, sheet_name) {
  openxlsx::addWorksheet(wb, sheet_name, gridLines = FALSE)
  options("openxlsx.maxWidth" = 40)
  openxlsx::setColWidths(
    wb, sheet_name,
    cols = seq_len(ncol(df)), widths = "auto"
  )
  openxlsx::writeDataTable(wb, sheet_name, df, tableStyle = "TableStyleLight9")
}


#' Save px object as an Excel workbook
#'
#' @param x A px object
#' @param path Path to save Excel workbook
#' @inheritParams px_save
#'
#' @returns Nothing
#' @keywords internal
save_px_as_xlsx <- function(x, path, save_data, data_path) {
  excel_table <-
    data.frame(
      keyword = "LANGUAGES",
      value = paste0(x$languages$language, collapse = ",")
    ) |>
    tidyr::drop_na("value") |>
    dplyr::bind_rows(x$table1) |>
    dplyr::arrange(.data$keyword)


  excel_table2 <-
    x$table2 |>
    tidyr::pivot_wider(
      names_from = "language",
      values_from = "value",
      names_glue = "{language}_value"
    )

  excel_variables <-
    x$variables2 |>
    tidyr::pivot_longer(
      cols = -all_of(intersect(
        c(
          "variable-code",
          "language",
          "contvariable",
          "timeval"
        ),
        names(x$variables2)
      )),
      names_to = "keyword",
      values_to = "value"
    ) |>
    tidyr::pivot_wider(
      names_from = c("language", "keyword"),
      values_from = "value",
      names_glue = "{language}_{keyword}"
    ) |>
    dplyr::relocate(
      "variable-code",
      ends_with("variable-label"),
      ends_with("domain"),
      ends_with("elimination"),
      ends_with("note")
    ) |>
    dplyr::full_join(x$variables1, by = "variable-code") |>
    dplyr::relocate(names(x$variables1))

  excel_cells <-
    x$cells2 |>
    dplyr::rename("code-label" = "value") |>
    tidyr::pivot_longer(
      cols = -c("variable-code", "code", "language"),
      names_to = "keyword",
      values_to = "value"
    ) |>
    tidyr::pivot_wider(
      names_from = c("language", "keyword"),
      values_from = "value",
      names_glue = "{language}_{keyword}"
    ) |>
    dplyr::relocate(
      "variable-code",
      "code",
      ends_with("code-label"),
      ends_with("valuenote")
    ) |>
    dplyr::full_join(x$cells1, by = c("variable-code", "code")) |>
    dplyr::relocate(names(x$cells1)) |>
    dplyr::rename("sortorder" = "order")


  empty_acrosscells <-
    dplyr::bind_cols(
      dplyr::tibble("language" = defined_languages(x)),
      lapply(x$acrosscells, function(x) NA) |>
        dplyr::as_tibble() |>
        dplyr::select(-"language")
    )

  excel_acrosscells <-
    x$acrosscells |>
    dplyr::bind_rows(empty_acrosscells) |>
    tidyr::pivot_longer(
      cols = setdiff(names(get_base_acrosscells()), "language"),
      names_to = "keyword",
      values_to = "value"
    ) |>
    tidyr::pivot_wider(
      names_from = c("language", "keyword"),
      values_from = "value",
      names_glue = "{language}_{keyword}"
    ) |>
    dplyr::relocate(ends_with("cellnote"),
      ends_with("cellnotex"),
      .after = last_col()
    ) |>
    drop_blank_rows()


  ### Make sheets in workbook
  wb <- openxlsx::createWorkbook()

  add_excel_sheet(wb, excel_table, "Table")
  add_excel_sheet(wb, excel_table2, "Table2")
  add_excel_sheet(wb, excel_variables, "Variables")
  add_excel_sheet(wb, excel_cells, "Cells")
  add_excel_sheet(wb, excel_acrosscells, "Acrosscells")

  if (save_data) {
    if (is.null(data_path)) {
      error_if_too_many_rows_for_excel(x$data)
      add_excel_sheet(wb, x$data, "Data")
    } else if (is_rds_file(data_path)) {
      saveRDS(x$data, data_path)
    } else if (is_parquet_file(data_path)) {
      arrow::write_parquet(x$data, data_path)
    } else {
      unexpected_error()
    }
  }

  openxlsx::saveWorkbook(wb, path, overwrite = TRUE)
}

test_that("px_save basic functionality", {
  x <- px(get_data_path("BEXSTA"))

  # Runs without errors
  px_save(x, temp_xlsx_file())
  px_save(x, temp_px_file())
  px_save(x, temp_r_file())

  px(population_gl) |>
    px_save(path = temp_r_file(), data_path = temp_rds_file())

  px(population_gl) |>
    px_save(path = temp_r_file(), data_path = temp_parquet_file())

  expect_error(px_save(x), regexp = 'argument "path" is missing')
  expect_error(px_save(x, temp_rds_file()), regexp = "Argument 'path'")
})

test_that("px_save save_data and data_path arguments works", {
  df <- readRDS(get_data_path("BEXLTALL"))

  set.seed(1)

  df_large <-
    purrr::map(1:20, ~df) |>
    dplyr::bind_rows() |>
    dplyr::ungroup() |>
    dplyr::mutate(
      row_number = dplyr::row_number()
    ) |> # add to avoid duplicated data
    dplyr::relocate(row_number)

  x <- px(df)

  x$data <- df_large

  expect_error(px_save(x, temp_xlsx_file()), regexp = "too many rows")

  temp_xlsx <- temp_xlsx_file()
  px_save(x = x, path = temp_xlsx, save_data = FALSE)
  expect_true(!excel_sheet_exists("Data", temp_xlsx))

  expect_error(px_save(x = x, path = temp_px_file(), save_data = FALSE),
    regexp = "'save_data' can only be"
  )

  expect_error(
    px_save(
      x = x,
      path = temp_xlsx_file(),
      save_data = FALSE,
      data_path = temp_rds_file()
    ),
    regexp = "'data_path' can only be used if 'save_data' is TRUE"
  )

  expect_error(
    px_save(x = x, path = temp_px_file(), data_path = temp_rds_file()),
    regexp = "'data_path' can only be used if 'path' is an .xlsx or .R file"
  )

  expect_error(
    px_save(x = x, path = temp_xlsx_file(), data_path = temp_px_file()),
    regexp = "'data_path' must be a path to an .rds or .parquet file"
  )

  expect_error(px_save(x = x, path = temp_xlsx_file(), save_data = "maybe"),
    regexp = "'save_data' must be TRUE or FALSE"
  )

  temp_rds <- temp_rds_file()
  px_save(x = x, path = temp_xlsx_file(), data_path = temp_rds)

  expect_true(!excel_sheet_exists("Data", temp_xlsx))
  expect_identical(x$data, readRDS(temp_rds))
})

#' @rdname px_valuenote.px
#' @export
px_valuenote <- function(x, value, validate) {
  UseMethod("px_valuenote")
}

#' @eval add_doc_keyword_function_intro("VALUENOTE")
#' @param value `r cells_param_value("VALUENOTE", "2")`
#' @eval param_validate()
#' @eval add_return_px_or_df()
#' @eval add_cells2_example(
#'  "VALUENOTE",
#'  "Counts are approximated",
#'  "Some of the figures are from 2003",
#'  "Kisitsisit ilaat 2003-imeersuupput"
#' )
#' @export
px_valuenote.px <- function(x, value, validate = TRUE) {
  handle_cells(x, value, "2", "valuenote", validate)
}


#' @rdname px_valuenotex.px
#' @export
px_valuenotex <- function(x, value, validate) {
  UseMethod("px_valuenotex")
}

#' @eval add_doc_keyword_function_intro("VALUENOTEX")
#' @param value `r cells_param_value("VALUENOTEX", "2")`
#' @eval param_validate()
#' @eval add_return_px_or_df()
#' @eval add_cells2_example(
#'  "VALUENOTEX",
#'  "Counts are approximated",
#'  "Some of the figures are from 2003",
#'  "Kisitsisit ilaat 2003-imeersuupput"
#' )
#' @export
px_valuenotex.px <- function(x, value, validate = TRUE) {
  handle_cells(x, value, "2", "valuenotex", validate)
}


#' @rdname px_values.px
#' @export
px_values <- function(x, value, validate) {
  UseMethod("px_values")
}

#' @eval add_doc_keyword_function_intro("VALUES")
#' @param value `r cells_param_value("VALUES", "2")`
#' @eval param_validate()
#' @eval add_return_px_or_df()
#' @eval add_cells2_example("VALUES", 'Year 2024', 'toddler', 'meeraaqqap')
#' @export
px_values.px <- function(x, value, validate = TRUE) {
  handle_cells(x, value, "2", "value", validate)
}

test_that("PX-file = px_save(px(PX-file))", {
  expect_px_px_save_preserves_everything(px_from_table_name("BEXSTA"))
  expect_px_px_save_preserves_everything(px_from_table_name("FOTEST"))
  expect_px_px_save_preserves_everything(
    px_from_table_name("no_timeval_or_codes")
  )
  expect_px_px_save_preserves_everything(px_from_table_name("zero_heading"))
  expect_px_px_save_preserves_everything(px_from_table_name("zero_stub"))
  expect_px_px_save_preserves_everything(
    px(get_px_file_path("semicolon_in_values"))
  )


  # Cell values with closing parenthesis )
  population_gl |>
    dplyr::mutate(across(age, ~ dplyr::case_when(
      . == "0-6" ~ "(0-6)",
      . == "7-16" ~ "(7)-16",
      . == "17-24" ~ "17-(24)",
      . == "25-64" ~ "25)-64",
      . == "65+" ~ "65(",
      TRUE ~ .
    ))) |>
    px() |>
    px_precision(dplyr::tribble(
      ~`variable-code`, ~precision,
      "age", 3
    )) |>
    px_cellnote(
      dplyr::tribble(
        ~gender, ~age, ~year, ~cellnote,
        "female", "*", "2014", "Uncertainty in ages",
        NA, "(0-6)", NA, "Note to value with parensthsis ()"
      )
    ) |>
    expect_px_px_save_preserves_everything()


  px(population_gl) |>
    # change order or data columns
    px_data(dplyr::relocate(population_gl, n)) |>
    expect_px_px_save_preserves_everything()
})

test_that("Semicolon inside value is preserved", {
  x <- px(get_px_file_path("semicolon_in_values"))

  expect_false("Public administration" %in% px_values(x)$value)
  expect_true("Public administration; defence" %in% px_values(x)$value)
})

test_that("px_valuenote() and px_valuenotex()", {
  x <-
    "BEXSTA" |>
    get_data_path() |>
    readRDS() |>
    px()

  x_lang <- px_languages(x, c("da", "kl"))

  expect_identical(px_valuenote(x), NULL)
  expect_identical(px_valuenote(x_lang), NULL)

  valuenote_df1 <- dplyr::tibble(
    `variable-code` = "gender",
    code = "K",
    language = "kl",
    valuenote = "A valuenote"
  )

  x2 <- px_valuenote(x_lang, valuenote_df1)
  expect_identical(px_valuenote(x2), valuenote_df1)

  valuenote_df2 <- dplyr::tibble(
    `variable-code` = "place of birth",
    code = "N",
    valuenote = "Second valuenote"
  )

  x3 <- px_valuenote(x_lang, valuenote_df2)

  valuenote_df2_expect <- tidyr::crossing(valuenote_df2,
    language = c("kl", "da")
  ) |>
    dplyr::relocate(valuenote, .after = last_col())

  expect_identical(px_valuenote(x3), valuenote_df2_expect)

  x4 <- px_valuenote(x_lang, NULL)
  expect_identical(px_valuenote(x4), NULL)

  valuenote_df3 <- valuenote_df2
  valuenote_df3_expect <-
    valuenote_df2 |>
    dplyr::relocate(valuenote, .after = last_col())

  x5 <- px_valuenote(x, valuenote_df3)

  expect_identical(px_valuenote(x5), valuenote_df3_expect)

  expect_error(px_valuenote(x, valuenote_df1), regex = "LANGUAGE")


  valuenotex_df1 <- dplyr::rename(valuenote_df1, valuenotex = valuenote)

  x6 <- px_valuenotex(x_lang, valuenotex_df1)

  expect_identical(px_valuenotex(x6), valuenotex_df1)
})

test_that("px_values()", {
  x <-
    "BEXSTA" |>
    get_data_path() |>
    readRDS() |>
    px()

  x_lang <- px_languages(x, c("da", "kl"))

  x1 <-
    x |>
    px_values(dplyr::tibble(
      `variable-code` = "gender",
      code = "K",
      value = "Kvinde"
    ))

  expect <-
    dplyr::tribble(
      ~`variable-code`, ~code, ~language, ~value,
      "place of birth", "N", NA_character_, "N",
      "place of birth", "S", NA_character_, "S",
      "place of birth", "T", NA_character_, "T",
      "gender", "K", NA_character_, "Kvinde",
      "gender", "M", NA_character_, "M",
      "gender", "T", NA_character_, "T",
      "time", "2018", NA_character_, "2018",
      "time", "2019", NA_character_, "2019",
      "time", "2020", NA_character_, "2020",
      "time", "2021", NA_character_, "2021",
      "time", "2022", NA_character_, "2022"
    ) |>
    dplyr::select(-language)

  expect_identical(px_values(x1), expect)

  values_df <-
    dplyr::tribble(
      ~`variable-code`, ~code, ~language, ~value,
      "time", "2018", "da", "år 2018",
      "time", "2018", "kl", "ukiumi 2018-m"
    )

  value_df_expect <-
    values_df |>
    dplyr::bind_rows(
      px_values(x) |>
        tidyr::crossing(language = c("da", "kl")) |>
        dplyr::filter(!code %in% c(2018))
    ) |>
    sort_cells2(
      data_table_names = unique(expect$`variable-code`),
      languages = c("da", "kl")
    )

  x2 <- x_lang |> px_values(values_df)

  expect_identical(px_values(x2), value_df_expect)

  x3 <- x2 |> px_values(NULL)

  expect_identical(px_values(x3), NULL)
})

#' Format df for px format
#'
#' Turn all variables, except figures variable, into character and replace NA
#' with dash.
#'
#' @param data_df A data frame with data.
#' @param figures_variable Character. The name of the figures variable.
#'
#' @returns A data frame
#' @keywords internal
format_data_df <- function(data_df, figures_variable) {
  data_df |>
    dplyr::ungroup() |>
    dplyr::mutate(across(
      -one_of(intersect(names(data_df), figures_variable)),
      as.character
    )) |>
    dplyr::mutate(dplyr::across(
      where(is.character),
      ~ tidyr::replace_na(.x, "-")
    ))
}


#' Create a minimal px object from a data frame
#'
#' @param df A data frame
#'
#' @returns A px object
#' @keywords internal
px_from_data_df <- function(df) {
  default_language <- NA

  mandatory_table_keywords <-
    pxmake::px_keywords |>
    dplyr::filter(.data$mandatory, .data$table_meta)

  table1 <-
    mandatory_table_keywords |>
    dplyr::filter(!.data$language_dependent) |>
    dplyr::select("keyword", "value" = "default_value") |>
    align_data_frames(get_base_table1()) |>
    sort_table1()

  table2 <-
    mandatory_table_keywords |>
    dplyr::filter(.data$language_dependent) |>
    dplyr::select("keyword", "value" = "default_value") |>
    dplyr::mutate(language = default_language) |>
    align_data_frames(get_base_table2()) |>
    sort_table2(languages = default_language)

  variable_names <- names(df)

  stub_variables <- c()
  heading_variables <- c()
  figures_variable <- c()

  figures_variable <- tail(variable_names, 1)

  data_df <- format_data_df(df, figures_variable)

  if (length(variable_names) >= 3) {
    heading_variables <- head(tail(variable_names, 2), 1)
    stub_variables <- head(variable_names, length(variable_names) - 2)
  } else {
    heading_variables <- NULL
    stub_variables <- head(variable_names, length(variable_names) - 1)
  }

  variables1 <-
    dplyr::tribble(
      ~`variable-code`, ~pivot,
      stub_variables, "STUB",
      heading_variables, "HEADING",
      figures_variable, "FIGURES"
    ) |>
    tidyr::unnest("variable-code") |>
    dplyr::group_by(.data$pivot) |>
    dplyr::mutate(
      order = ifelse(.data$pivot == "FIGURES",
        NA,
        dplyr::row_number()
      ),
      contvariable = FALSE
    ) |>
    dplyr::ungroup() |>
    align_data_frames(get_base_variables1()) |>
    sort_variables1()

  variables2 <-
    variables1 |>
    dplyr::select("variable-code") |>
    dplyr::mutate(
      language = default_language,
      `variable-label` = .data$`variable-code`
    ) |>
    align_data_frames(get_base_variables2()) |>
    sort_variables2(
      data_table_names = names(df),
      languages = default_language
    )

  if (length(df) == 0) {
    cells1 <- get_base_cells1()
  } else {
    cells1 <-
      dplyr::tibble(
        `variable-code` = setdiff(names(data_df), figures_variable)
      ) |>
      dplyr::rowwise() |>
      dplyr::mutate(
        code = df[[.data$`variable-code`]] |>
          unique() |>
          sort() |>
          as.character() |>
          list()
      ) |>
      dplyr::ungroup() |>
      dplyr::filter(!is.null(.data$code)) |>
      tidyr::unnest("code") |>
      dplyr::group_by(.data$`variable-code`) |>
      dplyr::mutate(order = as.numeric(dplyr::row_number())) |>
      dplyr::ungroup() |>
      align_data_frames(get_base_cells1()) |>
      sort_cells1(data_table_names = names(data_df))
  }

  cells2 <-
    cells1 |>
    dplyr::select("variable-code", "code") |>
    dplyr::mutate(
      language = default_language,
      value = .data$code
    ) |>
    align_data_frames(get_base_cells2()) |>
    sort_cells2(
      data_table_names = names(data_df),
      languages = default_language
    )

  new_px(
    languages = get_base_languages(),
    table1 = table1,
    table2 = table2,
    variables1 = variables1,
    variables2 = variables2,
    cells1 = cells1,
    cells2 = cells2,
    acrosscells = get_base_acrosscells(c(stub_variables, heading_variables)),
    data = data_df
  ) |>
    px_title("") |>
    px_charset("ANSI")
}

#' Get languages used in px object
#'
#' @param x A px object
#'
#' @returns A character vector
#' @keywords internal
defined_languages <- function(x) {
  if (!is.null(px_languages(x))) {
    px_languages(x)
  } else if (!is.null(px_language(x))) {
    px_language(x)
  } else {
    NA_character_
  }
}

#' Change language in table
#'
#' @param df A data frame to change languages in
#' @param new_languages A character vector
#' @param keep_vars A character vector
#' @param align_df A data frame to align with
#'
#' @returns A data frame
#' @keywords internal
modify_languages_in_table <- function(df, new_languages, keep_vars, align_df) {
  add_languages <- setdiff(new_languages, df$language)

  keep <- dplyr::filter(df, .data$language %in% new_languages)

  new <-
    df |>
    dplyr::select(all_of(keep_vars)) |>
    tidyr::expand_grid(language = add_languages) |>
    align_data_frames(align_df)

  dplyr::bind_rows(keep, new)
}

#' Change language in px object
#'
#' Changes languages in 3 of the px tables
#'
#' @param x A px object
#' @param new_languages A character vector
#'
#' @returns A px object
#' @keywords internal
modify_languages_in_px <- function(x, new_languages) {
  x$table2 <-
    modify_languages_in_table(
      df = x$table2,
      new_languages = new_languages,
      keep_vars = c("keyword", "value"),
      align_df = get_base_table2()
    )

  x$variables2 <-
    modify_languages_in_table(
      df = x$variables2,
      new_languages = new_languages,
      keep_vars = c("variable-code", "variable-label"),
      align_df = get_base_variables2()
    ) |>
    sort_variables2(
      data_table_names = names(px_data(x)),
      languages = new_languages
    )

  x$cells2 <-
    modify_languages_in_table(
      df = x$cells2,
      new_languages = new_languages,
      keep_vars = c(
        "variable-code", "code", "value",
        "valuenote", "valuenotex"
      ),
      align_df = get_base_cells2()
    ) |>
    sort_cells2(
      data_table_names = names(px_data(x)),
      languages = new_languages
    )

  x$acrosscells <-
    modify_languages_in_table(
      df = x$acrosscells,
      new_languages = new_languages,
      keep_vars = c(px_stub(x), px_heading(x), "cellnote"),
      align_df = get_base_acrosscells()
    )

  x
}

#' @rdname px_language.px
#' @export
px_language <- function(x, value, validate) {
  UseMethod("px_language")
}

#' LANGUAGE
#'
#' Inspect or change LANGUAGE.
#'
#' If LANGUAGES is defined, changing LANGUAGE will also add is to LANGUAGES.
#'
#' @param x A px object
#' @param value Optional. A character string. If missing, the current LANGUAGE
#' is returned. If NULL, LANGUAGE is removed.
#' @eval param_validate()
#'
#' @seealso \code{\link{px_languages}}
#'
#' @returns A px object
#'
#' @examples
#' # Set LANGUAGE to 'en'
#' x1 <-
#'   population_gl |>
#'   px() |>
#'   px_language("en")
#'
#' # Print LANGUAGE
#' px_language(x1)
#'
#' # Remove LANGUAGE
#' x2 <- px_language(x1, NULL)
#' px_language(x2)
#'
#' @export
px_language.px <- function(x, value, validate = TRUE) {
  if (missing(value)) {
    language <- get_table1_value(x, "LANGUAGE")

    if (identical(language, character(0))) {
      return(NULL)
    } else {
      return(language)
    }
  } else if (is.null(value)) {
    return(remove_keyword_table1(x, "LANGUAGE"))
  }

  if (!is.null(px_languages(x)) && !value %in% px_languages(x)) {
    x$languages <- modify_or_add_in_column(
      x$languages, "language", px_language(x), value
    )
  }

  x <- modify_table1(x, "LANGUAGE", value)

  x <- modify_languages_in_px(x, new_languages = defined_languages(x))

  return_px(x, validate)
}


#' @rdname px_languages.px
#' @export
px_languages <- function(x, value, validate) {
  UseMethod("px_languages")
}

#' LANGUAGES
#'
#' Inspect or change LANGUAGES.
#'
#' If LANGUAGE is defined it should be one of the values in LANGUAGES.
#'
#' If LANGUAGE is set, it is considered the main language.
#' If LANGUAGE is not set, the first language in LANGUAGES is considered the
#' main language.
#'
#' @param x A px object
#' @param value Optional. A character vector. If missing, the current LANGUAGES
#' are returned. If NULL, LANGUAGES are removed.
#' @eval param_validate()
#'
#' @seealso \code{\link{px_language}}
#'
#' @returns A px object
#'
#' @examples
#' # Set LANGUAGES to 'en' and 'kl', with 'en' as main language
#' x1 <-
#'   population_gl |>
#'   px() |>
#'   px_languages(c("en", "kl"))
#'
#' # Print LANGUAGES
#' px_languages(x1)
#'
#' # Remove LANGUAGES
#' x2 <- px_languages(x1, NULL)
#' px_languages(x2)
#'
#' @export
px_languages.px <- function(x, value, validate = TRUE) {
  if (missing(value)) {
    languages <- x$languages$language

    if (identical(languages, character(0))) {
      return(NULL)
    } else {
      return(languages)
    }
  } else if (is.null(value)) {
    x$languages <- get_base_languages()
    return(x)
  }

  x$languages <-
    dplyr::tibble(language = value) |>
    align_data_frames(get_base_languages())

  x <- modify_languages_in_px(x, new_languages = value)

  return_px(x, validate)
}

test_that("LANGUAGE modified", {
  x <-
    "BEXSTA" |>
    get_data_path() |>
    readRDS() |>
    px()

  expect_true(is.null(px_language(x)))
  expect_identical(px_languages(x), NULL)

  x2 <- px_language(x, "sv")
  expect_identical(px_language(x2), "sv")
  expect_identical(px_languages(x2), NULL)

  x3 <- px_languages(x2, "sv")
  expect_identical(px_language(x3), "sv")
  expect_identical(px_languages(x3), c("sv"))

  expect_error(px_languages(x3, "en"), regexp = "LANGUAGE is not in")

  x4 <-
    x3 |>
    px_language(NULL) |>
    px_languages("en")

  expect_identical(px_language(x4), NULL)
  expect_identical(px_languages(x4), c("en"))

  x5 <- px_languages(x4, NULL)

  expect_identical(px_languages(x5), NULL)

  # Creates mandatory keywords for all languages
  expect_all_mandatory_keywords_return_non_na_values <- function(x) {
    mandatory_keyword_values <-
      mandatory_keywords() |>
      sapply(function(keyword) {
        f <- get(keyword_to_function(keyword))
        f(x)
      })

    any_na_values <- function(l) {
      if (is.data.frame(l)) {
        any(is.na(l$value))
      } else {
        is.na(l)
      }
    }

    sapply(mandatory_keyword_values, any_na_values) |>
      any() |>
      expect_false()
  }

  x6 <-
    x |>
    px_language("da")

  x7 <-
    x |>
    px_languages(c("da", "kl"))

  expect_all_mandatory_keywords_return_non_na_values(x6)
  expect_all_mandatory_keywords_return_non_na_values(x7)

  # Runs without errors
  x |>
    px_languages(c("en", "kl")) |>
    px_language("en")

  x |>
    px_languages(c("en", "kl")) |>
    px_language("dk")

  # Setting language doesn't interfere with languages
  x_before <-
    population_gl |>
    px() |>
    px_languages(c("en", "da", "kl"))

  x_after <-
    x_before |>
    px_language("en")

  expect_equal(px_languages(x_after), c("en", "da", "kl"))

  expect_identical(x_before$table2, x_after$table2)
})

test_that("Setting languages preserves VALUENOTE(X)", {
  multilingual_valuenote <-
    dplyr::tibble(
      `variable-code` = "year",
      code = "2020",
      language = c("en", "da"),
      valuenote = "great year"
    )

  valuenote_after_language_change <-
    px(population_gl) |>
    px_valuenote(
      multilingual_valuenote |>
        dplyr::select(-language) |>
        dplyr::distinct_all()
    ) |>
    px_languages(c("en", "da")) |>
    px_valuenote()

  expect_equal(
    valuenote_after_language_change,
    multilingual_valuenote
  )
})

dummy_data_path <- function() {
  "not_a_file"
}

#' Save px object as an R script
#'
#' Running the R script creates the px object.
#'
#' @param x A px object
#' @param path Path to save R script at
#' @param data_path `.rds` or `.parquest` path to save data at
#'
#' @returns Nothing
#' @keywords internal
save_px_as_r_script <- function(x, path, data_path) {
  is_list_of_lists <- function(x) {
    all(purrr::map_lgl(x, is.list))
  }

  if (!is.null(data_path)) {
    if (is_rds_file(data_path)) {
      saveRDS(x$data, data_path)
    } else if (is_parquet_file(data_path)) {
      arrow::write_parquet(x$data, data_path)
    } else {
      unexpected_error()
    }
  } else {
    # Use dummy path because normalizePath() gives errors when data_path = NULL
    data_path <- dummy_data_path()
  }

  data_code <-
    pxmake::px_keywords |>
    # Remove unimplemented functions
    dplyr::filter(.data$px_function %in% getNamespaceExports("pxmake")) |>
    # Add px_order
    dplyr::bind_rows(data.frame(
      keyword = NA_character_,
      px_function = "px_order"
    )) |>
    dplyr::rowwise() |>
    dplyr::mutate(
      value = list(eval(parse(text = paste0(.data$px_function, "(x)"))))
    ) |>
    dplyr::ungroup() |>
    dplyr::filter(!purrr::map_lgl(.data$value, is.null)) |>
    # Remove rows where value is default value
    dplyr::filter(!purrr::map2_lgl(.data$value, default_value, identical)) |>
    # Expand values that are list of lists
    dplyr::mutate(value = purrr::map(.data$value, function(x) {
      if (is_list_of_lists(x)) x else list(x)
    })) |>
    dplyr::filter(lengths(.data$value) > 0) |>
    tidyr::unnest("value") |>
    dplyr::mutate(
      value_constructor = purrr::map_chr(.data$value, convert_value_to_code)
    ) |>
    dplyr::select("keyword", "px_function", "value_constructor") |>
    dplyr::arrange(.data$keyword != "DATA") |>
    dplyr::mutate(last_row = dplyr::row_number() == dplyr::n()) |>
    dplyr::mutate(
      code = dplyr::case_when(
        keyword == "DATA" & data_path == dummy_data_path() ~
          stringr::str_glue("px(input = {.data$value_constructor}) |>"),
        keyword == "DATA" & data_path != dummy_data_path() ~
          stringr::str_glue(
            'px(input = "',
            '{normalizePath(data_path, winslash = "/", mustWork = FALSE)}',
            '") |>'
          ),
        !last_row ~
          stringr::str_glue(
            "  {.data$px_function}({.data$value_constructor}) |>"
          ),
        last_row ~
          stringr::str_glue(
            "  {.data$px_function}({.data$value_constructor})"
          )
      )
    ) |>
    dplyr::pull(code) |>
    paste(collapse = "\n")

  code <-
    c(
      "library(dplyr)",
      "library(pxmake)",
      "",
      data_code
    ) |>
    paste(collapse = "\n")

  writeLines(code, path)
}

#' Create constructing code
#'
#' Creates code that construct input value.
#'
#' @param value Vector of values to create constructors for
#'
#' @returns A character vector
#' @keywords internal
convert_value_to_code <- function(value) {
  if (is.data.frame(value)) {
    convert_df_to_code(value)
  } else if (is.character(value)) {
    if (length(value) == 1) {
      shQuote(value)
    } else {
      paste0("c(", paste0(shQuote(value), collapse = ", "), ")")
    }
  } else {
    unexpected_error()
  }
}

#' Create code to construct data frame
#'
#' Convert data frame to the code nessasary to construct it as a tibble.
#'
#' @param df A data frame
#'
#' @returns A character vector
#' @keywords internal
convert_df_to_code <- function(df) {
  col_names <-
    names(df) |>
    purrr::map_chr(function(x) {
      if (make.names(x) == x) {
        x
      } else {
        paste0("`", x, "`")
      }
    }) |>
    (\(x) paste0("~", x, collapse = ", "))()

  rows <-
    df |>
    dplyr::mutate(
      across(
        where(~ is.factor(.) | is.character(.)),
        ~ dplyr::if_else(is.na(.), "NA", shQuote(.))
      ),
      across(
        where(is.numeric),
        ~ dplyr::if_else(is.na(.), "NA", as.character(.))
      ),
    ) |>
    tidyr::unite("rows", everything(), sep = ", ") |>
    dplyr::pull(1)

  c(
    "tribble(",
    paste0("  ", col_names, ","),
    paste(" ", paste0(rows, ","), collapse = "\n"),
    "  )"
  ) |>
    paste(collapse = "\n")
}

handle_table1_keyword <- function(x, value, keyword, validate) {
  if (missing(value)) {
    return(get_table1_value(x, keyword))
  } else if (is.null(value)) {
    error_if_mandatory_keyword(x, keyword)

    x <- remove_keyword_table1(x, keyword)
  } else {
    x <- modify_table1(x, keyword, value)
  }

  return_px(x, validate)
}

#' @rdname px_aggregallowed.px
#' @export
px_aggregallowed <- function(x, value, validate) {
  UseMethod("px_aggregallowed")
}

#' @eval add_documentation_table1("AGGREGALLOWED", "NO")
px_aggregallowed.px <- function(x, value, validate = TRUE) {
  handle_table1_keyword(x, value, "AGGREGALLOWED", validate)
}


#' @rdname px_autopen.px
#' @export
px_autopen <- function(x, value, validate) {
  UseMethod("px_autopen")
}

#' @eval add_documentation_table1("AUTOPEN", "YES")
px_autopen.px <- function(x, value, validate = TRUE) {
  handle_table1_keyword(x, value, "AUTOPEN", validate)
}


#' @rdname px_axis_version.px
#' @export
px_axis_version <- function(x, value, validate) {
  UseMethod("px_axis_version")
}

#' @eval add_documentation_table1("AXIS-VERSION", "2010")
px_axis_version.px <- function(x, value, validate = TRUE) {
  handle_table1_keyword(x, value, "AXIS-VERSION", validate)
}


#' @rdname px_charset.px
#' @export
px_charset <- function(x, value, validate) {
  UseMethod("px_charset")
}

#' @eval add_documentation_table1("CHARSET", "ANSI")
px_charset.px <- function(x, value, validate = TRUE) {
  handle_table1_keyword(x, value, "CHARSET", validate)
}


#' @rdname px_codepage.px
#' @export
px_codepage <- function(x, value, validate) {
  UseMethod("px_codepage")
}

#' @eval add_documentation_table1("CODEPAGE", "utf-8")
#' @details
#' CODEPAGE controls which encoding PX-files are read and stored in.
#' Use `iconvlist()` to see available encodings on your system.
#' @seealso [px_save()]
px_codepage.px <- function(x, value, validate = TRUE) {
  handle_table1_keyword(x, value, "CODEPAGE", validate)
}


#' @rdname px_confidential.px
#' @export
px_confidential <- function(x, value, validate) {
  UseMethod("px_confidential")
}

#' @eval add_documentation_table1("CONFIDENTIAL", "1")
px_confidential.px <- function(x, value, validate = TRUE) {
  handle_table1_keyword(x, value, "CONFIDENTIAL", validate)
}


#' @rdname px_copyright.px
#' @export
px_copyright <- function(x, value, validate) {
  UseMethod("px_copyright")
}

#' @eval add_documentation_table1("COPYRIGHT", "YES")
px_copyright.px <- function(x, value, validate = TRUE) {
  handle_table1_keyword(x, value, "COPYRIGHT", validate)
}


#' @rdname px_creation_date.px
#' @export
px_creation_date <- function(x, value, validate) {
  UseMethod("px_creation_date")
}

#' @eval add_documentation_table1("CREATION-DATE", "19960612 14:20")
px_creation_date.px <- function(x, value, validate = TRUE) {
  handle_table1_keyword(x, value, "CREATION-DATE", validate)
}


#' @rdname px_decimals.px
#' @export
px_decimals <- function(x, value, validate) {
  UseMethod("px_decimals")
}

#' @eval add_documentation_table1("DECIMALS", "3")
px_decimals.px <- function(x, value, validate = TRUE) {
  handle_table1_keyword(x, value, "DECIMALS", validate)
}


#' @rdname px_descriptiondefault.px
#' @export
px_descriptiondefault <- function(x, value, validate) {
  UseMethod("px_descriptiondefault")
}

#' @eval add_documentation_table1("DESCRIPTIONDEFAULT", "YES")
px_descriptiondefault.px <- function(x, value, validate = TRUE) {
  handle_table1_keyword(x, value, "DESCRIPTIONDEFAULT", validate)
}


#' @rdname px_matrix.px
#' @export
px_matrix <- function(x, value, validate) {
  UseMethod("px_matrix")
}

#' @eval add_documentation_table1("MATRIX", "POPULATION")
px_matrix.px <- function(x, value, validate = TRUE) {
  handle_table1_keyword(x, value, "MATRIX", validate)
}


#' @rdname px_next_update.px
#' @export
px_next_update <- function(x, value, validate) {
  UseMethod("px_next_update")
}

#' @eval add_documentation_table1("NEXT-UPDATE", "20240621 15:55")
px_next_update.px <- function(x, value, validate = TRUE) {
  handle_table1_keyword(x, value, "NEXT-UPDATE", validate)
}


#' @rdname px_official_statistics.px
#' @export
px_official_statistics <- function(x, value, validate) {
  UseMethod("px_official_statistics")
}

#' @eval add_documentation_table1("OFFICIAL-STATISTICS", "YES")
px_official_statistics.px <- function(x, value, validate = TRUE) {
  handle_table1_keyword(x, value, "OFFICIAL-STATISTICS", validate)
}


#' @rdname px_showdecimals.px
#' @export
px_showdecimals <- function(x, value, validate) {
  UseMethod("px_showdecimals")
}

#' @eval add_documentation_table1("SHOWDECIMALS", "2")
px_showdecimals.px <- function(x, value, validate = TRUE) {
  handle_table1_keyword(x, value, "SHOWDECIMALS", validate)
}


#' @rdname px_subject_code.px
#' @export
px_subject_code <- function(x, value, validate) {
  UseMethod("px_subject_code")
}

#' @eval add_documentation_table1("SUBJECT-CODE", "POP")
px_subject_code.px <- function(x, value, validate = TRUE) {
  handle_table1_keyword(x, value, "SUBJECT-CODE", validate)
}


#' @rdname px_tableid.px
#' @export
px_tableid <- function(x, value, validate) {
  UseMethod("px_tableid")
}

#' @eval add_documentation_table1("TABLEID", "POPGL")
px_tableid.px <- function(x, value, validate = TRUE) {
  handle_table1_keyword(x, value, "TABLEID", validate)
}


#' @rdname px_update_frequency.px
#' @export
px_update_frequency <- function(x, value, validate) {
  UseMethod("px_update_frequency")
}

#' @eval add_documentation_table1("UPDATE-FREQUENCY", "Yearly")
px_update_frequency.px <- function(x, value, validate = TRUE) {
  handle_table1_keyword(x, value, "UPDATE-FREQUENCY", validate)
}

test_that("pxjob exists without errors (exit code 0)", {
  skip_if_not_installed("pxjob64Win", minimum_version = "1.1.0")

  expect_that_pxjob_runs_without_errors <- function(x) {
    px_path <- temp_px_file()

    px_save(x, path = px_path)

    output <- get_pxjob_file_path(rlang::hash(x))

    pxjob_exit_code <- pxjob64Win::pxjob(px_path, output)
    expect_equal(0, pxjob_exit_code)

    # File is removed manually because pxjob_clean() cannot be used because the
    # exit code from pxjob64Win::pxjob is needed.
    file.remove(output)
  }

  expect_that_pxjob_runs_without_errors(px(input = get_metadata_path("FOTEST")))
  expect_that_pxjob_runs_without_errors(px(input = get_px_file_path("TUX01")))
  expect_that_pxjob_runs_without_errors(px(input = get_data_path("BEXSTA")))
  expect_that_pxjob_runs_without_errors(
    px(input = get_px_file_path("multilingual_no_codes"))
  )

  x1 <- px(input = women)
  x1$cells2 <- get_base_cells2()

  expect_that_pxjob_runs_without_errors(x1)

  expect_that_pxjob_runs_without_errors(
    px(input = women) |> px_languages(c("en", "sv"))
  )

  get_data_path("BEXSTA") |>
    px() |>
    px_language("en") |>
    px_charset("ANSI") |>
    px_creation_date("2020-01-01 10:00") |>
    px_decimals("1") |>
    px_heading("gender") |>
    px_stub("place of birth") |>
    px_last_updated("2020-01-01 10:00") |>
    px_next_update("2020-01-01 10:00") |>
    px_matrix("bexsta") |>
    px_timeval("time") |>
    px_valuenote(dplyr::tibble(
      `variable-code` = "gender",
      code = "K",
      valuenote = "Great value"
    )) |>
    px_valuenotex(dplyr::tibble(
      `variable-code` = "time",
      code = "2020",
      valuenotex = "What a great year"
    )) |>
    px_languages(c("en", "sv")) |>
    px_note(dplyr::tibble(
      `variable-code` = "time",
      language = c("en", "sv"),
      note = c("English time note", "Svensk tidsnote")
    )) |>
    px_notex("Mandatory note for entire table") |>
    expect_that_pxjob_runs_without_errors()
})

test_that("Table1 keywords are modified and removed", {
  x <-
    "BEXSTA" |>
    get_data_path() |>
    readRDS() |>
    px() |>
    px_charset("ANSI") |>
    px_creation_date("2019-01-01 10:00") |>
    px_matrix("BEXSTA") |>
    px_decimals("1") |>
    px_last_updated("2020-01-01 10:00") |>
    px_next_update(format("2022-01-01 10:00", format = "%Y%m%d %H:%M")) |>
    px_subject_code("BEXSTA") |>
    px_axis_version("2010") |>
    px_codepage("iso-8859-1") |>
    px_confidential("1") |>
    px_copyright("YES") |>
    px_showdecimals("3") |>
    px_tableid("BEXSTAid") |>
    px_update_frequency("yearly") |>
    px_aggregallowed("YES") |>
    px_autopen("NO") |>
    px_descriptiondefault("NO") |>
    px_official_statistics("YES")

  expect_identical(px_charset(x), "ANSI")
  expect_identical(px_creation_date(x), "2019-01-01 10:00")
  expect_identical(px_matrix(x), "BEXSTA")
  expect_identical(px_decimals(x), "1")
  expect_identical(px_last_updated(x), "2020-01-01 10:00")
  expect_identical(px_next_update(x), "2022-01-01 10:00")
  expect_identical(px_subject_code(x), "BEXSTA")
  expect_identical(px_axis_version(x), "2010")
  expect_identical(px_codepage(x), "iso-8859-1")
  expect_identical(px_confidential(x), "1")
  expect_identical(px_copyright(x), "YES")
  expect_identical(px_showdecimals(x), "3")
  expect_identical(px_tableid(x), "BEXSTAid")
  expect_identical(px_update_frequency(x), "yearly")
  expect_identical(px_aggregallowed(x), "YES")
  expect_identical(px_autopen(x), "NO")
  expect_identical(px_descriptiondefault(x), "NO")
  expect_identical(px_official_statistics(x), "YES")

  x2 <-
    x |>
    px_charset(NULL) |>
    px_creation_date(NULL) |>
    px_last_updated(NULL) |>
    px_next_update(NULL) |>
    px_axis_version(NULL) |>
    px_codepage(NULL) |>
    px_confidential(NULL) |>
    px_copyright(NULL) |>
    px_showdecimals(NULL) |>
    px_tableid(NULL) |>
    px_update_frequency(NULL) |>
    px_aggregallowed(NULL) |>
    px_autopen(NULL) |>
    px_descriptiondefault(NULL) |>
    px_official_statistics(NULL)

  expect_identical(px_charset(x2), NULL)
  expect_identical(px_creation_date(x2), NULL)
  expect_identical(px_last_updated(x2), NULL)
  expect_identical(px_next_update(x2), NULL)
  expect_identical(px_axis_version(x2), NULL)
  expect_identical(px_codepage(x2), NULL)
  expect_identical(px_confidential(x2), NULL)
  expect_identical(px_copyright(x2), NULL)
  expect_identical(px_showdecimals(x2), NULL)
  expect_identical(px_tableid(x2), NULL)
  expect_identical(px_update_frequency(x2), NULL)
  expect_identical(px_aggregallowed(x2), NULL)
  expect_identical(px_autopen(x2), NULL)
  expect_identical(px_descriptiondefault(x2), NULL)
  expect_identical(px_official_statistics(x2), NULL)

  expect_error(px_matrix(x, NULL), regexp = "mandatory")
  expect_error(px_decimals(x, NULL), regexp = "mandatory")
  expect_error(px_subject_code(x, NULL), regexp = "mandatory")
})

handle_notes <- function(x, value, keyword, validate) {
  colname <- tolower(keyword)
  error_msg <- stringr::str_glue(
    "Argument 'value' has wrong format. See ?{colname}."
  )

  if (missing(value)) {
    value_table2 <- get_table2_value(x, keyword)
    value_variables2 <- get_variables2_value(x, colname)

    if (is.null(value_table2)) {
      return(value_variables2)
    } else if (is.null(value_variables2)) {
      return(value_table2)
    } else {
      return(list(value_table2, value_variables2))
    }
  }

  if (is.null(value)) {
    x <- remove_keyword_table2(x, keyword)
    x <- remove_keyword_variables2(x, keyword)
  } else if (is.character(value)) {
    x <- modify_table2(x, keyword, value)
  } else if (is.data.frame(value)) {
    if (all(c("language", "value") %in% colnames(value))) {
      x <- modify_table2(x, keyword, value)
    } else if (all(c("variable-code", colname) %in% colnames(value))) {
      x <- modify_variables2(x, colname, value)
    } else {
      error(error_msg)
    }
  } else if (is.list(value)) {
    for (i in seq_along(value)) {
      x <- get(paste0("px_", colname, ".px"))(x, value[[i]])
    }
  } else {
    error(error_msg)
  }

  return_px(x, validate)
}


#' @rdname px_note.px
#' @export
px_note <- function(x, value, validate) {
  UseMethod("px_note")
}

#' @eval add_documentation_note("NOTE")
px_note.px <- function(x, value, validate = TRUE) {
  handle_notes(x, value, "NOTE", validate)
}


#' @rdname px_notex.px
#' @export
px_notex <- function(x, value, validate) {
  UseMethod("px_notex")
}

#' @eval add_documentation_note("NOTEX")
px_notex.px <- function(x, value, validate = TRUE) {
  handle_notes(x, value, "NOTEX", validate)
}

test_that("NOTE(X) is modified", {
  x <-
    "BEXSTA" |>
    get_data_path() |>
    readRDS() |>
    px()

  expect_identical(px_note(x), NULL)
  expect_identical(px_notex(x), NULL)

  x2 <- px_note(x, value = "Note for entire table")
  expect_error(px_note(x, value = 4), regex = "wrong format")

  expect_identical(px_note(x2), "Note for entire table")
  expect_true(all(is.na(x2$variables2$note)))

  x_lang <- px_languages(x, c("da", "kl"))

  expect_error(px_note(x_lang, value = c(4, 5)), regex = "wrong format")

  expect_identical(px_note(x_lang), NULL)

  dk_note_df <- dplyr::tibble(
    language = "da",
    value = "En dansk note"
  )

  x_lang2 <- px_note(x_lang, value = dk_note_df)
  expect_error(px_note(x_lang, value = data.frame(language = "da")),
    regex = "wrong format"
  )

  expect_identical(px_note(x_lang2), dk_note_df)

  entire_table_note <- "Same note for da and kl"
  x_lang3 <- px_note(x_lang2, value = entire_table_note)

  same_note_df <- dplyr::tibble(
    language = c("da", "kl"),
    value = entire_table_note
  )

  expect_identical(px_note(x_lang3), same_note_df)

  expect_error(
    px_note(x, value = dplyr::tibble(
      language = "da",
      value = "A variable note",
      wrong_name = "invalid"
    )),
    regex = "invalid columns: wrong_name"
  )

  variable_note_df <- dplyr::tibble(
    `variable-code` = "gender",
    note = "A variable note"
  )

  x3 <- px_note(x, value = variable_note_df)

  expect_error(
    px_note(x, value = dplyr::tibble(
      `variable-code` = "gender",
      note = "A variable note",
      wrong_name = "invalid"
    )),
    regex = "invalid columns: wrong_name"
  )

  expect_identical(px_note(x3), variable_note_df)


  variable_note_lang_df <- dplyr::tibble(
    `variable-code` = "gender",
    language = c("da", "kl"),
    note = c(
      "danish note",
      "kalaallisut note"
    )
  )

  x_lang4 <- px_note(x_lang, value = variable_note_lang_df)

  expect_identical(px_note(x_lang4), variable_note_lang_df)

  x_lang5 <-
    x_lang |>
    px_note(value = entire_table_note) |>
    px_note(value = variable_note_lang_df)

  expect_identical(px_note(x_lang5), list(
    same_note_df,
    variable_note_lang_df
  ))

  x_lang6 <- px_note(x_lang, NULL)

  expect_identical(px_note(x_lang6), NULL)

  x_lang7 <- px_note(x_lang, value = list(
    variable_note_lang_df,
    same_note_df
  ))

  expect_identical(px_note(x_lang7), list(
    same_note_df,
    variable_note_lang_df
  ))

  expect_error(
    px_note(x_lang, value = list(
      c(1),
      variable_note_lang_df
    )),
    regex = "wrong format"
  )
})

#' Regexp that matches a classification file section heading
#' @keywords internal
classification_file_section_heading_regexp <- function() {
  "^\\[[^\\]]+\\]$"
}

#' Smallest larger value
#'
#' @param vec A numeric vector
#' @param value A numeric value
#'
#' @returns The smallest value in 'vec' that is larger than 'value'
#' @keywords internal
smallest_larger_value <- function(vec, value) {
  min(vec[vec > value])
}

#' Get index of first and last line of section
#'
#' @param lines A character vector
#' @param head_line A numeric value, line number of section heading
#' @keywords internal
section_interval <- function(lines, head_line) {
  breaks <-
    c(
      stringr::str_which(
        lines,
        classification_file_section_heading_regexp()
      ),
      length(lines) + 1
    )

  (head_line + 1):(smallest_larger_value(breaks, head_line) - 1)
}

#' Return lines in section
#'
#' @inheritParams section_interval
#' @keywords internal
get_section <- function(lines, head_line) {
  lines[section_interval(lines, head_line)]
}

#' Get section in .vs or .agg file
#'
#' @param lines A character vector
#' @param heading Character, the section heading
#' @param key Optional. Character, subkey in section
#' @keywords internal
extract_section <- function(lines, heading, key = NULL) {
  head_lines <- stringr::str_which(lines, stringr::fixed(heading))

  if (identical(head_lines, integer(0))) {
    return(NULL)
  }

  colname <- stringr::str_remove_all(heading, "\\[|\\]") |> tolower()

  section <-
    head_lines |>
    purrr::map(~ get_section(lines, .x)) |>
    unlist() |>
    dplyr::as_tibble() |>
    dplyr::mutate(across("value", stringr::str_trim)) |>
    dplyr::filter(.data$value != "") |>
    tidyr::drop_na("value") |>
    tidyr::separate_wider_delim(
      cols = "value", delim = "=", names = c("id", colname)
    )


  if (!is.null(key)) {
    section <- section |> dplyr::filter(.data$id == key)
  }

  section
}

#' Create new classification object
#'
#' Constructor for internal functions
#'
#' @inheritParams px_classification
#'
#' @returns A classification object
#' @keywords internal
new_classification <- function(name, prestext, domain, df) {
  if (length(name) == 0 || length(prestext) == 0 || length(domain) == 0) {
    stop("name, prestext, and domain must be non-empty.")
  }

  vs <- list(
    name = as.character(name),
    prestext = as.character(prestext),
    domain = as.character(domain), # list
    # type     = "V", only support V
    df = df
  )

  structure(vs, class = "classification")
}

#' Get data set from aggregation file
#'
#' @param path to an aggregation file
#'
#' @returns A data frame with columns 'valuecode' (character) and a
#' second column (ordered) named after the aggregation
#' @keywords internal
aggregation_df <- function(path) {
  agg_lines <- read_lines_guess_encoding(path) # nolint: object_usage_linter

  error_if_aggregation_file_is_missing_mandatory_headings(agg_lines)

  aggregation_name <-
    basename(path) |>
    stringr::str_remove("\\.agg$")

  aggregation_groups_df <-
    extract_section(agg_lines, "[Aggreg]") |>
    dplyr::filter(stringr::str_detect(.data$id, "^\\d+$"))

  aggregation_text_df <- extract_section(agg_lines, "[Aggtext]")

  if (nrow(aggregation_groups_df) != nrow(aggregation_text_df)) {
    warning(paste0(
      "The number of aggregation groups (",
      nrow(aggregation_groups_df), ") and Aggtexts (",
      nrow(aggregation_text_df),
      ") differ in '", basename(path), "'."
    ))
  }

  aggregation_df <-
    dplyr::left_join(aggregation_groups_df, aggregation_text_df, by = "id") |>
    dplyr::select(-"id") |>
    dplyr::mutate(across(everything(), ~ dplyr::na_if(.x, "")))

  df <- dplyr::tibble(
    valuecode = as.character(),
    !!aggregation_name := factor(
      levels = aggregation_df$aggtext,
      ordered = TRUE
    )
  )

  for (aggregation_group in aggregation_df$aggreg) {
    section <- extract_section(agg_lines, paste0("[", aggregation_group, "]"))

    aggregation_text <-
      aggregation_df$aggtext[aggregation_df$aggreg == aggregation_group]

    if (is.null(section)) {
      warning(
        "No group with label '[", aggregation_group, "]' found in '",
        basename(path), "'."
      )
    } else {
      aggregation_values <-
        section |>
        dplyr::select(-"id") |>
        dplyr::pull(1)

      df <-
        dplyr::bind_rows(
          df,
          dplyr::tibble(
            valuecode = aggregation_values,
            !!aggregation_name := factor(aggregation_text,
              levels = aggregation_df$aggtext,
              ordered = TRUE
            )
          )
        )
    }
  }

  df
}

#' px classification from path
#'
#' Create px classification from .vs and .agg files
#'
#' @inheritParams px_classification
#'
#' @returns A classification object
#' @keywords internal
px_classification_from_path <- function(vs_path, agg_paths) {
  vs_lines <- read_lines_guess_encoding(vs_path)

  error_if_vs_file_is_missing_mandatory_headings(vs_lines)

  valuecode_df <- extract_section(vs_lines, "[Valuecode]")
  valuetext_df <- extract_section(vs_lines, "[Valuetext]")

  if (is.null(valuetext_df)) {
    vs_df <- dplyr::select(valuecode_df, -"id")
  } else {
    if (nrow(valuecode_df) != nrow(valuetext_df)) {
      warning(paste0(
        "[Valuecode] and [Valuetext] in '", basename(vs_path),
        "' have different number of rows."
      ))
    }

    vs_df <-
      dplyr::left_join(valuecode_df, valuetext_df, by = "id") |>
      dplyr::select(-"id") |>
      dplyr::mutate(across(everything(), ~ dplyr::na_if(.x, "")))
  }

  if (missing(agg_paths)) {
    vs_dir <- dirname(vs_path)

    agg_paths <- file.path(vs_dir, extract_section(vs_lines, "[Aggreg]")$aggreg)
  }

  if (any(!file.exists(agg_paths))) {
    warning(paste0(
      "Aggregation files: ",
      paste(basename(agg_paths[!file.exists(agg_paths)]),
        collapse = ", "
      ),
      " do not exist."
    ))

    agg_paths <- agg_paths[file.exists(agg_paths)]
  }

  if (length(agg_paths) == 0) {
    df <- vs_df
  } else {
    agg_df <-
      agg_paths |>
      purrr::map(aggregation_df) |>
      purrr::compact() |>
      purrr::reduce(dplyr::full_join, by = "valuecode")

    df <- dplyr::full_join(vs_df, agg_df, by = "valuecode")
  }

  new_classification(
    name = extract_section(vs_lines, "[Descr]", "Name") |> dplyr::pull(2),
    prestext =
      extract_section(vs_lines, "[Descr]", "Prestext") |> dplyr::pull(2),
    domain = extract_section(vs_lines, "[Domain]") |> dplyr::pull(2),
    df = df
  )
}

#' px classification from data frame
#'
#' Create px classification from a data frame.
#'
#' @inheritParams px_classification
#'
#' @returns A classification object
#' @keywords internal
px_classification_from_df <- function(name, prestext, domain, df) {
  character_columns <- intersect(names(df), c("valuecode", "valuetext"))

  df_formatted <-
    df |>
    dplyr::mutate(
      across(all_of(character_columns), as.character),
      across(-all_of(character_columns), ~ factor(.x, ordered = TRUE))
    )

  new_classification(
    name = name,
    prestext = prestext,
    domain = domain,
    df = df_formatted
  )
}


#' Create a classification object
#'
#' Create a classification object from a data frame or .vs and .agg files.
#'
#' A classification is a combination of a value set and zero, one, or more
#' aggregations. The classification can be saved as .vs and .agg files
#' (see [px_save_classification()]).
#'
#' If a classification is created from a data frame, the arguments `name` and
#' `prestext` and `domain` are required. If a classification is created from .vs
#' and .agg files, all other arguments should be empty.
#'
#' For aggregations, it's normally possible to have codes and values in the .agg
#' file under the sections '\[Aggreg\]' and '\[Aggtext\]' respectively. However,
#' in pxmake's implementation of classifications, it's not possible to
#' distinguish between these. When a new classification is created, the values
#' in the section '\[Aggtext\]' are used as aggregation values.
#'
#' Only value sets of type 'V' are supported. Type values sets with type 'H' and
#' 'N' are not supported.
#'
#' @param name Optional. Name of the classification.
#' @param prestext Optional. Presentation text.
#' @param domain Optional. Character vector with domain names. Used to link to
#' PX-file.
#' @param df Optional. A data frame with required column 'valuecode' and
#' optional column 'valuetext', if the codes have texts. Each additional column
#' represents an aggregation. The column name is the name of the aggregation.
#' If the column type is character the aggregation levels will be sorted
#' alphabetically; use factors to control the ordering.
#' @param vs_path Optional. Path to a values set (.vs) file.
#' @param agg_paths Optional.
#' \itemize{
#'   \item If NULL, aggregation paths are automatically taken from the
#'   \[Aggreg\] field in the .vs file.
#'   \item Use a vector of paths to one or more aggregation files (.agg) to
#'   manually choose aggregations.
#'   \item Use character(0) if aggregations from the .vs files should not be
#'   added automatically.
#' }
#' @returns A classification object
#'
#' @examples
#' # Create classification from data frame
#' library(tibble)
#'
#' c1 <- px_classification(
#'   name = "Age5",
#'   prestext = "Ages 0-9 - 60+",
#'   domain = "age",
#'   df = tribble(
#'     ~valuecode, ~valuetext, ~`25 years classes`,
#'     "0-4", "0-4 years", "0-24",
#'     "5-9", "5-9 years", "0-24",
#'     "10-14", "10-14 years", "0-24",
#'     "15-19", "15-19 years", "0-24",
#'     "20-24", "20-24 years", "0-24",
#'     "25-29", "25-29 years", "25-49",
#'     "30-34", "30-34 years", "25-49",
#'     "35-39", "35-39 years", "25-49",
#'     "40-44", "40-44 years", "25-49",
#'     "45-49", "45-49 years", "25-49",
#'     "50-54", "50-54 years", "50-74",
#'     "55-59", "55-59 years", "50-74",
#'     "60-64", "60-64 years", "50-74",
#'     "65-69", "65-69 years", "50-74",
#'     "70-74", "70-74 years", "50-74",
#'     "75+", "75+ years", "75+"
#'   )
#' )
#'
#' # Create classifications from files
#'
#' vs_file <- system.file("extdata", "Age5.vs", package = "pxmake")
#'
#' agg_files <- c(
#'   system.file("extdata", "10-years_classes.agg", package = "pxmake"),
#'   system.file("extdata", "25-years_classes.agg", package = "pxmake")
#' )
#'
#' if (vs_file != "" & all(agg_files != "")) {
#'   # Create classification from .vs file and use aggregations mentioned in .vs
#'   c2 <- px_classification(vs_path = vs_file)
#'
#'   # Create classification from .vs file and manually specify aggregation
#'   # files
#'   c3 <- px_classification(
#'     vs_path = vs_file,
#'     agg_paths = agg_files
#'   )
#'
#'   identical(c2, c3)
#' }
#' @export
px_classification <- function(name, prestext, domain, df, vs_path, agg_paths) {
  validate_px_classification_arguments(
    name, prestext, domain, df, vs_path, agg_paths
  )

  if (all(missing(vs_path), missing(agg_paths))) {
    c <- px_classification_from_df(name, prestext, domain, df)
  } else if (missing(agg_paths)) {
    c <- px_classification_from_path(vs_path)
  } else {
    c <- px_classification_from_path(vs_path, agg_paths)
  }

  c
}

#' Add lines numbers
#'
#' Add '1=', '2=', etc. at start of each line.
#'
#' @param lines A character vector
#'
#' @returns A character vector
#' @keywords internal
enumerate_lines <- function(lines) {
  stringr::str_glue("{seq_along(lines)}={lines}") |>
    paste(collapse = "\n")
}

blank_line <- function() " "

#' Save classification as .vs file
#'
#' Write value set part of classification to a file.
#'
#' @param c A classification object
#' @param directory Directory to save the file in
#'
#' @returns Nothing
#' @keywords internal
write_value_set <- function(c, directory) {
  filename <- file.path(directory, paste0(c$name, ".vs"))

  aggregation_df <-
    c$df |>
    dplyr::select(-"valuecode", -"valuetext")

  if (ncol(aggregation_df) == 0) {
    aggregation_text <- ""
  } else {
    aggregation_names <-
      aggregation_df |>
      names() |>
      paste0(".agg") |>
      enumerate_lines()

    aggregation_text <-
      stringr::str_glue("[Aggreg]",
        aggregation_names,
        "{blank_line()}\n",
        .sep = "\n"
      )
  }

  get_values <- function(var) {
    c$df |>
      tidyr::replace_na(setNames(list(""), var)) |>
      dplyr::pull(var) |>
      enumerate_lines()
  }

  value_codes <- get_values("valuecode")
  value_text <- get_values("valuetext")

  vs_lines <-
    stringr::str_glue(
      "[Descr]",
      "Name={c$name}",
      "Prestext={c$prestext}",
      "Type=V",
      blank_line(),
      aggregation_text,
      "[Domain]",
      enumerate_lines(c$domain),
      blank_line(),
      "[Valuecode]",
      value_codes,
      blank_line(),
      "[Valuetext]",
      value_text,
      .sep = "\n"
    )

  file_connection <- file(filename, encoding = "ISO-8859-1")
  writeLines(vs_lines, file_connection)
  close(file_connection)
}

#' Save classification as .agg file
#'
#' Write one aggregation column in a classification to an .agg file.
#'
#' @param aggregation Character, the column name of the aggregation in the
#' classification
#' @param c A classification object
#' @param directory Directory to save the file in
#'
#' @returns Nothing
#' @keywords internal
write_aggregation <- function(aggregation, c, directory) {
  filename <- file.path(directory, paste0(aggregation, ".agg"))

  groups <- levels(na.omit(c$df[[aggregation]]))

  agg_texts <-
    c$df |>
    dplyr::distinct(.data$valuecode, !!rlang::sym(aggregation)) |>
    tidyr::drop_na() |>
    dplyr::arrange(as.character(!!rlang::sym(aggregation))) |>
    tidyr::pivot_wider(
      names_from = all_of(aggregation),
      values_from = "valuecode",
      values_fn = list
    ) |>
    tidyr::pivot_longer(everything()) |>
    dplyr::rowwise() |>
    dplyr::mutate(
      str = stringr::str_glue(
        "[{name}]\n",
        "{paste(enumerate_lines(value), collapse = '\n')}"
      )
    ) |>
    dplyr::pull("str") |>
    paste0(collapse = paste0("\n", blank_line(), "\n"))

  agg_lines <-
    stringr::str_glue(
      "[Aggreg]",
      "Name={aggregation}",
      "Valueset={c$name}",
      enumerate_lines(groups),
      blank_line(),
      "[Aggtext]",
      enumerate_lines(groups),
      blank_line(),
      agg_texts,
      .sep = "\n"
    )

  file_connection <- file(filename, encoding = "ISO-8859-1")
  writeLines(agg_lines, file_connection)
  close(file_connection)
}

#' Save classification as .vs and .agg files
#'
#' Save a classification object as .vs and .agg files. The .vs file contains
#' the value set and the .agg files contain the aggregations.
#'
#' @param c A classification object
#' @param path Directory to save the files in
#'
#' @returns Nothing.
#'
#' @examples
#' # Save classification as .vs as .agg files
#' c <- px_classification(
#'   name = "Age5",
#'   prestext = "Ages 0-9 - 60+",
#'   domain = "age",
#'   df = age_classification
#' )
#'
#' px_save_classification(c, path = tempdir())
#' @export
px_save_classification <- function(c, path) {
  write_value_set(c, path)

  aggregations <-
    c$df |>
    dplyr::select(-"valuecode", -"valuetext") |>
    names()

  purrr::walk(aggregations, write_aggregation, c = c, directory = path)
}

test_that("A classification can be created from df and paths", {
  c1 <- px_classification(
    name = "Test",
    prestext = "Test",
    domain = "Test",
    df = data.frame(valuecode = 1:26, valuetext = letters)
  )

  expect_s3_class(c1, "classification")

  c2 <- px_classification(
    vs_path = vs_age5_path(),
    agg_paths = agg_25years_path()
  )

  expect_s3_class(c2, "classification")
})

test_that("Classification can be created without aggregation", {
  c1 <- px_classification(
    vs_path = vs_age5_path(),
    agg_paths = character(0)
  )

  expect_s3_class(c1, "classification")

  c2 <- px_classification(
    name = "No aggregation",
    prestext = "No aggregation",
    domain = "no_agg",
    df = dplyr::tribble(
      ~valuecode, ~valuetext,
      "0-4", "0-4 years",
      "5-9", "5-9 years",
      "10-14", "10-14 years"
    )
  )

  expect_s3_class(c2, "classification")
})

test_that("Classification automatically read agg files", {
  c <- px_classification(vs_path = vs_age5_path())

  expect_equal(c$df, age_classification)
})

test_that("Classification is robust to formatting differences in files", {
  target <- px_classification(vs_path = vs_age5_path())
  c <- px_classification(vs_path = vs_age5_strangely_formatted_path())

  expect_identical(target$name, c$name)
  expect_identical(target$prestext, c$prestext)
  expect_identical(target$domain, c$domain)
  expect_identical(target$df$valuecode, c$df$valuecode)
  expect_identical(target$df$valuetext, c$df$valuetext)
  expect_identical(target$df[[3]], c$df[[3]])
  expect_identical(target$df[[4]], c$df[[4]])
})

test_that("Classification can be create without [Valuetext]", {
  expect_s3_class(
    px_classification(
      vs_path = get_classification_path("vs_without_valuetext.vs")
    ),
    "classification"
  )
})

test_that("Aggregation values are taken from Aggtext section", {
  c <- px_classification(vs_path = vs_pxvsbrche_path())

  expect_equal(
    c$df$BRCH_BRCH_E |> as.character(),
    c(
      "All industries",
      "Agriculture, forestry and related industries",
      "Fishing and other related industries",
      "Mining and quarrying",
      "Manufacturing",
      "Energy and watersupply",
      "Construction",
      "Wholesale and retail trade",
      "Transportation and storage",
      "Accommodation and food service activities",
      "Information and communication",
      "Financial and insurance activities",
      "Real estate activities",
      "Professional, scientific and technical activities",
      "Administrative and support service activities",
      "Public administration and service",
      rep(NA, 17),
      "Other service industries",
      "Unknown"
    )
  )

  expect_equal(
    c$df$BRCH_OFF_E |> as.character(),
    c(
      rep(NA, 15),
      "Public administration and service",
      "General public services (COFOG 1)",
      "Defence (COFOG 2)",
      "Public order and safety (COFOG 3)",
      "Economic affairs (COFOG 4)",
      "Environmental protection (COFOG 5)",
      "Housing and community amenities (COFOG 6)",
      "Health (COFOG 7)",
      "Recreation, culture and religion (COFOG 8)",
      "Education (COFOG 9)",
      "Education other (COFOG 9.x)",
      "Primary education (COFOG 9.10)",
      "Youth-level education (COFOG 9.20)",
      "Social protection (COFOG 10)",
      "Social protection (COFOG 10.x)",
      "Old age (COFOG 10.20)",
      "Family and children (COFOG 10.40)",
      "Other (COFOG x)",
      rep(NA, 2)
    )
  )
})

test_that("Error if vs_file is missing heading", {
  expect_error(
    px_classification_from_path(
      vs_path =
        get_classification_path("vs_missing_mandatory_section.vs")
    ),
    regexp = "missing mandatory section"
  )
})

test_that("Warning if length of [Valuecode] and [Valuetext] differ", {
  expect_warning(
    px_classification_from_path(
      vs_path =
        get_classification_path("vs_different_lengths1.vs")
    ),
    regexp = "different number of rows"
  )

  expect_warning(
    px_classification_from_path(
      vs_path =
        get_classification_path("vs_different_lengths2.vs")
    ),
    regexp = "different number of rows"
  )
})

# ERRORS
test_that("Only paths OR name/domain/df/ can be defined", {
  expect_error(px_classification(name = "name", vs_path = "value_set.vs"),
    regexp = "path arguments"
  )

  expect_error(
    px_classification(
      prestext = "name",
      agg_paths = "aggregation.agg"
    ),
    regexp = "path arguments"
  )
})

test_that(".vs files exists", {
  expect_error(
    px_classification(
      vs_path = "not_a_valueset.vs",
      agg_paths = agg_25years_path()
    ),
    regexp = "does not exist"
  )
})

test_that(".agg files exists", {
  expect_warning(
    px_classification(
      vs_path = vs_age5_path(),
      agg_paths = "not_an_aggregation.agg"
    ),
    regexp = "do not exist"
  )
})

test_that("if agg_paths is defined vs_path should also", {
  expect_error(px_classification(agg_paths = agg_25years_path()),
    regexp = "'vs_path' is missing"
  )
})

test_that("if any non-path argument is defined, all must be defined", {
  expect_error(
    px_classification(
      name = "name",
      prestext = "prestext",
      domain = "domain"
    ),
    regexp = '"df" is missing'
  )

  expect_error(
    px_classification(
      name = "name",
      prestext = "prestext",
      df = age_classification
    ),
    regexp = '"domain" is missing'
  )

  expect_error(
    px_classification(
      name = "name",
      domain = "domain",
      df = age_classification
    ),
    regexp = '"prestext" is missing'
  )

  expect_error(
    px_classification(
      prestext = "prestext",
      domain = "domain",
      df = age_classification
    ),
    regexp = '"name" is missing'
  )
})

test_that("df has the right columns", {
  expect_error(px_classification(women,
    name = "name",
    prestext = "prestext",
    domain = "domain",
    regexp = "missing column 'valuecode'"
  ))
})

test_that("Error if wrong input formats", {
  expect_error(
    px_classification(
      name = c("name", "name2"),
      prestext = "prestext",
      domain = "domain",
      df = data.frame(
        valuecode = c(1, 2, 3),
        valuetext = c("a", "b", "c")
      )
    ),
    regexp = "'name' must be a character string"
  )

  expect_error(
    px_classification(
      name = "name",
      prestext = 5,
      domain = "domain",
      df = data.frame(
        valuecode = c(1, 2, 3),
        valuetext = c("a", "b", "c")
      )
    ),
    regexp = "'prestext' must be a character string"
  )

  expect_error(
    px_classification(
      name = "name",
      prestext = "prestext",
      domain = data.frame(),
      df = data.frame(
        valuecode = c(1, 2, 3),
        valuetext = c("a", "b", "c")
      )
    ),
    regexp = "'domain' must be a character string"
  )

  expect_error(
    px_classification(
      name = "name",
      prestext = "prestext",
      domain = "domani",
      df = c("valuetext")
    ),
    regexp = "'df' must be a data frame"
  )

  expect_error(
    px_classification(
      name = "name",
      prestext = "prestext",
      domain = "domain",
      df = data.frame(
        no_valuecode = c(1, 2, 3),
        valuetext = c("a", "b", "c")
      )
    ),
    regexp = "missing column 'valuecode'"
  )
})

# ERROR
# If trying to read a non-option (Aggtext, Valuetext) heading?
# E.g. when an agg file has a heading that doesn't match a group

# WARNINGS
# valuetext and valuecode have differing lengths
# A chunk contains invalid formatted lines.
# All should be digit|Name|Pretext=value

test_that("Aggregation df is created", {
  expect_identical(
    aggregation_df(agg_10years_path()),
    dplyr::select(age_classification, valuecode, `10-years_classes`)
  )

  expect_identical(
    aggregation_df(agg_25years_path()),
    dplyr::select(age_classification, valuecode, `25-years_classes`)
  )
})


test_that("Warning, aggreation groups and 'Aggtext' differ in length", {
  expect_warning(
    aggregation_df(get_classification_path("agg_different_lengths1.agg")),
    regexp = "number of aggregation groups.*differ"
  )

  expect_warning(
    aggregation_df(get_classification_path("agg_different_lengths2.agg")),
    regexp = "number of aggregation groups.*differ"
  )
})

test_that("Warning, if aggregation group is missing", {
  expect_warning(
    aggregation_df(get_classification_path("agg_missing_group.agg")),
    regexp = "No group with label"
  )
})

test_that("Error, if missing mandatory section", {
  expect_error(
    aggregation_df(
      get_classification_path("agg_missing_mandatory_section.agg")
    ),
    regexp = "missing mandatory section"
  )
})

# Check classification on many 'real-life' .vs and .agg files

test_that(" ", {
  vs_path_inverse <- function(vs_path) {
    px_classification(vs_path = vs_path) |>
      expect_save_read_preserves_classification()
  }

  vs_path_inverse(get_classification_path("BEXSTALLlcode.vs"))
  vs_path_inverse(get_classification_path("BEXSTALLlkode.vs"))
  vs_path_inverse(get_classification_path("OFAKD.vs"))
  vs_path_inverse(get_classification_path("OFAKN.vs"))
  vs_path_inverse(get_classification_path("PXVSAarmd.vs"))
  vs_path_inverse(get_classification_path("PXVSAge.vs"))
})

get_data_path <- function(table_name) {
  tables_with_data_in_excel <-
    c("FOTEST", "no_timeval_or_codes", "zero_heading", "zero_stub")

  tabels_with_data_in_parquet <- "BEXSTA_parquet"

  if (tolower(table_name) %in% tolower(tables_with_data_in_excel)) {
    NULL
  } else if (tolower(table_name) %in% tolower(tabels_with_data_in_parquet)) {
    testthat::test_path("fixtures", "data", paste0(table_name, ".parquet"))
  } else {
    testthat::test_path("fixtures", "data", paste0(table_name, ".rds"))
  }
}

get_classification_path <- function(name) {
  testthat::test_path("fixtures", "classification", name)
}

classification_path <- function(name) {
  function() {
    get_classification_path(name = name)
  }
}

agg_10years_path <- classification_path("10-years_classes.agg")
agg_25years_path <- classification_path("25-years_classes.agg")
vs_age5_path <- classification_path("Age5.vs")
vs_age5_strangely_formatted_path <- classification_path(
  "Age5_strangely_formatted.vs"
)
vs_agg_dont_exists <- classification_path("non_existing_agg.vs")
agg_different_lengths <- classification_path("agg_different_lengths.agg")
vs_different_lengths <- classification_path("agg_different_lengths.vs")
vs_pxvsbrche_path <- classification_path("PXVSBRCHE.vs")

get_metadata_path <- function(table_name) {
  testthat::test_path(
    "fixtures", "metadata", stringr::str_glue("metadata_{table_name}.xlsx")
  )
}

get_px_file_path <- function(table_name) {
  testthat::test_path("fixtures", "px", paste0(table_name, ".px"))
}

get_pxjob_file_path <- function(table_name) {
  testthat::test_path("fixtures", paste0(table_name, "_pxjob.px"))
}

expect_equal_lines <- function(path1, path2) {
  lines1 <- readLines(path1)
  lines2 <- readLines(path2)

  testthat::expect_equal(lines1, lines2)
}

px_from_table_name <- function(table_name) {
  px(
    input = get_metadata_path(table_name),
    data = get_data_path(table_name)
  )
}

expect_px_px_save_preserves_everything <- function(x) {
  px1 <- temp_px_file()
  px2 <- temp_px_file()

  px_save(x, path = px1)

  px1 |>
    px() |>
    px_save(path = px2)

  expect_equal_lines(px1, px2)
}

#' Run pxjob and delete created files when environment is killed
pxjob_clean <- function(input, output, env = parent.frame()) {
  pxjob64Win::pxjob(input, output)

  withr::defer(envir = env, {
    file.remove(output)
  })
}

#' Run px() and px_save() for a specific table. Return path to file.
create_px_file <- function(table_name) {
  px_path <- temp_px_file()

  px(
    input = get_metadata_path(table_name),
    data = get_data_path(table_name)
  ) |>
    px_save(path = px_path)

  px_path
}

expect_save_read_preserves_classification <- function(c) {
  tempdir <- temp_dir()
  px_save_classification(c, tempdir)

  c2 <-
    px_classification(
      vs_path = list.files(tempdir, pattern = ".*\\.vs", full.names = TRUE)
    )

  testthat::expect_identical(c, c2)
}

test_that("px_classification_save saves .vs and .agg files", {
  tmp_dir <- tempdir()

  px_classification(
    name = "Age",
    prestext = "Age table",
    domain = "age",
    df = age_classification
  ) |>
    px_save_classification(path = tmp_dir)

  expect_equal(
    list.files(tmp_dir, pattern = "(agg|vs)$"),
    c("10-years_classes.agg", "25-years_classes.agg", "Age.vs")
  )
})

test_that("NA values are not saved in .agg", {
  c <- px_classification(vs_path = vs_pxvsbrche_path())

  tmp_dir <- tempdir()

  px_save_classification(c, path = tmp_dir)

  agg_files <- list.files(tmp_dir, pattern = "\\.agg$", full.names = TRUE)

  for (agg_file in agg_files) {
    expect_false(any(readLines(agg_file) == "[NA]"))
  }
})

# Check that reading -> saving -> reading creates the same classification object

test_that("classification is preserved", {
  px_classification(
    name = "test",
    prestext = "test",
    domain = "test",
    df = tibble::tibble(
      valuecode = as.character(1:10),
      valuetext = letters[1:10]
    )
  ) |>
    expect_save_read_preserves_classification()

  px_classification(
    vs_path = vs_age5_path(),
    agg_paths = agg_25years_path()
  ) |>
    expect_save_read_preserves_classification()

  px_classification(vs_path = vs_age5_path()) |>
    expect_save_read_preserves_classification()

  px_classification(vs_path = vs_pxvsbrche_path()) |>
    expect_save_read_preserves_classification()
})

#' Add or modify value
#'
#' Modify a value in a row of a data frame based on its value in another column.
#' If the value is not found, a new row is added.
#'
#' @param df Data frame
#' @param lookup_column Column to look up
#' @param lookup_column_values Values to look up
#' @param modify_column Column to modify
#' @param new_value New value to modify/add to modify_column
#'
#' @returns A data frame
#' @keywords internal
modify_or_add_row <- function(df,
                              lookup_column,
                              lookup_column_values,
                              modify_column,
                              new_value) {
  if (any(df[[lookup_column]] %in% lookup_column_values)) {
    df[df[[lookup_column]] %in% lookup_column_values, modify_column] <-
      new_value
  } else {
    df <- dplyr::bind_rows(
      df,
      dplyr::tibble(
        !!lookup_column := lookup_column_values,
        !!modify_column := new_value
      )
    )
  }

  df
}

modify_or_add_in_column <- function(df,
                                    lookup_column,
                                    lookup_column_values,
                                    new_value) {
  if (any(df[[lookup_column]] %in% lookup_column_values)) {
    df[df[[lookup_column]] %in% lookup_column_values, lookup_column] <-
      new_value
  } else {
    df <- dplyr::bind_rows(
      df,
      dplyr::tibble(!!lookup_column := new_value)
    )
  }

  df
}

modify_with_df <- function(df1, df2, modify_column) {
  valid_names <- # nolint: object_usage_linter.
    names(df1) |>
    paste(collapse = ", ")

  invalid_columns <-
    setdiff(names(df2), names(df1)) |>
    paste(collapse = ", ")

  if (invalid_columns != "") {
    error(stringr::str_glue(
      "Data frame contains invalid columns: ",
      "{invalid_columns}, it can only ",
      "contain the columns: {valid_names}"
    ))
  }

  merge_variables <- setdiff(names(df2), modify_column)

  df1_without_modify_column <- dplyr::select(df1, -all_of(modify_column))

  if (length(merge_variables) == 0) {
    replace_values <- dplyr::cross_join(df1_without_modify_column, df2)
    keep_values <- dplyr::filter(df1, FALSE)
    add_values <- dplyr::filter(df2, FALSE)
  } else {
    replace_values <- dplyr::inner_join(df1_without_modify_column, df2,
      by = merge_variables
    )
    keep_values <- dplyr::anti_join(df1, df2, by = merge_variables)
    add_values <- dplyr::anti_join(df2, df1, by = merge_variables)
  }

  dplyr::bind_rows(replace_values, keep_values, add_values)
}

remove_keyword_from_table <- function(table_name) {
  function(x, keyword) {
    x[[table_name]] <- x[[table_name]] |>
      dplyr::filter(keyword != !!keyword)

    x
  }
}

remove_keyword_table1 <- remove_keyword_from_table("table1")
remove_keyword_table2 <- remove_keyword_from_table("table2")

remove_keyword_variables2 <- function(x, keyword) {
  x$variables2[[tolower(keyword)]] <- NA
  x
}

modify_table1 <- function(x, keyword, value) {
  x$table1 <-
    modify_or_add_row(x$table1, "keyword", keyword, "value", value) |>
    sort_table1()

  x
}

modify_table2 <- function(x, keyword, value) {
  if (is.data.frame(value)) {
    value$keyword <- keyword
  } else if (is.character(value)) {
    value <-
      data.frame(
        keyword = keyword, value = value, language = defined_languages(x)
      )
  }

  x$table2 <-
    x$table2 |>
    modify_with_df(value, "value") |>
    sort_table2(languages = defined_languages(x))

  x
}

get_cells_name <- function(number) {
  paste0("cells", number)
}

modify_cells <- function(x, number, column, value) {
  if ("values" %in% names(value)) {
    value <- dplyr::rename(value, "value" = "values")
  }

  x[[get_cells_name(number)]] <-
    modify_with_df(x[[get_cells_name(number)]], value, column) |>
    align_data_frames(get(paste0("get_base_cells", number))())

  if (number == "1") {
    x$cells1 <- sort_cells1(x$cells1, data_table_names = names(px_data(x)))
  } else if (number == "2") {
    x$cells2 <- sort_cells2(x$cells2,
      data_table_names = names(px_data(x)),
      languages = px_languages(x)
    )
  } else {
    unexpected_error()
  }

  x
}

modify_variables1 <- function(x, column, value) {
  x$variables1 <-
    modify_with_df(x$variables1, value, column) |>
    align_data_frames(get_base_variables1()) |>
    sort_variables1()

  x
}

modify_variables2 <- function(x, column, value) {
  if (is.character(value)) {
    value <- dplyr::tibble("{column}" := value)
  }

  x$variables2 <-
    modify_with_df(x$variables2, value, column) |>
    align_data_frames(get_base_variables2()) |>
    sort_variables2(
      data_table_names = names(px_data(x)),
      languages = px_languages(x)
    )
  x
}

modify_acrosscells <- function(x, value, keyword, na_to_star) {
  missing_columns <- setdiff(c(px_stub(x), px_heading(x)), names(value))

  value_completed <-
    value |>
    (\(.) {
      if (length(missing_columns) > 0) {
        if (na_to_star) {
          dummy_df <- asterisk_tibble(columns = missing_columns)
        } else {
          dummy_df <- na_tibble(columns = missing_columns)
        }

        dplyr::bind_cols(., dummy_df)
      } else {
        .
      }
    })()

  x$acrosscells <-
    modify_with_df(x$acrosscells, value_completed, tolower(keyword)) |>
    align_data_frames(get_base_acrosscells(c(px_stub(x), px_heading(x))))

  x
}


get_table1_value <- function(x, keyword) {
  value <-
    x$table1 |>
    dplyr::filter(keyword == !!keyword) |>
    dplyr::pull(value)

  if (identical(value, character(0))) {
    NULL
  } else {
    value
  }
}

get_table2_value <- function(x, keyword) {
  value <-
    x$table2 |>
    dplyr::filter(.data$keyword == !!keyword) |>
    dplyr::select("code", "language", "value") |>
    (\(.) {
      if (length(unique(.$code)) == 1) {
        dplyr::select(., -"code")
      } else {
        .
      }
    })()

  if (nrow(value) == 0) {
    NULL
  } else if (all(length(defined_languages(x)) == 1, nrow(value) == 1)) {
    unique(value$value)
  } else {
    value
  }
}

get_cells_value <- function(x, number, column) {
  if (number == "1") {
    language_col <- NULL
  } else if (number == "2") {
    language_col <- "language"
  } else {
    unexpected_error()
  }

  x[[get_cells_name(number)]] |>
    dplyr::select(
      all_of(c("variable-code", "code", language_col, !!column))
    ) |>
    tidyr::drop_na(!!column) |>
    dplyr::select(where(~ !all(is.na(.))))
}

get_variable1_value <- function(x, column) {
  value <-
    x$variables1 |>
    dplyr::select("variable-code", !!column) |>
    tidyr::drop_na(!!column)

  if (nrow(value) == 0) {
    NULL
  } else {
    value
  }
}

get_variable1_logic_value <- function(x, column) {
  value <-
    x$variables1 |>
    dplyr::filter(.data[[column]] == TRUE) |>
    dplyr::pull("variable-code")

  if (length(value) == 0) {
    NULL
  } else {
    value
  }
}

get_variables2_value <- function(x, column) {
  value <-
    x$variables2 |>
    dplyr::select("variable-code", "language", all_of(column)) |>
    tidyr::drop_na(all_of(column))

  if (nrow(value) == 0) {
    NULL
  } else if (length(unique(x$variables2[[column]])) == 1) {
    unique(x$variables2[[column]])
  } else {
    if (length(defined_languages(x)) == 1) {
      dplyr::select(value, -"language")
    } else {
      value
    }
  }
}

get_acrosscells_value <- function(x, keyword) {
  column_name <- tolower(keyword)

  value <-
    x$acrosscells |>
    dplyr::select(all_of(c(
      intersect(names(x$acrosscells), names(x$data)),
      "language",
      column_name
    ))) |>
    tidyr::drop_na(all_of(column_name))

  if (nrow(value) == 0) {
    NULL
  } else if (length(defined_languages(x)) == 1) {
    dplyr::select(value, -"language")
  } else {
    value
  }
}

test_that("Value is added/modified", {
  df1 <- data.frame(
    a = c(1, 2, 1),
    b = c(4, 5, 6)
  )

  df2 <- data.frame(
    a = c(1, 2, 1),
    b = c(4, 7, 6)
  )

  expect_equal(
    modify_or_add_row(df1,
      lookup_column = "a", lookup_column_values = 2,
      modify_column = "b", new_value = 7
    ),
    df2
  )

  df3 <- data.frame(
    a = c(1, 2, 1),
    b = c(8, 5, 8)
  )

  expect_equal(
    modify_or_add_row(df1,
      lookup_column = "a", lookup_column_values = 1,
      modify_column = "b", new_value = 8
    ),
    df3
  )

  df4 <- data.frame(
    a = c(1, 2, 1),
    b = c(8, 8, 8)
  )

  expect_equal(
    modify_or_add_row(df1,
      lookup_column = "a", lookup_column_values = 1:2,
      modify_column = "b", new_value = 8
    ),
    df4
  )

  df5 <- data.frame(
    a = c(1, 2, 1),
    b = c(4, 5, 6)
  )

  df6 <- data.frame(
    a = c(7, 2, 7),
    b = c(4, 5, 6)
  )


  expect_equal(
    modify_or_add_in_column(df5,
      lookup_column = "a",
      lookup_column_values = 1,
      new_value = 7
    ),
    df6
  )

  df7 <- data.frame(
    a = c(1, 2, 1, 7),
    b = c(4, 5, 6, NA)
  )

  expect_equal(
    modify_or_add_in_column(df5,
      lookup_column = "a",
      lookup_column_values = 4,
      new_value = 7
    ),
    df7
  )
})

test_that("Data frame modifier works", {
  arrange_and_expect_equal <- function(df1, df2) {
    expect_identical(
      dplyr::arrange(df1, dplyr::across(everything())),
      dplyr::arrange(df2, dplyr::across(everything()))
    )
  }

  df1 <- data.frame(
    a = c(1, 1, 2),
    b = c(3, 4, 3),
    c = c(7, 8, 9)
  )

  df2 <- data.frame(
    a = c(1, 1),
    b = c(3, 4),
    c = c(1, 2)
  )

  df3 <- data.frame(
    a = c(1, 1, 2),
    b = c(3, 4, 3),
    c = c(1, 2, 9)
  )

  arrange_and_expect_equal(modify_with_df(df1, df2, "c"), df3)

  df4 <- data.frame(
    a = c(3),
    b = c(1),
    c = c(5)
  )

  df5 <- data.frame(
    a = c(1, 1, 2, 3),
    b = c(3, 4, 3, 1),
    c = c(7, 8, 9, 5)
  )

  arrange_and_expect_equal(modify_with_df(df1, df4, "c"), df5)

  df6 <- data.frame(
    a = c(1, 1, 3),
    b = c(3, 5, 3),
    c = c(6, 7, 9)
  )

  df7 <- data.frame(
    a = c(1, 1, 2, 3, 1),
    b = c(3, 4, 3, 3, 5),
    c = c(6, 8, 9, 9, 7)
  )

  arrange_and_expect_equal(modify_with_df(df1, df6, "c"), df7)

  df8 <- data.frame(
    a = c(1, 2, 3),
    c = c(5, 6, 7)
  )

  df9 <- data.frame(
    a = c(1, 1, 2, 3),
    b = c(3, 4, 3, NA),
    c = c(5, 5, 6, 7)
  )

  arrange_and_expect_equal(modify_with_df(df1, df8, "c"), df9)

  df10 <- data.frame(
    d = c(1, 1, 2, 3),
    e = c(3, 4, 3, 3),
    c = c(7, 8, 9, 9)
  )

  expect_error(modify_with_df(df1, df10, "c"), regexp = "invalid columns: d, e")

  df11 <- data.frame(
    a = c(1, 1, 2),
    b = c(3, 4, 3),
    c = NA
  )

  df12 <- data.frame(c = 1)

  df13 <- data.frame(
    a = c(1, 1, 2),
    b = c(3, 4, 3),
    c = c(1, 1, 1)
  )

  arrange_and_expect_equal(modify_with_df(df11, df12, "c"), df13)
})

handle_variables2_keyword <- function(x, value, keyword, validate) {
  colname <- tolower(keyword)

  if (missing(value)) {
    return(get_variables2_value(x, colname))
  } else if (is.null(value)) {
    return(remove_keyword_variables2(x, keyword))
  } else if (is.character(value)) {
    value <- dplyr::tibble(
      `variable-code` = c(px_stub(x), px_heading(x)),
      !!colname := value
    )
  }

  x <- modify_variables2(x, colname, value)

  return_px(x, validate)
}

#' @rdname px_domain.px
#' @export
px_domain <- function(x, value, validate) {
  UseMethod("px_domain")
}

#' @eval add_documentation_variables2(
#'  "DOMAIN",
#'  "aggregation1",
#'  "aggregation2",
#'  "aggregation3"
#' )
px_domain.px <- function(x, value, validate = TRUE) {
  handle_variables2_keyword(x, value, "DOMAIN", validate)
}

#' @rdname px_map.px
#' @export
px_map <- function(x, value, validate) {
  UseMethod("px_map")
}

#' @eval add_documentation_variables2(
#'  "MAP",
#'  "greenland",
#'  "cities",
#'  "municipalities"
#' )
px_map.px <- function(x, value, validate = TRUE) {
  handle_variables2_keyword(x, value, "MAP", validate)
}

#' @rdname px_variable_label.px
#' @export
px_variable_label <- function(x, value, validate) {
  UseMethod("px_variable_label")
}

#' @title Change VARIABLE-LABEL
#' @eval add_documentation_variables2("VARIABLE-LABEL", NA, NA, NA)
#' @description
#' The variable label is the name that is shown in the PX-file.
#' @examples
#' # Set VARIABLE-LABEL for individual variables
#' library(tibble)
#' x1 <-
#'   px(population_gl) |>
#'   px_variable_label(tribble(
#'     ~`variable-code`, ~`variable-label`,
#'     "gender", "Gender",
#'     "age", "Age"
#'   ))
#' px_variable_label(x1)
#'
#' # Set VARIABLE-LABEL for individual languages
#' x2 <-
#'   x1 %>%
#'   px_languages(c("en", "kl")) |>
#'   px_variable_label(tribble(
#'     ~`variable-code`, ~language, ~`variable-label`,
#'     "gender", "en", "Gender",
#'     "gender", "kl", "Suiaassuseq",
#'     "age", "en", "Age",
#'     "age", "kl", "Ukiut"
#'   ))
#' px_variable_label(x2)
#'
#' # Remove VARIABLE-LABEL
#' x3 <- px_variable_label(x2, NULL)
#' px_variable_label(x3)
px_variable_label.px <- function(x, value, validate = TRUE) {
  handle_variables2_keyword(x, value, "variable-label", validate)
}

test_that("variables 2 is modified", {
  x <-
    "BEXSTA" |>
    get_data_path() |>
    readRDS() |>
    px()

  expect_identical(px_domain(x), NULL)
  expect_identical(px_map(x), NULL)

  domain_str <- "aggregation list"

  x2 <- px_domain(x, domain_str)

  domain_df1 <- tibble::tribble(
    ~`variable-code`, ~domain,
    "place of birth", "aggregation list",
    "gender", "aggregation list",
    "time", "aggregation list"
  )

  expect_identical(px_domain(x2), domain_df1)

  x3 <- px_languages(x, c("da", "en"))
  expect_identical(px_domain(x3), NULL)

  x4 <- px_domain(x3, domain_str)
  expect_identical(
    px_domain(x4) |> dplyr::arrange(`variable-code`),
    tidyr::crossing(domain_df1, language = c("en", "da")) |>
      dplyr::relocate(domain, .after = last_col())
  )

  domain_df2 <- dplyr::tibble(
    `variable-code` = "time",
    language = c("da", "en"),
    domain = c(
      "aggregation list1",
      "aggregation list2"
    )
  )

  x5 <- px_domain(x3, domain_df2)
  expect_identical(px_domain(x5), domain_df2)

  domain_df2 <- dplyr::tibble(
    `variable-code` = "gender",
    language = c("en"),
    domain = c("aggregation list1")
  )

  x6 <- px_domain(x3, domain_df2)
  expect_identical(px_domain(x6), domain_df2)

  x7 <- px_domain(x5, NULL)
  expect_identical(px_domain(x7), NULL)


  variable_label_df <- dplyr::tribble(
    ~`variable-code`, ~`variable-label`,
    "place of birth", "birthplace",
    "gender", "sex",
    "time", "year",
    "persons", "count"
  )

  x8 <- px_variable_label(x, variable_label_df)

  expect_identical(px_variable_label(x8), variable_label_df)

  map_df <- dplyr::tribble(
    ~`variable-code`, ~`map`,
    "time", "not really a map, just an example"
  )

  x9 <- px_map(x, map_df)

  expect_identical(px_map(x9), map_df)
})

create_micro_file <- function(
  micro_var, x, filenames, keyword_values_long, out_dir
) {
  figures_var <- "n"

  new_data <-
    x$data |>
    dplyr::select(all_of(c(px_heading(x), micro_var))) |>
    dplyr::count(across(everything()), name = figures_var) |>
    dplyr::arrange_all() |>
    format_data_df(figures_variable = figures_var)

  headings_with_non_na_values <-
    new_data |>
    dplyr::filter(!!rlang::sym(micro_var) != "-") |>
    dplyr::select(all_of(px_heading(x))) |>
    dplyr::distinct_all() |>
    tidyr::complete(!!!rlang::syms(px_heading(x)))

  headings_with_only_na_values <-
    new_data |>
    dplyr::select(all_of(px_heading(x))) |>
    (\(.) {
      if (length(px_heading(x)) > 0) {
        dplyr::anti_join(., headings_with_non_na_values, by = px_heading(x))
      } else {
        dplyr::filter(headings_with_non_na_values, FALSE)
      }
    })()

  if (nrow(headings_with_non_na_values) == 0) {
    # Edge case: If all headings only contain NA values, all are kept
    headings_with_only_na_values <- headings_with_non_na_values
  }

  if (length(px_heading(x)) > 0) {
    # Remove headings where the figures variable only has NA values
    new_data <-
      new_data |>
      dplyr::anti_join(headings_with_only_na_values, by = px_heading(x))
  }

  data_names <- names(new_data)

  headings_with_only_na_values_long <-
    headings_with_only_na_values |>
    (\(.) {
      if (nrow(.) > 0) {
        tidyr::pivot_longer(.,
          cols = px_heading(x),
          names_to = "variable-code",
          values_to = "code"
        )
      } else {
        dplyr::tibble(`variable-code` = character(), code = character())
      }
    })()

  x_micro <-
    new_px(
      languages = x$languages,
      table1 = x$table1,
      table2 = x$table2,
      variables1 = dplyr::filter(
        x$variables1, .data$`variable-code` %in% data_names
      ),
      variables2 = dplyr::filter(
        x$variables2, .data$`variable-code` %in% data_names
      ),
      cells1 = dplyr::filter(x$cells1, .data$`variable-code` %in% data_names) |>
        dplyr::anti_join(headings_with_only_na_values_long,
          by = c("variable-code", "code")
        ),
      cells2 = dplyr::filter(x$cells2, .data$`variable-code` %in% data_names) |>
        dplyr::anti_join(headings_with_only_na_values_long,
          by = c("variable-code", "code")
        ),
      acrosscells = dplyr::select(
        x$acrosscells,
        all_of(c(
          setdiff(data_names, figures_var),
          names(get_base_acrosscells())
        ))
      ) |>
        tidyr::drop_na({{ micro_var }}),
      data = new_data
    ) |>
    fix_px() |>
    px_figures(figures_var, validate = FALSE)

  if (all(!is.null(keyword_values_long), nrow(keyword_values_long) > 0)) {
    extra_keywords <-
      keyword_values_long |>
      dplyr::filter(.data$variable %in% micro_var)

    keyword_functions <- unique(extra_keywords$keyword_function)

    for (fnc in keyword_functions) {
      language_dependent_keyword <-
        function_to_keyword(fnc) %in% language_dependant_keywords()

      value <-
        extra_keywords |>
        dplyr::filter(.data$keyword_function == fnc) |>
        (\(.) {
          if (language_dependent_keyword && "language" %in% names(.)) {
            dplyr::distinct(., .data$language, .data$value)
          } else {
            dplyr::distinct(., value) |>
              dplyr::pull(value)
          }
        })()

      modifying_function <- get(fnc)

      x_micro <- modifying_function(x = x_micro, value = value)
    }
  }

  if (any(is.null(filenames[micro_var]), is.na(filenames[micro_var]))) {
    filename <- paste0(micro_var, ".px")
  } else {
    filename <- filenames[micro_var]
  }

  px_save(x = x_micro, path = file.path(out_dir, filename))
}

#' Create micro PX-files
#'
#' Split one px object into many small PX-files (micro files), with count of
#' the variables in it.
#'
#' The HEADING variables are use in all the micro files, and a file is created
#' for each non-HEADING variable. The new PX-files are saved in a directory
#' specified by `out_dir`.
#'
#' The main loop uses the furrr package for parallelisation. Use future::plan()
#' to choose how to parallelise.
#'
#' @param x A px object.
#' @param out_dir Directory to save PX-files in.
#' @param keyword_values Optional. A data frame with column 'variable' and one
#' or more of: 'px_contents', 'px_title', 'px_description', and 'px_matrix'. The
#' columns will be added as keywords to the table for each non-HEADING variable
#' that match the 'variable' column. It probably work for other keywords as
#' well.
#'
#' Use the column 'filename' to control the filename of each micro file. The
#' filename path is relative to 'out_dir'.
#'
#' Use the column 'language' if the PX-file has multiple languages.
#'
#' @returns Nothing
#'
#' @examples
#' # Create px object with cohort as HEADING
#' x <-
#'   greenlanders |>
#'   px() |>
#'   px_stub(names(greenlanders)) |>
#'   px_heading("cohort")
#'
#' # Create micro files, one for each of the non-HEADING variables (gender, age,
#' # municipality)
#' px_micro(x)
#'
#' @export
px_micro <- function(x, out_dir = NULL, keyword_values = NULL) {
  validate_px_micro_arguments(x, out_dir)

  print_out_dir <- is.null(out_dir)

  if (is.null(out_dir)) out_dir <- temp_dir()

  micro_vars <- setdiff(names(x$data), px_heading(x))

  if (!is.null(keyword_values)) {
    if ("filename" %in% colnames(keyword_values)) {
      filenames <-
        keyword_values |>
        dplyr::select("variable", "filename") |>
        tibble::deframe()
    } else {
      filenames <- NULL
    }

    keyword_values_long <-
      keyword_values |>
      tidyr::pivot_longer(
        cols = setdiff(names(keyword_values), c("variable", "language")),
        names_to = "keyword_function"
      ) |>
      tidyr::drop_na("value") |>
      dplyr::filter(.data$keyword_function != "filename")
  } else {
    keyword_values_long <- NULL
    filenames <- NULL
  }

  furrr::future_walk(micro_vars,
    create_micro_file,
    x = px_stub(x, micro_vars, validate = FALSE),
    filenames = filenames,
    keyword_values_long = keyword_values_long,
    out_dir = out_dir
  )

  if (print_out_dir) print(paste("Created PX-files in:", out_dir))
}

#' Create a px object
#'
#' Create a px object from a PX-file, an Excel metadata workbook, or a data
#' frame.
#'
#' @param input Optional character string. Can be:
#' \itemize{
#'   \item Path to a PX-file
#'   \item Path to an Excel metadata workbook
#'   \item A data frame
#'   \item Path to an `.rds` or `.parquet` file with a data frame
#'   \item Direct download URL of a PX-file
#'   \item Direct download URL of a parquet file
#' }
#'
#' If input is a data frame or NULL, a px object with minimal
#' metadata is created.
#' @param data Either a data frame or a path to an `.rds` or `.parquet` file
#' with a data frame. This can only be used if `input` is an Excel metadata
#' workbook.
#' @eval param_validate()
#'
#' @returns A px object
#'
#' @examples
#' # Create px object from dataset
#' x1 <- px(population_gl)
#'
#' # Download PX-file for example
#' download_succeeded <- tryCatch(
#'   {
#'     px_path <- tempfile(fileext = ".px")
#'     url <- "https://bank.stat.gl:443/sq/0cf06962-19f1-4d5c-8d43-b7ed0009617d"
#'     download.file(url, px_path)
#'     TRUE
#'   },
#'   error = function(e) FALSE
#' )
#'
#' # Run examples only if file was downloaded
#' if (download_succeeded) {
#'   # Create px object from PX-file
#'   x2 <- px(px_path)
#'
#'   # Create px object from URL
#'   x3 <- px(url)
#' }
#'
#' # Create minimal px object
#' x4 <- px()
#'
#' @export
px <- function(input = NULL, data = NULL, validate = TRUE) {
  validate_px_arguments(input, data)

  if (is_url(input)) {
    input <- download_px_or_parquet_and_return_path(url = input)

    validate_px_arguments(input, data)
  }

  if (is_rds_file(input)) {
    input <- readRDS(input)
  }

  if (is_parquet_file(input)) {
    input <- arrow::read_parquet(input)
  }

  if (is.null(input)) {
    input <- data.frame()
  }

  if (is_rds_file(data)) {
    data <- readRDS(data)
  }

  if (is_parquet_file(data)) {
    data <- arrow::read_parquet(data)
  }

  if (is_px_file(input)) {
    px <- px_from_px_file(input)
  } else if (is_xlsx_file(input)) {
    px <- px_from_excel(input, data)
  } else if (is.data.frame(input)) {
    px <- px_from_data_df(input)
  } else {
    unexpected_error()
  }

  return_px(px, validate)
}

#' Save px object to file
#'
#' @param x A px object.
#' @param path Path to file. The file extension determines the format. Can be:
#' - `.px` to save as a PX-file
#' - `.xlsx` to save as an Excel workbook
#' - `.R` to save an R-script that creates the px object
#' @param save_data If FALSE, no 'Data' sheet is created in the Excel workbook.
#' Can only be used if `path` is an `.xlsx` file.
#' @param data_path Path to an `.rds` or `.parquet` file to save data table at.
#' Can only be used if `path` is an `.xlsx` or `.R` file, and `save_data` is
#' `TRUE`.
#' @details
#' Use `px_codepage()` to change file encoding.
#'
#' @seealso [px_codepage()]
#' @returns Nothing
#'
#' @examples
#'
#' # Save px object to PX-file
#' tmp_dir <- tempdir()
#'
#' x <- px(population_gl)
#'
#' px_save(x, file.path(tmp_dir, "population.px"))
#'
#' # Save px object to Excel workbook
#' px_save(x, file.path(tmp_dir, "population.xlsx"))
#'
#' # Save px object as R-script that creates the px object
#' px_save(x, file.path(tmp_dir, "population.R"))
#'
#' @export
px_save <- function(x, path, save_data = TRUE, data_path = NULL) {
  validate_px_save_arguments(x, path, save_data, data_path)

  if (is_px_file(path)) {
    save_px_as_px_file(x, path)
  } else if (is_xlsx_file(path)) {
    save_px_as_xlsx(x, path, save_data, data_path)
  } else if (is_r_file(path)) {
    save_px_as_r_script(x, path, data_path)
  } else {
    unexpected_error()
  }
}

#' Create new px object
#'
#' px constructor for internal functions
#'
#' @param languages A data frame with language metadata.
#' @param table1 A data frame with language independent table metadata.
#' @param table2 A data frame with language dependent table metadata.
#' @param variables1 A data frame with language independent variable metadata.
#' @param variables2 A data frame with language dependent variable metadata.
#' @param cells1 A data frame with language independent cells metadata.
#' @param cells2 A data frame with language dependent cells metadata.
#' @param acrosscells A data frame with metadata that spans multiple cells.
#' @param data A data frame with data.
#'
#' @returns A px object
#' @keywords internal
new_px <- function(languages, table1, table2, variables1, variables2,
                   cells1, cells2, acrosscells, data) {
  x <- list(
    languages = dplyr::as_tibble(languages),
    table1 = dplyr::as_tibble(table1),
    table2 = dplyr::as_tibble(table2),
    variables1 = dplyr::as_tibble(variables1),
    variables2 = dplyr::as_tibble(variables2),
    cells1 = dplyr::as_tibble(cells1),
    cells2 = dplyr::as_tibble(cells2),
    acrosscells = dplyr::as_tibble(acrosscells),
    data = dplyr::as_tibble(data)
  )

  structure(x, class = "px")
}

#' Fix some common issues in px objects
#'
#' @param x A px object.
#'
#' @returns A px object
#' @keywords internal
fix_px <- function(x) {
  undefined_variables <- setdiff(colnames(x$data), x$variables2$`variable-code`)

  if (length(undefined_variables) > 0) {
    x$variables2 <-
      dplyr::bind_rows(
        x$variables2,
        dplyr::tibble(`variable-code` = undefined_variables)
      )
  }

  # Add missing variable-labels to variables2
  x$variables2 <-
    x$variables2 |>
    dplyr::mutate(`variable-label` = ifelse(is.na(.data$`variable-label`),
      .data$`variable-code`,
      .data$`variable-label`
    ))

  x
}

#' Check px object
#'
#' Runs a number of checks on px object to see if it is valid.
#'
#' This check is run by default by all `px_*` functions, but can be skipped by
#' using `validate = FALSE`. This can be useful on large px objects where the
#' checks are time consuming. Instead of validating on every modifying function
#' `px_validate()` can be run as the final step to validate the object.
#'
#' @param x A supposed px object.
#'
#' @returns A valid px object.
#' @examples
#' # Turn off validation for modifying functions, and manually
#' # run validation as final step in creating px object.
#' x1 <-
#'   px(population_gl, validate = FALSE) |>
#'   px_title("Test", validate = FALSE) |>
#'   px_validate()
#'
#' @export
px_validate <- function(x) {
  error_if_not_list(x)
  error_if_not_class_px(x)
  error_if_not_list_of_data_frames(x)
  error_if_list_names_are_wrong(x)
  error_if_data_frame_is_missing_column(x)
  error_if_multiple_time_variables(x)
  error_if_timeval_isnot_heading_or_stub(x)
  error_if_contvariable_isnot_heading_or_stub(x)
  error_if_variable_label_is_na(x)
  error_if_misplaced_keywords_in_table(x, table_name = "table1")
  error_if_misplaced_keywords_in_table(x, table_name = "table2")
  error_if_variable_code_not_in_data(x, "variables1")
  error_if_variable_code_not_in_data(x, "variables2")
  error_if_variable_code_not_in_data(x, "cells1")
  error_if_variable_code_not_in_data(x, "cells2")
  error_if_data_column_is_not_defined(x, "variables1")
  error_if_data_column_is_not_defined(x, "variables2")
  error_if_used_languages_are_not_defined(x)
  error_if_language_not_in_languages(x)
  error_if_value_contains_quotation_marks(x)

  x
}

#' Return px object
#'
#' Wrapper potentially validate it and return px object.
#'
#' @inheritParams px
#'
#' @keywords internal
return_px <- function(x, validate) {
  if (validate) {
    px_validate(x)
  } else {
    x
  }
}

test_that("that PX-file doesn't change second time through px -> px_save", {
  expect_px_save_does_not_change_on_second_pass <- function(table_name) {
    px_in <- get_px_file_path(table_name)
    px_out1 <- temp_px_file()
    px_out2 <- temp_px_file()

    px_in |>
      px() |>
      px_save(path = px_out1)

    px_out1 |>
      px() |>
      px_save(path = px_out2)

    px_file_as_lines <- function(path) {
      lines <-
        path |>
        readLines()

      lines
    }

    expect_equal(px_file_as_lines(px_out1), px_file_as_lines(px_out2))
  }

  expect_px_save_does_not_change_on_second_pass("BEXSTA_windows_1252")
  expect_px_save_does_not_change_on_second_pass("CONTVARIABLE")
  expect_px_save_does_not_change_on_second_pass(
    "CONTVARIABLE_multiple_languages"
  )
  expect_px_save_does_not_change_on_second_pass("no_timeval_or_codes2")
  expect_px_save_does_not_change_on_second_pass("SOXATI4")
  expect_px_save_does_not_change_on_second_pass("TUX01")
})

test_that(
  "px(PX-file) and px(px_save(px(PX-file), path = excel)) are equivalent",
  {
    # Some files don't give an exact match in the PX-file because there can be
    # subtle differences in formatting. PxJob is run on these file to test
    # that they are equivalent.
    skip_if_not_installed("pxjob64Win", minimum_version = "1.1.0")

    run_px_px_save_excel_pxjob_and_compare <- function(table_name) {
      px1 <- get_px_file_path(table_name)
      excel1 <- temp_xlsx_file()
      px2 <- temp_px_file()
      pxjob1 <- temp_px_file()
      pxjob2 <- temp_px_file()

      px1 |>
        px() |>
        px_save(path = excel1)

      excel1 |>
        px() |>
        px_save(path = px2)

      pxjob_clean(input = px1, output = pxjob1)
      pxjob_clean(input = px2, output = pxjob2)

      #' Read pxjob file and remove lines that doesn't need to be equal
      read_pxjobfile <- function(path) {
        lines <-
          path |>
          readLines() |>
          # Improve when implementing #163
          stringr::str_subset("^VARIABLECODE.+", negate = TRUE) |>
          stringr::str_subset("^META-ID.+", negate = TRUE) |>
          (\(.) {
            if (table_name == "CONTVARIABLE_multiple_languages") {
              .
              stringr::str_subset(., '^VARIABLE-TYPE.+Time"', negate = TRUE)
            } else {
              .
            }
          })()

        if (table_name %in% c("no_timeval_or_codes2")) {
          lines <- stringr::str_subset(lines, "^CODES.+", negate = TRUE)
        }

        if (table_name %in% c("CONTVARIABLE")) {
          lines <- stringr::str_subset(
            lines, "^CODES\\(.year.\\).+",
            negate = TRUE
          )
        }

        lines
      }

      expect_equal(read_pxjobfile(pxjob1), read_pxjobfile(pxjob2))
    }

    run_px_px_save_excel_pxjob_and_compare("SOXATI4")
    run_px_px_save_excel_pxjob_and_compare("BEXSTA_windows_1252")
    run_px_px_save_excel_pxjob_and_compare("no_timeval_or_codes2")
    run_px_px_save_excel_pxjob_and_compare("CONTVARIABLE")
    run_px_px_save_excel_pxjob_and_compare("CONTVARIABLE_multiple_languages")
    run_px_px_save_excel_pxjob_and_compare("TUX01")
  }
)

test_that("px runs without errors (excel file)", {
  expect_runs_without_errors <- function(table_name) {
    px(
      input = get_metadata_path(table_name),
      data = get_data_path(table_name)
    )

    expect_true(TRUE)
  }

  expect_runs_without_errors("BEXLTALL")
  expect_runs_without_errors("BEXSTA")
  expect_runs_without_errors("FOTEST")
  expect_runs_without_errors("no_timeval_or_codes")
  expect_runs_without_errors("zero_heading")
  expect_runs_without_errors("zero_stub")
})

test_that("px runs without errors (PX-file)", {
  expect_runs_without_errors <- function(name) {
    px(input = get_px_file_path(name))

    expect_true(TRUE)
  }

  expect_runs_without_errors("BEXSTA_windows_1252")
  expect_runs_without_errors("SOXATI4")
  expect_runs_without_errors("TUX01")
  expect_runs_without_errors("CONTVARIABLE")
  expect_runs_without_errors("CONTVARIABLE_multiple_languages")
  expect_runs_without_errors("PRXPRISH")
  expect_runs_without_errors("multilingual_no_codes")
})

test_that("px runs without errors (data frame and rds path)", {
  expect_runs_without_errors <- function(name) {
    px(input = readRDS(get_data_path(name)))
    px(input = get_data_path(name))

    expect_true(TRUE)
  }

  expect_runs_without_errors("BEXLTALL")
  expect_runs_without_errors("BEXSTA_WITHOUT_TOTALS")
  expect_runs_without_errors("BEXSTA")
})

test_that("px runs without errors (data frame and parquet path)", {
  expect_runs_without_errors <- function(name) {
    px(input = arrow::read_parquet(get_data_path(name)))
    px(input = get_data_path(name))

    expect_true(TRUE)
  }

  expect_runs_without_errors("BEXSTA_parquet")
})

test_that("px can run on an Excel workbook without a 'Data' sheet", {
  px(input = get_metadata_path("BEXSTA"))

  expect_true(TRUE)
})

test_that("Minimal px object can be created without data", {
  x <- px()

  px_validate(x)

  expect_true(TRUE)
})


test_that("Validation can be turned off", {
  x0 <- px(women)

  x1 <- px_figures(x0, "fisk", validate = FALSE)

  expect_error(px_validate(x1), regexp = "variable-codes not in x")

  expect_true(TRUE)
})

test_that("ordered factors are sorted", {
  tmp_px <- temp_px_file()

  age_levels <- c("17-24", "25-64", "7-16", "65+", "0-6")

  df <-
    population_gl |>
    dplyr::mutate(age = factor(age,
      levels = age_levels,
      ordered = TRUE
    ))

  px_ordering <-
    px(df)$cells1 |>
    dplyr::filter(`variable-code` == "age") |>
    dplyr::arrange(order) |>
    dplyr::pull(code)


  expect_identical(age_levels, px_ordering)
})

test_that("numerics are sorted", {
  tmp_px <- temp_px_file()

  heights <- c(1, 2, 10, 20, 100, 200)

  population_gl |>
    dplyr::mutate(height = rep(heights, 5)) |>
    dplyr::relocate(n, .after = dplyr::last_col()) |>
    px() |>
    px_save(tmp_px)

  height_order_in_px <-
    px(tmp_px)$cells1 |>
    dplyr::filter(`variable-code` == "height") |>
    dplyr::pull(code)

  expect_identical(as.character(heights), height_order_in_px)
})

test_that("px created from data frame orders variables properly", {
  df <-
    dplyr::tibble(
      sick = c(TRUE, FALSE, FALSE, NA, TRUE),
      age = as.integer(c(1, 10, 100, NA, 42)),
      income = c(9999, 20.1, 100, 999, NA),
      name = c("E", "A", "B", "D", "C"),
      gender = as.factor(c("male", "female", "male", NA, "female")),
      education = factor(
        c(
          "elementary", NA, "high school",
          "middle school", "high school"
        ),
        levels = c("high school", "middle school", "elementary"),
        ordered = TRUE
      ),
      birth_date = as.Date(c(
        "1966-07-30", NA, "1973-06-11",
        "1980-11-06", "1994-03-02"
      )),
      n = c(1, 20, 42, 49, 2)
    )

  cells1_expect <- tibble::tribble(
    ~`variable-code`, ~code, ~order, ~precision,
    "sick", "FALSE", 1L, NA_real_,
    "sick", "TRUE", 2L, NA_real_,
    "age", "1", 1L, NA_real_,
    "age", "10", 2L, NA_real_,
    "age", "42", 3L, NA_real_,
    "age", "100", 4L, NA_real_,
    "income", "20.1", 1L, NA_real_,
    "income", "100", 2L, NA_real_,
    "income", "999", 3L, NA_real_,
    "income", "9999", 4L, NA_real_,
    "name", "A", 1L, NA_real_,
    "name", "B", 2L, NA_real_,
    "name", "C", 3L, NA_real_,
    "name", "D", 4L, NA_real_,
    "name", "E", 5L, NA_real_,
    "gender", "female", 1L, NA_real_,
    "gender", "male", 2L, NA_real_,
    "education", "high school", 1L, NA_real_,
    "education", "middle school", 2L, NA_real_,
    "education", "elementary", 3L, NA_real_,
    "birth_date", "1966-07-30", 1L, NA_real_,
    "birth_date", "1973-06-11", 2L, NA_real_,
    "birth_date", "1980-11-06", 3L, NA_real_,
    "birth_date", "1994-03-02", 4L, NA_real_
  ) |>
    dplyr::mutate(order = as.numeric(order))

  expect_identical(px(df)$cells1, cells1_expect)
})

test_that("px file is preserved when saved as R script", {
  save_as_r_and_open <- function(x, save_dataset = FALSE) {
    tmp_r <- temp_r_file()
    tmp_rds <- temp_rds_file()

    if (isTRUE(save_dataset)) {
      px_save(x, path = tmp_r, data_path = tmp_rds)
    } else {
      px_save(x, path = tmp_r)
    }

    x2 <- source(tmp_r)$value

    x2
  }

  expect_equal_all_px_element <- function(x, save_dataset = FALSE) {
    x2 <- save_as_r_and_open(x, save_dataset)

    expect_equal(x$languages, x2$languages)
    expect_equal(x$table1, x2$table1)
    expect_equal(x$table2, x2$table2)
    expect_equal(x$variables1, x2$variables1)
    expect_equal(x$variables2, x2$variables2)
    expect_equal(x$cells1, x2$cells1)
    expect_equal(x$cells2, x2$cells2)
    expect_equal(x$acrosscells, x2$acrosscells)
    expect_equal(x$data, x2$data)
    expect_equal(x, x2)
  }

  expect_equal_all_px_element(px(population_gl))
  expect_equal_all_px_element(px(population_gl), save_dataset = TRUE)
  expect_equal_all_px_element(px(get_px_file_path("BEXSTA_windows_1252")))
  expect_equal_all_px_element(
    px(get_px_file_path("BEXSTA_windows_1252")),
    save_dataset = TRUE
  )
  expect_equal_all_px_element(px(get_px_file_path("SOXATI4")))
  expect_equal_all_px_element(
    px(get_px_file_path("SOXATI4")),
    save_dataset = TRUE
  )
  expect_equal_all_px_element(px(get_px_file_path("TUX01")))
  expect_equal_all_px_element(
    px(get_px_file_path("TUX01")),
    save_dataset = TRUE
  )

  # A couple of tests doens't work, but the functionality is 80 there

  # UNITS in wrong because of (#364)
  # expect_equal_all_px_element(px(get_px_file_path('CONTVARIABLE')))
  # expect_equal_all_px_element(
  #   px(get_px_file_path('CONTVARIABLE_multiple_languages'))
  # )

  # TITLE differ, which is fine ("" vs NULL)
  # TIMEVAL differ, easy fix
  # expect_equal_all_px_element(px(get_px_file_path('no_timeval_or_codes2')))

  # Something about TIMEVAL causes problems
  # expect_equal_all_px_element(px(get_px_file_path('PRXPRISH')))
})

test_that("Source data variable names are preserved", {
  table_name <- "FOTEST"
  excel1 <- get_metadata_path(table_name)
  excel2 <- temp_xlsx_file()

  excel1 |>
    px() |>
    px_save(path = excel2)

  get_data_sheet_variable_names <- function(path) {
    path |>
      readxl::read_excel(sheet = "Data") |>
      names()
  }

  expect_equal(
    get_data_sheet_variable_names(excel1),
    get_data_sheet_variable_names(excel2)
  )
})

test_that("PX-file and px_save(px(PX-file), path='*.px') are equivalent", {
  # Some files don't give an exact match in the PX-file because there can be
  # subtle differences in formatting. PxJob is run on these file to test
  # that they are equivalent.
  skip_if_not_installed("pxjob64Win", minimum_version = "1.1.0")

  run_px_px_save_and_expect_equal <- function(table_name) {
    px_in <- get_px_file_path(table_name)
    px_out <- temp_px_file()

    px_in |>
      px() |>
      px_save(path = px_out)

    pxjob_clean(input = px_in, output = pxjob_in <- temp_px_file())
    pxjob_clean(input = px_out, output = pxjob_out <- temp_px_file())

    px_file_as_lines <- function(path) {
      ignore_keywords_regex <-
        c("VARIABLECODE") |>
        (\(.) {
          paste0("^", ., ".+")
        })() |>
        paste(collapse = "|")

      lines <-
        path |>
        readLines() |>
        stringr::str_subset(ignore_keywords_regex, negate = TRUE)

      if (table_name %in% c("no_timeval_or_codes2")) {
        lines <- stringr::str_subset(lines, "^CODES.+", negate = TRUE)
      }

      if (table_name %in% c("CONTVARIABLE_multiple_languages")) {
        lines <- stringr::str_subset(
          lines, '^VARIABLE-TYPE.+Time"',
          negate = TRUE
        )
        lines <- stringr::str_subset(lines, "META-ID", negate = TRUE)
      }

      lines
    }

    expect_equal(px_file_as_lines(pxjob_in), px_file_as_lines(pxjob_out))
  }

  run_px_px_save_and_expect_equal("BEXSTA_windows_1252")
  run_px_px_save_and_expect_equal("CONTVARIABLE")
  run_px_px_save_and_expect_equal("CONTVARIABLE_multiple_languages")
  run_px_px_save_and_expect_equal("no_timeval_or_codes2")
  run_px_px_save_and_expect_equal("SOXATI4")
  run_px_px_save_and_expect_equal("TUX01")
})

#' Add total level to variable
#'
#' Add a new level to a variables which is the sum of all other levels.
#'
#' @inheritParams add_totals_to_df
#' @param variable Name of variable to add total level to.
#' @param level_name Total level name.
#'
#' @returns A data frame
#' @keywords internal
add_total_level_to_var <- function(df,
                                   variable,
                                   level_name = "Total",
                                   sum_var = "value",
                                   na_rm = TRUE) {
  column_order <- names(df)

  df |>
    dplyr::group_by(across(-any_of(c(variable, sum_var)))) |>
    dplyr::summarise(
      !!sum_var := sum(!!rlang::sym(sum_var), na.rm = !!na_rm)
    ) |>
    dplyr::ungroup() |>
    dplyr::mutate({{ variable }} := level_name) |>
    dplyr::bind_rows(df) |>
    dplyr::relocate(all_of(column_order))
}

#' Add total levels to multiple variables
#'
#' Wrapper around \link{add_total_level_to_var} to add levels to total level to
#' multiple variables.
#'
#' @inheritParams px_add_totals
#' @param df Data frame to add total levels to.
#' @param variables List of variables to add total levels to.
#' @param level_names Names of total levels. Should have length 1 or same length
#' as `variables`.
#' @param sum_var String, name of variable to sum over.
#' @keywords internal
add_totals_to_df <- function(df,
                             variables,
                             level_names,
                             sum_var = "value",
                             na_rm = TRUE) {
  params <- data.frame(variables = variables, level_names = level_names)

  for (i in seq_len(nrow(params))) {
    df <- add_total_level_to_var(df,
      variable = params$variables[i],
      level_name = params$level_names[i],
      sum_var = sum_var,
      na_rm = !!na_rm
    )
  }

  df
}

#' @rdname px_add_totals.px
#' @export
px_add_totals <- function(x, value, na_rm = TRUE, validate) {
  UseMethod("px_add_totals")
}

#' @title Add total levels to variables
#'
#' @description
#' Adds a total level, which is the sum of the figures for all other levels of
#' the variable.
#'
#' The default name of the total level is 'Total', unless \link{px_elimination}
#' is set, in which case the elimination code is used.
#'
#' @param x A px object
#' @param value A character vector of variables to add total levels to.
#' @param na_rm Optional. Logical. If TRUE, NAs are removed before summing.
#' @eval param_validate()
#'
#' @returns A px object
#'
#' @seealso [px_elimination]
#'
#' @examples
#' # Create small px object example
#' x0 <- px(subset(population_gl, age == "65+"))
#' px_data(x0)
#'
#' # Add total level to one variable
#' x1 <- px_add_totals(x0, "gender")
#' px_data(x1)
#'
#' # Add total level to multiple variables
#' x2 <- px_add_totals(x0, c("gender", "age"))
#' px_data(x2)
#'
#' # The name of the total level can be changed with px_elimination()
#' x3 <-
#'   x0 |>
#'   px_elimination("T") |>
#'   px_add_totals("gender")
#'
#' px_data(x3)
#'
#' @export
px_add_totals.px <- function(x, value, na_rm = TRUE, validate = TRUE) {
  elimination <-
    px_elimination(x)

  values_not_in_elimination <-
    value[!value %in% elimination$`variable-code`]

  if (length(values_not_in_elimination) > 0) {
    default_elimination <-
      dplyr::tibble(
        `variable-code` = values_not_in_elimination,
        elimination = "Total"
      )

    x <- px_elimination(x, dplyr::bind_rows(elimination, default_elimination))
  }

  params <-
    x$variables1 |>
    dplyr::filter(.data$`variable-code` %in% value) |>
    dplyr::distinct(.data$`variable-code`, .data$elimination)

  x$data <-
    add_totals_to_df(x$data,
      variables = params$`variable-code`,
      level_names = params$elimination,
      sum_var = px_figures(x),
      na_rm = !!na_rm
    )

  return_px(x, validate)
}

test_that("Total is added to one variable", {
  input <- tibble::tribble(
    ~cohort, ~sex, ~value,
    "a", "m", NA,
    "a", "f", 3,
    "b", "m", 2,
    "b", "f", 4
  )

  expect <- tibble::tribble(
    ~cohort, ~sex, ~value,
    "a", "m", NA,
    "a", "f", 3,
    "b", "m", 2,
    "b", "f", 4,
    "Total", "m", 2,
    "Total", "f", 7
  )

  expect_equal(
    add_total_level_to_var(input, "cohort") |> dplyr::arrange_all(),
    expect |> dplyr::arrange_all()
  )
})


test_that("Totals are added to two variables", {
  read_data <- function(table_name) {
    table_name |>
      get_data_path() |>
      readRDS() |>
      dplyr::as_tibble()
  }

  bexsta <- read_data("BEXSTA")
  bexsta_without_totals <- read_data("BEXSTA_WITHOUT_TOTALS")

  output <-
    add_totals_to_df(
      df = bexsta_without_totals,
      variables = c("place of birth", "gender"),
      level_name = "T",
      sum_var = "persons"
    ) |>
    dplyr::arrange_all()

  expect_equal(dplyr::arrange_all(output), dplyr::arrange_all(bexsta))

  # Test having multiple level_names
  bexsta2 <-
    bexsta |>
    dplyr::mutate(gender = ifelse(gender == "T", "Total", gender))

  output2 <-
    add_totals_to_df(
      df = bexsta_without_totals,
      variables = c("place of birth", "gender"),
      level_names = c("T", "Total"),
      sum_var = "persons"
    ) |>
    dplyr::arrange_all()

  expect_equal(dplyr::arrange_all(output2), dplyr::arrange_all(bexsta2))
})


test_that("adds total levels to data without them", {
  metadata_path <- get_metadata_path("BEXSTA")
  px1 <- temp_px_file()
  px2 <- temp_px_file()

  px(input = metadata_path, data = get_data_path("BEXSTA")) |>
    px_save(path = px1)

  px(input = metadata_path, data = get_data_path("BEXSTA_WITHOUT_TOTALS")) |>
    px_add_totals(c("place of birth", "gender")) |>
    px_save(path = px2)

  expect_equal_lines(px1, px2)
})

test_that("metadata without elimination values use default value", {
  levels <-
    px(
      input = get_metadata_path("BEXSTA"),
      data = get_data_path("BEXSTA_WITHOUT_TOTALS")
    ) |>
    px_elimination(NULL) |>
    px_add_totals(value = "gender") |>
    magrittr::extract2("data") |>
    dplyr::pull(gender) |>
    unique()

  expect_equal(levels, c("Total", "K", "M"))
})

test_that("do not ignore NA when summing", {
  df <- tibble::tribble(
    ~category, ~n,
    "a", 10,
    "b", 20,
    "c", NA
  )

  result <-
    px(df) |>
    px_add_totals("category", na_rm = FALSE) |>
    magrittr::extract2("data") |>
    dplyr::pull(n)

  expect_identical(result, c(NA, 10, 20, NA))
})

test_that("Add totals work for multiple languages", {
  # Create px object with multilingual add_totals
  x <-
    population_gl |>
    px() |>
    px_languages(c("en", "da")) |>
    px_elimination(dplyr::tribble(
      ~`variable-code`, ~elimination,
      "gender", "Total"
    )) |>
    px_add_totals("gender")

  # px_add_totals adds both 'en' and 'da' level in data. This is a mistake.
  gender_levels <-
    x |>
    px_data() |>
    dplyr::distinct(gender) |>
    dplyr::pull(gender)

  expect_identical(gender_levels, c("Total", "male", "female"))
})

test_that("Add totals adds elimination", {
  default_elimination <-
    population_gl |>
    px() |>
    px_add_totals("gender") |>
    px_elimination()

  expect_identical(
    default_elimination,
    dplyr::tibble(
      "variable-code" = "gender",
      "elimination" = "Total"
    )
  )
})

test_that("Elimination and values are preserved when changing data frame", {
  tmp_px_file <- temp_px_file()

  population_gl |>
    px() |>
    px_add_totals("gender") |>
    px_save(tmp_px_file)

  x <-
    px(tmp_px_file) |>
    px_data(population_gl)

  expect_identical(
    px_elimination(x),
    dplyr::tibble(
      "variable-code" = "gender",
      "elimination" = "Total"
    )
  )

  elimination_values <-
    x |>
    px_values() |>
    dplyr::semi_join(px_elimination(x),
      by = c(
        "variable-code" = "variable-code",
        "code" = "elimination"
      )
    ) |>
    dplyr::select("variable-code", "elimination" = "code")

  expect_identical(elimination_values, px_elimination(x))
})

#' Change pivot variables
#'
#' @param x A px object
#' @param pivot Pivot type (STUB, HEADING, FIGURES)
#' @param value A character vector of variable codes to change to the pivot
#' type
#'
#' @returns A px object
#' @keywords internal
change_pivot_variables <- function(x, value, pivot) {
  old_pivot_variables <- get_pivot_variables(x, pivot)

  new_pivot_variables <- unique(c(value, old_pivot_variables))

  order_df <- dplyr::tibble(
    `variable-code` = new_pivot_variables,
    order = seq_along(new_pivot_variables)
  )

  x$variables1 <-
    modify_or_add_row(
      df = x$variables1,
      lookup_column = "variable-code",
      lookup_column_values = value,
      modify_column = "pivot",
      new_value = pivot
    ) |>
    dplyr::left_join(order_df, by = "variable-code") |>
    dplyr::mutate(order = ifelse(toupper(pivot) == !!pivot,
      .data$order.y,
      .data$order.x
    )) |>
    dplyr::select(-all_of(c("order.y", "order.x"))) |>
    align_data_frames(get_base_variables1()) |>
    sort_variables1()

  new_acrosscells_base <- get_base_acrosscells(c(px_stub(x), px_heading(x)))

  x$acrosscells <-
    x$acrosscells |>
    align_data_frames(new_acrosscells_base) |>
    dplyr::select(names(new_acrosscells_base))

  x
}

#' Get names of pivot variables
#'
#' @param x A px object
#' @param pivot A string, either "STUB", "HEADING" or "FIGURES"
#'
#' @returns A character vector of variable codes
#' @keywords internal
get_pivot_variables <- function(x, pivot) {
  x$variables1 |>
    dplyr::filter(toupper(.data$pivot) == !!pivot) |>
    dplyr::arrange(.data$order, .data$`variable-code`) |>
    dplyr::pull("variable-code")
}

#' @rdname px_stub.px
#' @export
px_stub <- function(x, value, validate) {
  UseMethod("px_stub")
}

#' @eval add_documentation_head_stub("STUB")
#' @seealso \code{\link{px_heading}} \code{\link{px_figures}}
#' @examples
#' x1 <- px(population_gl)
#' # Print STUB
#' px_stub(x1)
#' # Add 'year' to STUB
#' x2 <- px_stub(x1, "year")
#' px_stub(x2)
#'
#' # Change order of STUB
#' x3 <- px_stub(x2, c("age", "gender"))
#' px_stub(x3)
px_stub.px <- function(x, value, validate = TRUE) {
  if (missing(value)) {
    return(get_pivot_variables(x, "STUB"))
  }

  return_px(change_pivot_variables(x, value, "STUB"), validate)
}

#' @rdname px_heading.px
#' @export
px_heading <- function(x, value, validate) {
  UseMethod("px_heading")
}

#' @eval add_documentation_head_stub("HEADING")
#' @seealso \code{\link{px_stub}} \code{\link{px_figures}}
#' @examples
#' x1 <- px(population_gl)
#'
#' # Print HEADING
#' px_heading(x1)
#'
#' # Add 'gender' to HEADING
#' x2 <- px_heading(x1, "gender")
#' px_heading(x2)
#'
#' # Change order of HEADING
#' x3 <- px_heading(x2, "year")
#' px_heading(x3)
px_heading.px <- function(x, value, validate = TRUE) {
  if (missing(value)) {
    return(get_pivot_variables(x, "HEADING"))
  }

  return_px(change_pivot_variables(x, value, "HEADING"), validate)
}

#' @rdname px_figures.px
#' @export
px_figures <- function(x, value, validate) {
  UseMethod("px_figures")
}

#' @title Change figures variable
#' @description
#' Inspect or change which variable is used as figures. The previous figures
#' variable is changed to STUB. There can only be one figures variable.
#'
#' @param x A px object
#' @param value Optional. Name of variable to use as FIGRUES. If missing, the
#' current PX_FIGURES variable is returned.
#' @eval param_validate()
#'
#' @returns A px object or a character string
#'
#' @seealso \code{\link{px_stub}} \code{\link{px_heading}}
#'
#' @examples
#' x1 <- px(population_gl)
#'
#' # Print FIGURES
#' px_figures(x1)
#'
#' # Change 'age' to FIGURES variable, 'n' i changed to STUB
#' x2 <- px_figures(x1, "age")
#' px_figures(x2)
#' px_stub(x2)
#' @export
px_figures.px <- function(x, value, validate = TRUE) {
  if (missing(value)) {
    return(get_pivot_variables(x, "FIGURES"))
  }

  error_if_not_exactly_one_figures_variable(value)

  old_figures_variable <- px_figures(x)

  x <- change_pivot_variables(x, value, "FIGURES")

  x <- change_pivot_variables(x, old_figures_variable, "STUB")

  x$cells1 <-
    x$cells1 |>
    dplyr::filter(!.data$`variable-code` %in% !!value)

  x$cells2 <-
    x$cells2 |>
    dplyr::filter(!.data$`variable-code` %in% !!value)

  return_px(x, validate)
}

#' @rdname px_timeval.px
#' @export
px_timeval <- function(x, value, validate) {
  UseMethod("px_timeval")
}

#' @eval add_documentation_table1("TIMEVAL", "year")
#' @description There can only be one time variable.
px_timeval.px <- function(x, value, validate = TRUE) {
  if (missing(value)) {
    return(get_variable1_logic_value(x, "timeval"))
  } else if (is.null(value)) {
    x$variables1$timeval <- FALSE
  } else {
    x$variables1$timeval <- FALSE

    x <- modify_variables1(
      x, "timeval",
      dplyr::tibble(
        `variable-code` = value,
        timeval = TRUE
      )
    )
  }

  return_px(x, validate)
}

#' @rdname px_contvariable.px
#' @export
px_contvariable <- function(x, value, validate) {
  UseMethod("px_contvariable")
}

#' @eval add_doc_keyword_function_intro("CONTVARIABLE")
#' @description Setting CONTVARIABLE indexes several variables in table2.
#' Removing CONTVARIABLE removes the indexing from table2.
#' @param value `r table1_param_value("CONTVARIABLE")`
#' @eval param_validate()
#' @eval add_return_px_or_char_str()
#' @examples
#' # Set CONTVARIABLE
#' x1 <-
#'   px(population_gl) |>
#'   px_contvariable("gender")
#'
#' # After setting CONTVARIABLE some variables are index by it, e.g. UNITS
#' px_units(x1)
#'
#' # Remove CONTVARIABLE
#' x2 <- px_contvariable(x1, NULL)
#' px_contvariable(x2)
#'
#' # Removing CONTVARIABLE also removes the index from UNITS
#' px_units(x2)
#' @export
px_contvariable.px <- function(x, value, validate = TRUE) {
  if (missing(value)) {
    return(get_variable1_logic_value(x, "contvariable"))
  } else if (is.null(value)) {
    x$variables1$contvariable <- FALSE

    previously_indexed_by_contvariable <-
      x$table2 |>
      dplyr::filter(.data$keyword %in% keywords_indexed_by_contvariable()) |>
      dplyr::group_by(.data$keyword, .data$language) |>
      dplyr::slice(1)

    x$table2 <-
      x$table2 |>
      dplyr::filter(!.data$keyword %in% keywords_indexed_by_contvariable()) |>
      dplyr::bind_rows(previously_indexed_by_contvariable) |>
      dplyr::mutate(code = NA_character_)
  } else {
    x$variables1$contvariable <- FALSE

    x <- modify_variables1(
      x, "contvariable",
      dplyr::tibble(
        "variable-code" = value,
        "contvariable" = TRUE
      )
    )


    contvariable_values <-
      x |>
      px_values() |>
      dplyr::filter(.data$`variable-code` %in% c(!!value)) |>
      dplyr::select(-c("variable-code", "code")) |>
      dplyr::rename("code" = "value") |>
      (\(.) {
        if (!"language" %in% colnames(.)) {
          dplyr::mutate(., language = NA_character_)
        } else {
          .
        }
      })()

    indexed_by_contvariable <-
      x$table2 |>
      dplyr::filter(.data$keyword %in% keywords_indexed_by_contvariable()) |>
      dplyr::select(-"code") |>
      dplyr::left_join(contvariable_values,
        by = "language",
        relationship = "many-to-many"
      )

    x$table2 <-
      x$table2 |>
      dplyr::bind_rows(indexed_by_contvariable) |>
      sort_table2(languages = defined_languages(x))
  }

  return_px(x, validate)
}

#' @rdname px_elimination.px
#' @export
px_elimination <- function(x, value, validate) {
  UseMethod("px_elimination")
}

#' @eval add_doc_keyword_function_intro("ELIMINATION")
#' @param value Optional. Use character to set ELIMINATION for all STUB and
#' HEADING variables. Use a data frame with columns 'variable-code' and
#' 'elimination' to set ELIMINATION for individual variables. If value is
#' missing, the current ELIMINATION is returned. If NULL, ELIMINATION is removed
#' for all variables.
#' @eval param_validate()
#' @eval add_return_px_or_df()
#' @examples
#' library(tibble)
#'
#' # Set ELIMINATION
#' x1 <-
#'   px(population_gl) |>
#'   px_elimination(tribble(
#'     ~`variable-code`, ~elimination,
#'     "gender", "T",
#'     "age", "YES"
#'   ))
#'
#' # Print ELIMINATION
#' px_elimination(x1)
#'
#' # Remove ELIMINATION
#' x2 <- px_elimination(x1, NULL)
#' px_elimination(x2)
#'
#' @export
px_elimination.px <- function(x, value, validate = TRUE) {
  if (missing(value)) {
    return(get_variable1_value(x, "elimination"))
  } else if (is.null(value)) {
    x$variables1$elimination <- NA
  } else if (is.character(value)) {
    x$variables1$elimination <- NA

    x <- modify_variables1(
      x, "elimination",
      dplyr::tibble(
        `variable-code` = c(px_stub(x), px_heading(x)),
        elimination = value
      )
    )
  } else {
    x <- modify_variables1(x, "elimination", value)
  }

  return_px(x, validate)
}


#' @rdname px_variable_type.px
#' @export
px_variable_type <- function(x, value, validate) {
  UseMethod("px_variable_type")
}

#' @eval add_doc_keyword_function_intro("VARIABLE-TYPE")
#' @param value A data frame with columns 'variable-code' and 'variable-type'.
#' If value is missing, the current VARIABLE-TYPE is returned. If NULL, all
#' VARIABLE-TYPE is removed.
#' @eval param_validate()
#' @eval add_return_px_or_df()
#' @examples
#' library(tibble)
#'
#' # Set VARIABLE-TYPE
#' x1 <-
#'   px(population_gl) |>
#'   px_variable_type(
#'     tibble("variable-code" = "year", "variable-type" = "time")
#'   )
#'
#' # Print VARIABLE-TYPE
#' px_variable_type(x1)
#'
#' # Remove VARIABLE-TYPE
#' x2 <- px_variable_type(x1, NULL)
#' px_variable_type(x2)
#'
#' @export
px_variable_type.px <- function(x, value, validate = TRUE) {
  if (missing(value)) {
    return(get_variable1_value(x, "variable-type"))
  } else if (is.null(value)) {
    x$variables1$`variable-type` <- NA
  } else {
    x <- modify_variables1(x, "variable-type", value)
  }

  return_px(x, validate)
}

test_that("Data table can be returned sorted", {
  x1 <- px_from_table_name("BEXSTA")

  expect_error(px_data(x1, value = population_gl, sort = TRUE),
    regexp = "'sort' can only be used if 'value' is missing"
  )

  expect_error(px_data(x1, sort = "yes"),
    regexp = "must be TRUE or FALSE."
  )

  # Change order
  total_level_last <- dplyr::tribble(
    ~`variable-code`, ~code, ~order,
    "gender", "T", 4
  )

  x2 <-
    x1 |>
    px_order(
      px_order(x1) |>
        dplyr::anti_join(total_level_last,
          by = dplyr::join_by(
            `variable-code`,
            code
          )
        ) |>
        dplyr::bind_rows(total_level_last)
    )

  expect_identical(px_data(x1), px_data(x2))
  expect_identical(px_data(x1, labels = TRUE), px_data(x2, labels = TRUE))

  expect_false(identical(
    px_data(x1, sort = TRUE),
    px_data(x2, sort = TRUE)
  ))

  expect_false(identical(
    px_data(x1, labels = TRUE, sort = TRUE),
    px_data(x2, labels = TRUE, sort = TRUE)
  ))

  x2_data_expected_codes <-
    px_data(x2) |>
    dplyr::arrange(
      match(`place of birth`, c("T", "N", "S")),
      match(gender, c("M", "K", "T"))
    )

  x2_data_expected_labels <-
    px_data(x2, labels = TRUE) |>
    dplyr::arrange(
      match(`place of birth`, c("Total", "Greenland", "Outside Greenland")),
      match(gender, c("Men", "Women", "Total"))
    )

  expect_identical(
    x2_data_expected_codes,
    px_data(x2, sort = TRUE)
  )

  expect_identical(
    x2_data_expected_labels,
    px_data(x2, label = TRUE, sort = TRUE)
  )
})


test_that("data is modified", {
  x <-
    "BEXSTA" |>
    get_data_path() |>
    readRDS() |>
    px()

  expect_identical(x$data, px_data(x))

  mini_data <- x$data[1:5, ]

  x2 <- px_data(x, mini_data)

  expect_identical(mini_data, px_data(x2))

  x3 <- px_data(x, NULL)

  expect_identical(dplyr::filter(x$data, FALSE), px_data(x3))
})

test_that("modifying data updates metadata", {
  # Use a new data set where one variable changes name and one changes levels.
  population_gl_new <-
    population_gl |>
    dplyr::mutate(age = dplyr::if_else(age %in% c("0-6", "7-16"),
      "0-16",
      age
    )) |>
    dplyr::group_by(gender, age, year) |>
    dplyr::summarise(n = sum(n), .groups = "keep") |>
    dplyr::ungroup() |>
    dplyr::mutate(year = stringr::str_sub(year, 3, 4)) |>
    dplyr::rename(shortyear = year)

  x1 <-
    population_gl |>
    px()

  x2 <-
    x1 |>
    px_data(population_gl_new)


  ## Variables
  old_names <- names(population_gl)
  new_names <- names(population_gl_new)
  new_names_stub_heading <- setdiff(new_names, "n")
  removed_vars <- setdiff(old_names, new_names)

  expect_identical(x1$languages, x2$languages)
  expect_identical(x1$table1, x2$table1)
  expect_identical(x1$table2, x2$table2)

  expect_true(all(new_names %in% x2$variables1$`variable-code`))
  expect_true(all(new_names %in% x2$variables2$`variable-code`))
  expect_true(all(new_names_stub_heading %in% x2$cells1$`variable-code`))
  expect_true(all(new_names_stub_heading %in% x2$cells2$`variable-code`))

  expect_false(any(removed_vars %in% x2$variables1$`variable-code`))
  expect_false(any(removed_vars %in% x2$variables2$`variable-code`))
  expect_false(any(removed_vars %in% x2$cells1$`variable-code`))
  expect_false(any(removed_vars %in% x2$cells2$`variable-code`))

  expect_identical(
    x2$acrosscell,
    get_base_acrosscells(c(px_stub(x2), px_heading(x2)))
  )

  ## Codes
  data_set_codes <- function(df) {
    df |>
      dplyr::select(-n) |>
      tidyr::pivot_longer(everything(),
        names_to = "variable-code",
        values_to = "code"
      ) |>
      dplyr::distinct_all()
  }

  old_codes <- data_set_codes(population_gl)

  df_codes <- data_set_codes(population_gl_new)

  removed_codes <-
    old_codes |>
    dplyr::anti_join(df_codes, by = dplyr::join_by(`variable-code`, code))

  new_codes <-
    df_codes |>
    dplyr::anti_join(old_codes, by = dplyr::join_by(`variable-code`, code))

  dplyr::semi_join(removed_codes, x2$cells1,
    by = dplyr::join_by(`variable-code`, code)
  ) |>
    nrow() |>
    expect_equal(0)

  expect_no_matches <- function(df1, df2) {
    dplyr::semi_join(df1, df2,
      by = dplyr::join_by(`variable-code`, code)
    ) |>
      nrow() |>
      expect_equal(0)
  }

  expect_no_matches(removed_codes, x2$cells1)
  expect_no_matches(removed_codes, x2$cells2)

  expect_all_matches <- function(df1, df2) {
    dplyr::anti_join(df1, df2,
      by = dplyr::join_by(`variable-code`, code)
    ) |>
      nrow() |>
      expect_equal(0)
  }

  expect_all_matches(new_codes, x2$cells1)
  expect_all_matches(new_codes, x2$cells2)

  ## Multilingual
  x_lang1 <-
    population_gl |>
    px() |>
    px_languages(c("en", "fr"))

  x_lang2 <-
    x_lang1 |>
    px_data(population_gl_new)

  x_lang2$variables2 |>
    dplyr::filter(`variable-code` == "shortyear") |>
    dplyr::pull(language) |>
    expect_identical(c("en", "fr"))

  expect_identical(nrow(x_lang2$cells2), 18L)
  expect_identical(x_lang2$cells2$language, rep(c("en", "fr"), 9))
})

test_that("Elimination values and order is preserved ", {
  elimination_df <- dplyr::tibble(
    `variable-code` = "gender",
    elimination = "all genders"
  )

  elimination_order <- dplyr::tibble(
    `variable-code` = "gender",
    code = "all genders",
    order = 0
  )
  x1 <-
    population_gl |>
    px() |>
    px_elimination(elimination_df) |>
    px_order(elimination_order)

  population_gl_2024 <-
    population_gl |>
    dplyr::filter(year == 2024)

  x2 <-
    x1 |>
    px_data(population_gl_2024)

  expect_identical(px_elimination(x2), elimination_df)

  px_order(x2) |>
    dplyr::semi_join(elimination_df,
      by = c(
        "variable-code" = "variable-code",
        "code" = "elimination"
      )
    ) |>
    expect_identical(elimination_order)
})

test_that("Labels can be returned", {
  x1 <- px_from_table_name("BEXSTA")

  expect_error(px_data(x1, value = population_gl, labels = TRUE),
    regexp = "can only be used"
  )

  expect_error(px_data(x1, labels = "fr"),
    regexp = "Language.*is not defined"
  )

  expect_identical(
    px_data(x1, labels = TRUE),
    px_data(x1, labels = "en")
  )

  expect_identical(
    head(px_data(x1, labels = "kl"), 3),
    tibble::tribble(
      ~`place of birth`,  ~gender,  ~time, ~persons,
      "Kalaallit Nunaat", "Arnat", "2018",    24392,
      "Kalaallit Nunaat", "Arnat", "2019",    24434,
      "Kalaallit Nunaat", "Arnat", "2020",    24452
    )
  )

  x2 <- px(population_gl)

  expect_identical(
    px_data(x2, labels = TRUE),
    px_data(x2)
  )
})

test_that("px_micro creates PX-files correctly", {
  skip_if_not_installed("pxjob64Win", minimum_version = "1.1.0")

  expect_that_micro_files_are_correct <- function(x) {
    n_microvars <- length(names(x$data)) - length(px_heading(x))

    out_dir <- temp_dir()

    px_micro(x, out_dir = out_dir)

    px_paths <- list.files(out_dir, full.names = TRUE)

    expect_identical(n_microvars, length(px_paths))

    for (px_path in px_paths) {
      x_micro <- px(px_path)
      micro_var <- px_stub(x_micro)

      expect_true(stringr::str_detect(px_path, micro_var))

      expect_identical(px_heading(x), px_heading(x_micro))

      expect_identical(px_figures(x_micro), "n")

      expect_identical(
        px_note(x_micro),
        list(
          "Table note",
          dplyr::tibble(
            `variable-code` = micro_var,
            note = paste0("note for ", micro_var)
          )
        )
      )
    }
  }

  get_data_path("micro") |>
    readRDS() |>
    dplyr::as_tibble() |>
    dplyr::select(taar, sex, civst, alder) |>
    dplyr::mutate(alder = cut(alder,
      breaks = c(0, 20, 40, 60, 80, 100),
      labels = c("0-19", "20-39", "40-59", "60-79", "80+")
    )) |>
    px() |>
    px_stub("civst") |>
    px_timeval("taar") |>
    px_heading(c("taar", "sex")) |>
    px_note("Table note") |>
    px_note(dplyr::tibble(
      `variable-code` = c("civst", "alder"),
      note = paste0("note for ", `variable-code`)
    )) |>
    expect_that_micro_files_are_correct()
})

test_that("px_micro creates valid PX-files", {
  skip_if_not_installed("pxjob64Win", minimum_version = "1.1.0")

  expect_that_pxjob_runs_without_errors <- function(px) {
    out_dir <- temp_dir()

    px_micro(px, out_dir = out_dir)

    px_paths <- list.files(out_dir, full.names = TRUE)

    for (px_path in px_paths) {
      output <- temp_px_file()
      pxjob_exit_code <- pxjob64Win::pxjob(px_path, output)

      expect_equal(0, pxjob_exit_code, info = px_path)
    }
  }

  get_data_path("micro") |>
    readRDS() |>
    dplyr::as_tibble() |>
    dplyr::mutate(
      sidedoer = stringr::str_trim(sidedoer),
      sidedoer = dplyr::na_if(sidedoer, ""),
      pnr = NA
    ) |>
    px() |>
    px_timeval("taar") |>
    expect_that_pxjob_runs_without_errors()
})

test_that("px_micro can control data for individual tables", {
  set.seed(1)

  df <-
    get_data_path("micro") |>
    readRDS() |>
    dplyr::as_tibble() |>
    dplyr::select(taar, civst) |>
    dplyr::mutate(
      study = sample(c("A", "B"), size = dplyr::n(), replace = TRUE)
    )


  table_level <-
    dplyr::tribble(
      ~variable, ~px_description,
      "taar", "Year",
      "civst", "Civil status"
    ) |>
    dplyr::mutate(px_matrix = variable)

  out_dir <- temp_dir()

  px(df) |>
    px_stub(names(df)) |>
    px_heading("study") |>
    px_micro(
      out_dir = out_dir,
      keyword_values = table_level
    )

  px_paths <- list.files(out_dir, full.names = TRUE)

  for (px_path in px_paths) {
    x_micro <- px(px_path)
    micro_var <- px_stub(x_micro)

    micro_table_level <-
      table_level |>
      dplyr::filter(variable == micro_var)

    expect_equal(px_description(x_micro), micro_table_level$px_description)
    expect_equal(px_matrix(x_micro), micro_table_level$px_matrix)
  }
})

test_that("keyword_values are multilingual", {
  x <-
    greenlanders |>
    px() |>
    px_language("en") |>
    px_languages(c("en", "kl")) |>
    px_stub(names(greenlanders)) |>
    px_heading("cohort")

  keyword_values <-
    dplyr::tribble(
      ~variable, ~language, ~px_description, ~px_matrix,
      "age", "en", "Age", "gl",
      "age", "kl", "Ukiut", NA,
      "gender", "en", "Gender", "ge",
      "gender", "kl", "Suiaassuseq", "ge"
    )

  out_dir <- temp_dir()

  px_micro(x, out_dir = out_dir, keyword_values = keyword_values)

  px_age <- px(file.path(out_dir, "age.px"))
  keyword_values_age <- dplyr::filter(keyword_values, variable == "age")

  expect_identical(
    px_description(px_age),
    keyword_values_age |>
      dplyr::select(language, value = px_description)
  )

  expect_identical(
    px_matrix(px_age),
    keyword_values_age |>
      tidyr::drop_na(px_matrix) |>
      dplyr::pull(px_matrix)
  )

  px_gender <- px(file.path(out_dir, "gender.px"))
  keyword_values_gender <- dplyr::filter(keyword_values, variable == "gender")

  expect_identical(
    px_description(px_gender),
    keyword_values_gender |>
      dplyr::select(language, value = px_description)
  )

  expect_identical(
    px_matrix(px_gender),
    keyword_values_gender |>
      tidyr::drop_na(px_matrix) |>
      dplyr::distinct(px_matrix) |>
      dplyr::pull(px_matrix)
  )
})

test_that("px_micro can control filenames", {
  df <-
    get_data_path("micro") |>
    readRDS() |>
    dplyr::as_tibble() |>
    dplyr::select(1:4)

  out_dir <- temp_dir()

  filename_df <-
    dplyr::tibble(
      variable = names(df),
      filename = paste0("micro_", variable, ".px")
    ) |>
    dplyr::arrange(variable) |>
    head(2)

  px(df) |>
    px_stub(names(df)) |>
    px_micro(out_dir = out_dir, keyword_values = filename_df)

  expect_equal(
    list.files(out_dir),
    c(filename_df$filename, "pnrmor.px", "taar.px")
  )
})

test_that("px_micro removes headings where all values are NA", {
  df <-
    get_data_path("micro") |>
    readRDS() |>
    dplyr::as_tibble() |>
    dplyr::select(1:2) |>
    dplyr::mutate(pnr = ifelse(taar == 1994, NA, pnr)) |>
    dplyr::arrange_all()

  out_dir <- temp_dir()

  px(df) |>
    px_stub("pnr") |>
    px_heading("taar") |>
    px_micro(out_dir = out_dir)

  micro_df <-
    px(list.files(out_dir, full.names = TRUE))$data

  target <-
    df |>
    dplyr::count(taar, pnr) |>
    tidyr::drop_na(pnr) |>
    dplyr::select(pnr, taar, n) |>
    dplyr::mutate(n = as.double(n))

  expect_identical(micro_df, target)
})

test_that("Variables is modified", {
  x <-
    "BEXSTA" |>
    get_data_path() |>
    readRDS() |>
    px()

  expect_equal(px_stub(x), c("place of birth", "gender"))

  # Change time to stub variable
  x2 <- px_stub(x, value = "time")
  expect_equal(px_stub(x2), c("time", "place of birth", "gender"))
  expect_equal(names(x2$variables1), names(get_base_variables1()))

  # Change order
  new_order <- c("gender", "time", "place of birth")
  x3 <- px_stub(x, value = new_order)
  expect_equal(px_stub(x3), new_order)

  x4 <- px_heading(x, value = new_order)
  expect_equal(px_heading(x4), new_order)
  expect_equal(px_stub(x4), as.character())

  x5 <- px_figures(x, value = "time")

  expect_equal(px_figures(x5), "time")
  expect_equal(px_heading(x5), as.character())
  expect_equal(px_stub(x5), c("persons", "place of birth", "gender"))

  expect_equal(px_timeval(x), NULL)

  x6 <- px_heading(x5, c("gender"))

  x6_stub_vars <- px_stub(x6)

  expect_identical(
    dplyr::filter(x5$variables1, `variable-code` %in% x6_stub_vars),
    dplyr::filter(x6$variables1, `variable-code` %in% x6_stub_vars)
  )

  x7 <- px_timeval(x, value = "time")
  expect_equal(px_timeval(x7), "time")

  x8 <- px_timeval(x7, value = "gender")
  expect_equal(px_timeval(x8), "gender")

  x9 <- px_timeval(x8, NULL)
  expect_equal(px_timeval(x9), NULL)

  elimination_df <- dplyr::tribble(
    ~`variable-code`, ~elimination,
    "gender", "Total"
  )

  x10 <- px_elimination(x, elimination_df)
  expect_equal(px_elimination(x10), elimination_df)

  x11 <- px_elimination(x10, "T")

  expect_equal(px_elimination(x11), dplyr::tribble(
    ~`variable-code`, ~elimination,
    "place of birth", "T",
    "gender", "T",
    "time", "T"
  ))

  x11 <- px_elimination(x10, NULL)

  expect_equal(px_elimination(x11), NULL)
})

test_that("VARIABLE-TYPE is changed", {
  x <-
    "BEXSTA" |>
    get_data_path() |>
    readRDS() |>
    px()

  expect_equal(px_variable_type(x), NULL)

  variable_type_df1 <- dplyr::tibble(
    `variable-code` = "gender",
    `variable-type` = "Time"
  )

  x1 <- px_variable_type(x, variable_type_df1)

  expect_equal(px_variable_type(x1), variable_type_df1)

  x2 <- px_variable_type(x, NULL)

  expect_equal(px_variable_type(x2), NULL)

  variable_type_df2_empty <- dplyr::filter(variable_type_df1, FALSE)

  x3 <- px_variable_type(x1, variable_type_df2_empty)

  expect_equal(px_variable_type(x3), variable_type_df1)
})

test_that("CONTVARIABLE is changed", {
  x <-
    "BEXSTA" |>
    get_data_path() |>
    readRDS() |>
    px()

  x_lang <- px_languages(x, c("kl", "da"))

  expect_equal(px_contvariable(x), NULL)

  x1 <- px_contvariable(x, "gender")

  expect_equal(px_contvariable(x1), "gender")

  # Contvariable indexes some variables in table2
  expect_equal(px_units(x1), dplyr::tibble(
    code = c(NA, "K", "M", "T"),
    language = NA_character_,
    value = default_value("UNITS")
  ))

  x1_lang <- px_contvariable(x_lang, "gender")

  expect_equal(px_units(x1_lang), dplyr::tibble(
    `code` = c(
      rep(NA, 2),
      rep("K", 2),
      rep("M", 2),
      rep("T", 2)
    ),
    language = rep(c("kl", "da"), 4),
    value = default_value("UNITS")
  ))

  x2 <- px_contvariable(x1, NULL)

  expect_equal(px_contvariable(x2), NULL)

  # Setting CONTVARIABLE to NULL removes indexing in table2
  expect_equal(px_units(x2), "units")
})

test_that("stub and heading modifies acrosscells", {
  cellnote_df1 <- dplyr::tibble(
    `place of birth` = "*",
    gender = "K",
    time = "2018",
    cellnote = "This is a cellnote"
  )

  x <-
    "BEXSTA" |>
    get_data_path() |>
    readRDS() |>
    px() |>
    px_cellnote(cellnote_df1)

  x1 <-
    x |>
    px_stub("time")

  expect1 <-
    x$acrosscells |>
    dplyr::relocate(time)

  expect_identical(expect1, x1$acrosscells)

  x2 <-
    x1 |>
    px_heading(c("gender", "place of birth"))

  expect2 <-
    x$acrosscells |>
    dplyr::relocate(time, gender, `place of birth`)

  expect_identical(expect2, x2$acrosscells)
})

test_that("changing px_figures removes it from cells", {
  x <-
    population_gl |>
    dplyr::relocate(n) |>
    px() |>
    px_figures("n")

  x$cells1 |>
    dplyr::filter(`variable-code` == "n") |>
    nrow() |>
    expect_equal(0)


  x$cells2 |>
    dplyr::filter(`variable-code` == "n") |>
    nrow() |>
    expect_equal(0)
})

test_that("Error data is specified when input is not .xlsx", {
  expect_data_input_error <- function(input, data) {
    expect_error(px(input = input, data = data),
      regexp = "'data' can only be used if 'input' is an .xlsx file"
    )
  }

  expect_data_input_error(
    input = data.frame(),
    data = data.frame()
  )

  expect_data_input_error(
    input = get_px_file_path("TUX01"),
    data = data.frame()
  )
})

test_that("Error if data is not NULL, a data frame, .rds or .parquet file", {
  expect_data_error <- function(data) {
    expect_error(px(input = get_metadata_path("FOTEST"), data = data),
      regexp = "'data' has wrong format"
    )
  }

  expect_data_error(data = base::matrix())
  expect_data_error(data = list())
  expect_data_error(data = get_metadata_path("BEXSTA"))
})

test_that("Error if multiple time vars in variable", {
  expect_error(px(input = get_metadata_path("error_two_timevals")),
    regexp = "time"
  )
})

test_that("Error if timeval is not stub or heading", {
  expect_error(px(population_gl) |> px_timeval("n"),
    regexp = "TIMEVAL has to be a STUB or"
  )
})

test_that("Error if contvariable is not stub or heading", {
  expect_error(px(population_gl) |> px_contvariable("n"),
    regexp = "CONTVARIABLE has to be a STUB or"
  )
})

test_that("Error wrong input", {
  expect_input_error <- function(input, data = NULL) {
    expect_error(px(input = input, data = data),
      regexp = "'input' has wrong format"
    )
  }

  expect_input_error(input = list())
})

test_that("Error if any value contains quotation marks", {
  x <- px(input = population_gl)

  expect_quotation_error <- function(input) {
    expect_error(input,
      regexp = "quotation marks"
    )
  }

  expect_quotation_error(x |> px_language(value = 'd"a'))
  expect_quotation_error(x |> px_matrix(value = 'quotation"marks"'))
  expect_quotation_error(x |> px_contents(value = 'val " with quo'))
  expect_quotation_error(x |> px_note(value = 'val " with quo'))
  expect_quotation_error(x |> px_valuenote(value = data.frame(valuenote = '"')))
  expect_quotation_error(x |> px_cellnote(value = data.frame(cellnote = '"')))
})

test_that("px_micro arguments are validated", {
  expect_error(px_micro(x = data.frame()),
    regexp = "px object"
  )

  expect_error(px_micro(x = px(women), out_dir = 5),
    regexp = "character"
  )

  expect_error(px_micro(x = px(women), out_dir = "not/a/directory"),
    regexp = "directory does not exist"
  )
})

test_that("px with duplicates in data table cannot be saved", {
  duplicated_df <-
    dplyr::bind_rows(population_gl, population_gl)

  # px object can be created with duplicated data, which is necessary when
  # micro data is created
  x <- px(input = duplicated_df)

  # However, an error is raised when saving, becuase it will created lists
  # in figures.
  expect_error(
    x |>
      px_save(temp_px_file()),
    regexp = "contains duplicates"
  )

  expect_error(
    x |>
      px_save(temp_xlsx_file()),
    regexp = "contains duplicates"
  )
})

# Define base structures for px object
get_base_languages <- function() {
  dplyr::tibble(language = as.character())
}

get_base_table1 <- function() {
  dplyr::tibble(
    keyword = as.character(),
    value = as.character()
  )
}

get_base_table2 <- function() {
  dplyr::tibble(
    keyword = as.character(),
    code = as.character(),
    language = as.character(),
    value = as.character()
  )
}

get_base_variables1 <- function() {
  dplyr::tibble(
    `variable-code` = as.character(), # key
    pivot = as.character(),
    order = as.numeric(),
    `variable-type` = as.character(),
    contvariable = as.logical(),
    timeval = as.logical(),
    elimination = as.character()
  )
}

get_base_variables2 <- function() {
  dplyr::tibble(
    `variable-code` = as.character(), # key
    language = as.character(), # key
    `variable-label` = as.character(),
    domain = as.character(),
    map = as.character(),
    note = as.character(),
    notex = as.character()
  )
}

get_base_cells1 <- function() {
  dplyr::tibble(
    `variable-code` = as.character(), # key
    code = as.character(), # key
    order = as.numeric(),
    precision = as.numeric()
  )
}

get_base_cells2 <- function() {
  dplyr::tibble(
    `variable-code` = as.character(), # key
    code = as.character(), # key
    language = as.character(), # key
    value = as.character(),
    valuenote = as.character(),
    valuenotex = as.character()
  )
}

get_base_acrosscells <- function(stub_heading_variables = NULL) {
  data_columns <- character0_tibble(stub_heading_variables)

  dplyr::bind_cols(
    data_columns,
    dplyr::tibble(
      language = as.character(),
      cellnote = as.character(),
      cellnotex = as.character()
    )
  )
}

get_base_data <- function() {
  dplyr::tibble()
}

get_base_px <- function() {
  structure(
    list(
      languages = get_base_languages(),
      table1 = get_base_table1(),
      table2 = get_base_table2(),
      variables1 = get_base_variables1(),
      variables2 = get_base_variables2(),
      cells1 = get_base_cells1(),
      cells2 = get_base_cells2(),
      acrosscells = get_base_acrosscells(),
      data = get_base_data()
    ),
    class = "px"
  )
}

sort_in_keyword_order <- function(df) {
  df |>
    dplyr::left_join(dplyr::select(pxmake::px_keywords, "keyword", "order"),
      by = "keyword"
    ) |>
    dplyr::arrange(.data$order) |>
    dplyr::select(-"order")
}

sort_table1 <- sort_in_keyword_order

sort_table2 <- function(df, languages) {
  df |>
    dplyr::left_join(dplyr::select(pxmake::px_keywords, "keyword", "order"),
      by = "keyword"
    ) |>
    dplyr::arrange(
      .data$order, !is.na(.data$code), .data$code,
      match(.data$language, languages)
    ) |>
    dplyr::select(-"order")
}

sort_variables1 <- function(df) {
  df |>
    dplyr::arrange(desc(.data$pivot), .data$order)
}

sort_variables2 <- function(df, data_table_names, languages) {
  df |>
    dplyr::arrange(
      match(.data$`variable-code`, data_table_names),
      match(.data$language, languages)
    )
}

sort_cells1 <- function(df, data_table_names) {
  df |>
    dplyr::arrange(
      match(.data$`variable-code`, data_table_names),
      .data$order
    )
}

sort_cells2 <- function(df, data_table_names, languages) {
  df |>
    dplyr::arrange(
      match(.data$`variable-code`, data_table_names),
      .data$code,
      match(.data$language, languages)
    )
}

test_that("error - invalid px object", {
  expect_error(px_validate(1),
    regexp = "px object must be a list"
  )

  base_px <- get_base_px()
  base_px$cells1 <- NULL

  expect_error(px_validate(base_px),
    regexp = "missing these names:"
  )

  base_px2 <- get_base_px()
  base_px2$fisk <- data.frame()

  expect_error(px_validate(base_px2),
    regexp = "invalid names:"
  )

  base_px3 <- get_base_px()
  base_px3$languages <- c("en")

  expect_error(px_validate(base_px3),
    regexp = "list of data frames"
  )

  base_px4 <- get_base_px()
  base_px4$table1 <- data.frame(Keyword = as.character())

  expect_error(px_validate(base_px4),
    regexp = "px object: 'table1' is missing these columns: keyword"
  )

  base_px5 <- get_base_px()

  base_px5$table1 <- data.frame(
    keyword = "LAST-UPDATED",
    value = "2020-01-01 10:00"
  )

  expect_error(px_validate(base_px5),
    regexp = "'table1' contains misplaced keywords"
  )

  base_px6 <- get_base_px()
  base_px6$table2 <-
    data.frame(
      keyword = "NEXT-UPDATE",
      value = "2020-01-01 10:00"
    ) |>
    align_data_frames(get_base_table2())

  expect_error(px_validate(base_px6),
    regexp = "'table2' contains misplaced keywords"
  )

  base_px7 <- get_base_px()

  base_px7$languages <-
    data.frame(language = "en") |>
    align_data_frames(get_base_languages())

  base_px7$table2 <-
    dplyr::bind_rows(
      base_px7$table2,
      data.frame(keyword = "TITLE", language = "dk")
    )

  expect_error(px_validate(base_px7),
    regexp = "languages that are not defined"
  )

  bexsta <- px(get_data_path("BEXSTA"))

  bexsta1 <- bexsta
  bexsta1$variables1 <- dplyr::bind_rows(
    bexsta1$variables1,
    dplyr::tibble(`variable-code` = "fisk")
  )

  expect_error(
    px_validate(bexsta1),
    regexp = "variables1.*not in x\\$data:\\s*fisk"
  )

  bexsta2 <- bexsta
  bexsta2$variables2 <- dplyr::bind_rows(
    bexsta2$variables2,
    dplyr::tibble(
      `variable-code` = "ost",
      `variable-label` = "Ost"
    )
  )

  expect_error(
    px_validate(bexsta2),
    regexp = c("variables2.*not in x\\$data:\\s*ost")
  )

  bexsta3 <- bexsta
  bexsta3$cells1 <- dplyr::bind_rows(
    bexsta3$cells1,
    dplyr::tibble(
      `variable-code` = "sovs",
      code = "1"
    )
  )

  expect_error(
    px_validate(bexsta3),
    regexp = "cells1.*not in x\\$data:\\s*sovs"
  )

  bexsta4 <- bexsta
  bexsta4$cells2 <- dplyr::bind_rows(
    bexsta4$cells2,
    dplyr::tibble(
      `variable-code` = "ice",
      code = "1"
    )
  )

  expect_error(px_validate(bexsta4), regexp = "cells2.*not in x\\$data:\\s*ice")

  bexsta5 <- bexsta
  bexsta5$data$new_variable <- 1

  expect_error(
    px_validate(bexsta5),
    regexp = "not defined in x\\$variables1:\\s*new_variable"
  )

  bexsta5$variables1 <- dplyr::bind_rows(
    bexsta5$variables1,
    dplyr::tibble(
      `variable-code` = "new_variable",
      pivot = "STUB"
    )
  )

  expect_error(
    px_validate(bexsta5),
    regexp = "not defined in x\\$variables2:\\s*new_variable"
  )

  bexsta5$variables2 <- dplyr::bind_rows(
    bexsta5$variables2,
    dplyr::tibble(
      `variable-code` = "new_variable",
      `variable-label` = "New variable"
    )
  )

  expect_identical(px_validate(bexsta5), bexsta5)
})

#' Population Greenland
#'
#' A subset of the population count data available in Statistic Greenland's
#' BEESTA table.
#'
#' @source <https://bank.stat.gl/pxweb/en/Greenland/Greenland__BE__BE01__BE0120/BEXSTA.px/>
"population_gl"

#' Greenlanders
#'
#' A fictive data set with demographic data for Greenlanders split in two cohorts.
#'
"greenlanders"

#' px keywords
#'
#' Properties of all px keywords. Used internally by the package.
#'
#' @format A data frame:
#' \describe{
#'   \item{keyword}{Name}
#'   \item{px_function}{Function name}
#'   \item{mandatory}{Is required in a PX-file}
#'   \item{table_meta}{Is metadata for entire table; not individual variables or cells}
#'   \item{language_dependent}{Is language dependent}
#'   \item{indexed_by_contvariable}{Is indexed if CONTVARIABLE is set}
#'   \item{quote_value}{Value should be quoted in PX-file}
#'   \item{default_value}{Default value for mandatory keywords}
#'   \item{documentation}{URL to Statistic Sweden's documentation}
#'   \item{order}{Recommended order}
#' }
#'
#'
#' @source <https://www.scb.se/globalassets/vara-tjanster/px-programmen/PX-file_format_specification_2013.pdf>
"px_keywords"

#' Age classification
#'
#' Example data set to create age classification with aggreations form 10 and
#' 25 years classes.
#'
#' @format A data frame:
#' \describe{
#'   \item{valuecode}{Value code}
#'   \item{valuetext}{Value text}
#'   \item{10-years_classes}{Aggregtation into 10 years classes}
#'   \item{25-years_classes}{Aggregtation into 25 years classes}
#' }
"age_classification"

handle_cells <- function(x, value, number, column, validate) {
  if (missing(value)) {
    result <- get_cells_value(x, number, column)

    if (nrow(result) == 0) {
      return(NULL)
    } else {
      return(result)
    }
  } else if (is.null(value)) {
    x[[get_cells_name(number)]][[column]] <- NA
    return(x)
  }

  return_px(modify_cells(x, number, column, value), validate)
}

#' @rdname px_precision.px
#' @export
px_precision <- function(x, value, validate) {
  UseMethod("px_precision")
}

#' @eval add_doc_keyword_function_intro("PRECISION")
#' @param value `r cells_param_value("PRECISION", "1")`
#' @eval param_validate()
#' @eval add_return_px_or_df()
#' @eval add_cells1_example("PRECISION", 2, 3)
#' @export
px_precision.px <- function(x, value, validate = TRUE) {
  handle_cells(x, value, "1", "precision", validate)
}


#' @rdname px_order.px
#' @export
px_order <- function(x, value, validate) {
  UseMethod("px_order")
}

#' @title Change value order
#' @eval add_doc_keyword_function_intro("ORDER")
#' @param value `r cells_param_value("ORDER", "1")`
#' @eval param_validate()
#' @eval add_return_px_or_df()
#' @eval add_cells1_example("ORDER", 8, 9)
#' @export
px_order.px <- function(x, value, validate = TRUE) {
  handle_cells(x, value, "1", "order", validate)
}

test_that("cells1 is modified", {
  x <-
    "BEXSTA" |>
    get_data_path() |>
    readRDS() |>
    px()

  expect_identical(px_precision(x), NULL)

  precision_df1 <- dplyr::tibble(
    `variable-code` = "gender",
    code = "K",
    precision = 1
  )

  x2 <- px_precision(x, precision_df1)
  expect_identical(px_precision(x2), precision_df1)

  precision_df2 <- dplyr::tibble(
    `variable-code` = "gender",
    precision = 2
  )

  x3 <- px_precision(x, precision_df2)

  precision_df2_expect <- tidyr::crossing(precision_df2,
    code = c("K", "M", "T")
  ) |>
    dplyr::relocate(precision, .after = last_col())

  expect_identical(px_precision(x3), precision_df2_expect)

  x4 <- px_precision(x, NULL)
  expect_identical(px_precision(x4), NULL)


  reverse_order <-
    px_precision(x, data.frame(precision = 1)) |>
    px_precision() |>
    dplyr::mutate(order = as.numeric(rev(dplyr::row_number()))) |>
    dplyr::select(-precision) |>
    dplyr::arrange(
      match(.data$`variable-code`, names(px_data(x))),
      order
    )

  x5 <- px_order(x, reverse_order)

  expect_identical(px_order(x5), reverse_order)

  x6 <- px_order(x, NULL)
  expect_identical(px_order(x6), NULL)
})


test_that("cells1 are properly created for multilingual file without CODES", {
  x <- px(get_px_file_path("multilingual_no_codes"))

  expect_identical(
    x$cells1,
    tibble::tribble(
      ~`variable-code`, ~code, ~order, ~precision,
      "type", "a", 1L, NA_real_,
      "type", "b", 2L, NA_real_,
      "type", "c", 3L, NA_real_,
      "type", "d", 4L, NA_real_,
      "gender", "F", 1L, NA_real_,
      "gender", "M", 2L, NA_real_
    )
  )
})

handle_acrossnote_keyword <- function(x, value, keyword, na_to_star, validate) {
  if (missing(value)) {
    return(get_acrosscells_value(x, keyword))
  } else if (is.null(value)) {
    x$acrosscells <- dplyr::filter(x$acrosscells, FALSE)
  } else if (nrow(value) == 0) {
    return(x)
  } else {
    x <- modify_acrosscells(x, value, keyword, na_to_star)
  }

  return_px(x, validate)
}

#' @rdname px_cellnote.px
#' @export
px_cellnote <- function(x, value, na_to_star, validate) {
  UseMethod("px_cellnote")
}

#' @eval add_documentation_acrosscells("CELLNOTE")
px_cellnote.px <- function(x, value, na_to_star = TRUE, validate = TRUE) {
  handle_acrossnote_keyword(x, value, "CELLNOTE", na_to_star, validate)
}

#' @rdname px_cellnotex.px
#' @export
px_cellnotex <- function(x, value, na_to_star, validate) {
  UseMethod("px_cellnotex")
}

#' @eval add_documentation_acrosscells("CELLNOTEX")
px_cellnotex.px <- function(x, value, na_to_star = TRUE, validate = TRUE) {
  handle_acrossnote_keyword(x, value, "CELLNOTEX", na_to_star, validate)
}

test_that("cellnote is modified and removed", {
  x <-
    "BEXSTA" |>
    get_data_path() |>
    readRDS() |>
    px()

  expect_identical(px_cellnote(x), NULL)

  cellnote_df1 <- dplyr::tibble(
    `place of birth` = "*",
    gender = "K",
    time = "2018",
    cellnote = "This is a cellnote"
  )

  x1 <- px_cellnote(x, cellnote_df1)
  expect_identical(px_cellnote(x1), cellnote_df1)

  # Stars * are added for variables that aren't supplied
  cellnote_df2 <- dplyr::select(cellnote_df1, -`place of birth`)
  x2 <- px_cellnote(x, cellnote_df2)
  expect_identical(px_cellnote(x2), cellnote_df1)

  # Adding stars can be turned off (used for px_micro)
  cellnote_df2 <- dplyr::select(cellnote_df1, -`place of birth`)
  x3 <- px_cellnote(x, cellnote_df2, na_to_star = FALSE)

  cellnote_df_no_star <-
    cellnote_df1 |>
    dplyr::mutate(`place of birth` = NA_character_)

  expect_identical(px_cellnote(x3), cellnote_df_no_star)


  # Changing languages adds cellnote for each language
  language_list <- c("da", "en", "kl")

  x1_lang <- px_languages(x1, language_list)

  cellnote_df2_lang <-
    tidyr::crossing(cellnote_df1, language = language_list) |>
    dplyr::relocate(language, .before = "cellnote")

  expect_identical(px_cellnote(x1_lang), cellnote_df2_lang)

  # Set cellnote for multiple languages
  x2_lang <-
    x |>
    px_languages(language_list) |>
    px_cellnote(cellnote_df2_lang)

  expect_identical(px_cellnote(x2_lang), cellnote_df2_lang)

  # error if column is not in the data
  cellnote_error1 <- dplyr::tibble(
    not_a_column = "fisk",
    `place of birth` = "*",
    gender = "K",
    time = "2018",
    cellnote = "This is a cellnote"
  )

  expect_error(px_cellnote(x, cellnote_error1), regexp = "invalid column")

  # cellnotex
  expect_identical(px_cellnotex(x), NULL)

  cellnotex_df1 <- cellnote_df1 |> dplyr::rename(cellnotex = cellnote)

  x4 <- px_cellnote(x, dplyr::filter(cellnote_df1, FALSE))

  expect_identical(
    px_cellnote(x4), px_cellnote(x)
  ) # cellnote which no rows should not modify

  x5 <- px_cellnotex(x, cellnotex_df1)

  expect_identical(px_cellnotex(x5), cellnotex_df1)

  x6 <- px_cellnotex(x, NULL)

  expect_identical(px_cellnotex(x6), NULL)
})

#' Return data table from px object
#'
#' @inheritParams px_data
#'
#' @keywords internal
get_data_table <- function(x, value, labels, sort) {
  if (isTRUE(sort)) {
    data_table <- sort_data_table_by_order(x)
  } else {
    data_table <- x$data
  }

  if (isFALSE(labels)) {
    return(data_table)
  } else if (isTRUE(labels)) {
    m_language <- defined_languages(x)[1]
  } else {
    m_language <- labels
  }

  error_if_language_is_undefined(m_language, x)

  data_with_labels <-
    data_table |>
    dplyr::mutate(id_ = dplyr::row_number()) |>
    tidyr::pivot_longer(-c(px_figures(x), "id_"),
      names_to = "variable-code",
      values_to = "code"
    ) |>
    dplyr::left_join(px_values(x),
      by = c("variable-code", "code"),
      relationship = "many-to-many"
    ) |>
    (\(.) {
      if (!"language" %in% names(.)) {
        dplyr::mutate(., language = m_language) # add dummy language
      } else {
        .
      }
    })() |>
    dplyr::mutate(
      language = ifelse(is.na(.data$language),
        m_language,
        .data$language
      ),
      value = ifelse(is.na(.data$value),
        .data$code,
        .data$value
      )
    ) |>
    dplyr::filter(.data$language %in% m_language) |>
    dplyr::select(-"code", -"language") |>
    tidyr::pivot_wider(
      names_from = "variable-code",
      values_from = "value"
    ) |>
    dplyr::select(-"id_") |>
    dplyr::relocate(names(data_table))

  data_with_labels
}

#' Add value as data table in x
#'
#' @inheritParams px_data
#'
#' @keywords internal
update_data_table <- function(x, value) {
  old_df_columns <- names(px_data(x))
  df_columns <- names(value)
  new_columns <- setdiff(df_columns, old_df_columns)

  dummy_px <-
    px_from_data_df(value) |>
    px_languages(defined_languages(x))

  if (!is.null(px_elimination(x))) {
    elimination_values <-
      px_values(x) |>
      dplyr::semi_join(px_elimination(x),
        by = c(
          "variable-code" = "variable-code",
          "code" = "elimination"
        )
      )

    elimination_order <-
      px_order(x) |>
      dplyr::semi_join(px_elimination(x),
        by = c(
          "variable-code" = "variable-code",
          "code" = "elimination"
        )
      )

    dummy_px <-
      dummy_px |>
      px_values(elimination_values) |>
      px_order(elimination_order)
  }

  # Add new variables from x2 to x1, and remove variables from x1 that are
  # not in x2.
  swap_in_variables_metadata <- function(x1, x2, element) {
    x1[[element]] <-
      x1[[element]] |>
      dplyr::filter(.data$`variable-code` %in% df_columns) |>
      dplyr::bind_rows(dplyr::filter(
        x2[[element]],
        .data$`variable-code` %in% new_columns
      ))
    x1
  }

  x <- swap_in_variables_metadata(x, dummy_px, element = "variables1")
  x <- swap_in_variables_metadata(x, dummy_px, element = "variables2")

  # Add new codes from x2 not in x1, and remove codes from x1 that are not
  # in x2.
  swap_in_cells_metadata <- function(x1, x2, element) {
    join_vars <- c("variable-code", "code")

    if (element == "cells2") {
      join_vars <- c(join_vars, "language")
    }

    new_codes <-
      x2[[element]] |>
      dplyr::anti_join(x1[[element]], by = join_vars)

    x1[[element]] <-
      x1[[element]] |>
      dplyr::semi_join(x2[[element]], by = join_vars) |>
      dplyr::bind_rows(new_codes)

    x1
  }

  x <- swap_in_cells_metadata(x, dummy_px, element = "cells1")
  x <- swap_in_cells_metadata(x, dummy_px, element = "cells2")

  x$acrosscells <- get_base_acrosscells(c(px_stub(x), px_heading(x)))

  x$cells1 <- x$cells1 |> dplyr::filter(.data$`variable-code` != px_figures(x))
  x$cells2 <- x$cells2 |> dplyr::filter(.data$`variable-code` != px_figures(x))

  x$data <- value

  x
}

#' Sort data table by order
#'
#' @keywords internal
sort_data_table_by_order <- function(x) {
  data_table <- x$data
  order_df <- px_order(x)

  columns_to_sort <- intersect(
    names(data_table),
    dplyr::pull(order_df, .data$`variable-code`)
  )

  for (column in rev(columns_to_sort)) {
    tmp <-
      order_df |>
      dplyr::filter(.data$`variable-code` == column) |>
      tidyr::pivot_wider(names_from = "variable-code", values_from = "code")

    data_table <-
      data_table |>
      dplyr::left_join(tmp, by = column) |>
      dplyr::arrange(order) |>
      dplyr::select(-order)
  }

  data_table
}

#' @rdname px_data.px
#' @export
px_data <- function(x, value, labels, sort, validate) {
  UseMethod("px_data")
}

#' @eval add_doc_keyword_function_intro("DATA")
#' @param value Optional. A data frame. If missing, the current
#' DATA is returned. If NULL, all data rows are removed.
#' @eval add_return_px_or_df()
#' @param labels Optional. Logic or character vector. If TRUE, the data table
#' is returned with VALUES instead of CODES. By default the VALUES of the main
#' language are returned, use a character language code to return VALUES for a
#' specific language.
#' @param sort Optional. If TRUE, the data table is returned in the sort order
#' defined by [px_order()]. If FALSE, the data table is returned as is.
#' @eval param_validate()
#'
#' @details If adding a new data frame, metadata is generated for the new
#' columns and removed for columns that are no longer present.
#'
#' @examples
#' x1 <- px(population_gl)
#'
#' # Print data table
#' px_data(x1)
#'
#' # Change data table
#' population_gl_2024 <- subset(population_gl, year == 2024)
#'
#' x2 <- px_data(x1, population_gl_2024)
#'
#' # Return data table with VALUES instead of CODES
#' px_data(x1, labels = TRUE)
#'
#' # Return VALUES for a specific language
#' x_mult <-
#'   x1 |>
#'   px_languages(c("en", "gl"))
#'
#' px_data(x_mult, labels = "gl")
#'
#' @export
px_data.px <- function(
  x, value, labels = FALSE, sort = FALSE, validate = TRUE
) {
  validate_px_data_arguments(x, value, labels, sort, validate)

  if (missing(value)) {
    return(get_data_table(x = x, value = value, labels = labels, sort = sort))
  } else if (is.null(value)) {
    x$data <- dplyr::filter(x$data, FALSE)
  } else {
    x <- update_data_table(x = x, value = value)
  }

  return_px(x, validate)
}

#' Regular expression to parse header in PX-file
#'
#' @returns A character vector
#' @keywords internal
get_px_metadata_regex <- function() {
  paste0(
    "(?<keyword>[[:upper:]-]+)", # Leading keyword
    "(?:\\[)?", # Maybe opening language bracket [
    "(?<language>[[:alpha:]_-]+)?", # Maybe language
    "(?:\\])?", # Maybe closing language bracket ]
    "(?:\\(\")?", # Maybe opening sub-key parentheses (
    "(?<variable>[^\"=]+)?", # Maybe sub-key
    "(?:\",\")?", # Maybe comma before cell value
    "(?<cell>[^=]+?)?(?=\\)=|=)", # Maybe cell value
    "(?:\")?", # Maybe closing " after cell value
    "(?:\"\\))?", # Maybe closing sub-key parentheses )
    "=", # definitely =
    '(?<value>(?:"[^"]*"|[^;])*)', # Val is quoted string or non-semicolon chars
    "(?:;$)?" # Maybe ;
  )
}

#' Get metadata df from px lines
#'
#' @param metadata_lines A character vector with the header of a PX-file.
#'
#' @returns A data frame
#' @keywords internal
get_metadata_df_from_px_lines <- function(metadata_lines) {
  acrosscells_keywords <- c("CELLNOTE", "CELLNOTEX")

  metadata_lines |>
    # Remove newlines in file. Use semi-colon as line separator
    paste0(collapse = "") |>
    # Split by semicolons not inside a pair of quotes ""
    stringr::str_split(';(?=(?:[^"]*"[^"]*")*[^"]*$)') |>
    unlist() |>
    stringr::str_match(get_px_metadata_regex()) |>
    dplyr::as_tibble(
      .name_repair = ~ vctrs::vec_as_names(...,
        repair = "unique",
        quiet = TRUE
      )
    ) |>
    dplyr::select(-starts_with("...")) |>
    dplyr::mutate(
      cell = stringr::str_replace_all(.data$cell, '"\\)', ""),
      cell = dplyr::na_if(.data$cell, "")
    ) |>
    # remove leading and trailing " on all keywords except TIMEVAL
    dplyr::mutate(value = ifelse(.data$keyword != "TIMEVAL",
      stringr::str_replace_all(.data$value, '^"|"$', ""),
      .data$value
    )) |>
    # Variables indexed by CONTVARIABLE are cell values not variable
    dplyr::mutate(
      cell = ifelse(.data$keyword %in% keywords_indexed_by_contvariable(),
        .data$variable,
        .data$cell
      ),
      variable = ifelse(.data$keyword %in% keywords_indexed_by_contvariable(),
        NA,
        .data$variable
      )
    ) |>
    dplyr::mutate(
      variable = dplyr::if_else(.data$keyword %in% acrosscells_keywords,
        stringr::str_glue('"{variable}","{cell}"'),
        .data$variable
      ),
      cell = ifelse(.data$keyword %in% acrosscells_keywords,
        NA,
        .data$cell
      )
    ) |>
    # remove double quotes caused by collapsing values spanning multiple lines
    dplyr::mutate(value = stringr::str_replace_all(.data$value, '""', "")) |>
    dplyr::mutate(value = ifelse(.data$keyword != "TIMEVAL",
      stringr::str_split(.data$value, '","'),
      .data$value
    )) |>
    dplyr::filter(.data$keyword != "DATA")
}

#' Get variable code and label
#'
#' Get a data frame with variable codes and label in all languages.
#' If VARIABLECODE is used, the relation is defined there, but otherwise the
#' variable code is set to be the main languages' label.
#'
#' @inheritParams sort_metadata_df
#'
#' @returns A data frame
#' @keywords internal
get_variable_label <- function(metadata_df) {
  metadata_df <- add_main_language(metadata_df)

  head_stub <-
    metadata_df |>
    dplyr::filter(.data$keyword %in% c("HEADING", "STUB")) |>
    tidyr::unnest("value") |>
    dplyr::rename("variable-label" = "value") |>
    dplyr::group_by(.data$keyword, .data$language) |>
    dplyr::mutate(index = dplyr::row_number()) |>
    dplyr::ungroup() |>
    dplyr::select(
      "keyword", "language", "main_language", "variable-label", "index"
    )

  head_stub_main_language <-
    head_stub |>
    dplyr::filter(.data$main_language) |>
    dplyr::rename("main_language_label" = "variable-label") |>
    dplyr::select("keyword", "index", "main_language_label")

  metadata_df |>
    dplyr::filter(.data$keyword == "VARIABLECODE") |>
    tidyr::unnest("value") |>
    dplyr::select(
      "variable-code" = "value",
      "variable-label" = "variable",
      "language",
      "main_language"
    ) |>
    dplyr::full_join(head_stub,
      by = c("variable-label", "language", "main_language")
    ) |>
    dplyr::left_join(head_stub_main_language, by = c("keyword", "index")) |>
    dplyr::mutate(`variable-code` = ifelse(is.na(.data$`variable-code`),
      .data$main_language_label,
      .data$`variable-code`
    )) |>
    dplyr::select(-"main_language_label")
}

#' Add boolean main_language column
#'
#' @inheritParams sort_metadata_df
#'
#' @returns A data frame
#' @keywords internal
add_main_language <- function(metadata_df) {
  m_language <- get_main_language(metadata_df)

  if (is.na(m_language)) {
    metadata_df |>
      dplyr::mutate(main_language = is.na(m_language))
  } else {
    metadata_df |>
      replace_na_language_with_main_language() |>
      dplyr::mutate(main_language = .data$language == m_language)
  }
}

#' Get the main language from metadata
#'
#' @inheritParams sort_metadata_df
#'
#' @returns Character
#' @keywords internal
get_main_language <- function(metadata_df) {
  main_language <-
    metadata_df |>
    dplyr::filter(.data$keyword %in% c("LANGUAGE", "LANGUAGES")) |>
    dplyr::arrange(.data$keyword) |>
    tidyr::unnest("value") |>
    dplyr::slice(1) |>
    dplyr::pull(.data$value)

  if (length(main_language) == 0) {
    main_language <- NA
  }

  main_language
}

#' Impute missing language
#'
#' Add main language as language if language is missing.
#'
#' @inheritParams sort_metadata_df
#'
#' @returns A data frame
#' @keywords internal
replace_na_language_with_main_language <- function(metadata_df) {
  metadata_df |>
    dplyr::mutate(language = tidyr::replace_na(
      .data$language,
      get_main_language(metadata_df)
    ))
}

#' Sort metadata data frame
#'
#' The data frame is first sorted by the keyword order defined in the
#' px-specification and then by the language order.
#'
#' @param metadata_df Data frame with metadata.
#'
#' @returns A data frame
#' @keywords internal
sort_metadata_df <- function(metadata_df) {
  languages <-
    metadata_df |>
    dplyr::filter(.data$keyword == "LANGUAGES") |>
    tidyr::unnest("value") |>
    dplyr::mutate(language_order = dplyr::row_number()) |>
    dplyr::select("language" = "value", "language_order")

  stub_heading <-
    metadata_df |>
    dplyr::filter(
      .data$keyword %in% c("STUB", "HEADING"),
      .data$language %in% get_main_language(metadata_df)
    ) |>
    dplyr::arrange(desc(.data$keyword)) |>
    tidyr::unnest("value") |>
    dplyr::mutate(stub_heading_order = dplyr::row_number()) |>
    dplyr::select("variable" = "value", "stub_heading_order")

  metadata_df |>
    dplyr::left_join(pxmake::px_keywords |> dplyr::select("keyword", "order"),
      by = "keyword"
    ) |>
    dplyr::left_join(languages, by = "language") |>
    dplyr::left_join(stub_heading, by = "variable") |>
    dplyr::arrange(
      .data$order,
      .data$keyword,
      .data$stub_heading_order,
      .data$language_order,
      !is.na(.data$variable),
      .data$variable,
      .data$cell
    ) |>
    dplyr::select(-"order", -"language_order", -"stub_heading_order")
}

#' Get metadata df from px object
#'
#' @param x A px object
#'
#' @returns A data frame
#' @keywords internal
get_metadata_df_from_px <- function(x) {
  metadata_template <- dplyr::tibble(
    keyword = character(),
    language = character(),
    variable = character(),
    cell = character(),
    value = list(character())
  )

  languages <-
    x$languages |>
    dplyr::summarise(value = list(paste0(.data$language, sep = ""))) |>
    dplyr::mutate(keyword = "LANGUAGES") |>
    dplyr::filter(.data$value != "")

  table <-
    x$table1 |>
    tidyr::drop_na("value") |>
    wrap_varaible_in_list("value")

  table_language_dependent <-
    x$table2 |>
    dplyr::rename("variable" = "code") |>
    tidyr::drop_na("value") |>
    wrap_varaible_in_list("value")

  variablecode <-
    x$variables2 |>
    dplyr::distinct(
      "keyword" = "VARIABLECODE",
      .data$language,
      "variable" = .data$`variable-label`,
      "value" = .data$`variable-code`
    ) |>
    tidyr::drop_na("value") |>
    wrap_varaible_in_list("value")

  note_etc <-
    x$variables2 |>
    tidyr::pivot_longer(
      cols = c(-"variable-code", -"language", -"variable-label"),
      names_to = "keyword",
      values_to = "value"
    ) |>
    dplyr::mutate(keyword = toupper(.data$keyword)) |>
    dplyr::select(
      "keyword",
      "variable" = "variable-label", "language", "value"
    ) |>
    dplyr::arrange_all() |>
    tidyr::drop_na("value") |>
    wrap_varaible_in_list("value")

  name_relation <-
    dplyr::select(x$variables2, "variable-code", "language", "variable-label")

  variables1 <-
    x$variables1 |>
    dplyr::left_join(name_relation, by = "variable-code")

  stub_heading_variables <-
    variables1 |>
    sort_variables1() |>
    dplyr::filter(toupper(.data$pivot) %in% c("STUB", "HEADING")) |>
    dplyr::distinct(.data$`variable-code`) |>
    dplyr::pull("variable-code")

  head_stub <-
    variables1 |>
    dplyr::filter(.data$`variable-code` %in% stub_heading_variables) |>
    dplyr::mutate(keyword = toupper(.data$pivot)) |>
    dplyr::arrange(.data$keyword, .data$order) |>
    dplyr::group_by(.data$keyword, .data$language) |>
    dplyr::summarise(
      value = list(paste(.data$`variable-label`, sep = ", ")),
      .groups = "keep"
    )

  variable_type <-
    variables1 |>
    tidyr::drop_na("variable-type") |>
    tidyr::pivot_longer(
      cols = "variable-type",
      names_to = "keyword",
      values_to = "value"
    ) |>
    dplyr::mutate(keyword = toupper(.data$keyword)) |>
    dplyr::select(
      "keyword",
      "variable" = "variable-label", "language", "value"
    ) |>
    wrap_varaible_in_list("value")

  contvariable <-
    variables1 |>
    dplyr::filter(.data$contvariable) |>
    dplyr::mutate(keyword = "CONTVARIABLE") |>
    dplyr::select("keyword", "language", "value" = "variable-label") |>
    wrap_varaible_in_list("value")

  time_metadata <-
    variables1 |>
    dplyr::filter(.data$timeval)

  time_variable <-
    time_metadata |>
    dplyr::distinct(.data$`variable-code`) |>
    dplyr::pull("variable-code")

  if (length(time_variable) == 0) {
    timeval <- NULL
  } else {
    error_if_more_than_one_time_variable(time_variable)

    time_values <-
      x$data |>
      dplyr::distinct(across(all_of(time_variable))) |>
      dplyr::pull(1)

    timeval <-
      time_metadata |>
      dplyr::mutate(
        keyword = "TIMEVAL",
        value = format_time_values(time_values)
      ) |>
      dplyr::select(
        "keyword", "language",
        "variable" = "variable-label", "value"
      ) |>
      wrap_varaible_in_list("value")
  }

  codes_not_in_cells <-
    x$data |>
    dplyr::select(
      dplyr::all_of(intersect(stub_heading_variables, names(x$data)))
    ) |>
    tidyr::pivot_longer(
      cols = everything(),
      names_to = "variable-code",
      values_to = "code"
    ) |>
    dplyr::distinct_all() |>
    align_data_frames(get_base_cells2()) |>
    dplyr::select("variable-code", "code") |>
    dplyr::anti_join(x$cells2, by = c("variable-code", "code")) |>
    dplyr::mutate(value = .data$code) |>
    tidyr::crossing(language = defined_languages(x))

  cells <-
    x$cells2 |>
    dplyr::bind_rows(codes_not_in_cells) |>
    dplyr::left_join(x$cells1, by = c("variable-code", "code")) |>
    dplyr::left_join(name_relation, by = c("variable-code", "language")) |>
    dplyr::mutate(value = ifelse(is.na(.data$value), .data$code, .data$value))

  code_value <-
    cells |>
    tidyr::pivot_longer(cols = c("code", "value"), names_to = "type") |>
    dplyr::mutate(keyword = toupper(paste0(.data$type, "s"))) |>
    dplyr::arrange(.data$keyword, .data$order) |>
    dplyr::rename("variable" = "variable-label") |>
    dplyr::group_by(.data$keyword, .data$language, .data$variable) |>
    dplyr::summarise(
      value = list(paste(.data$value, sep = ", ")), .groups = "keep"
    ) |>
    dplyr::ungroup()

  elimination <-
    variables1 |>
    tidyr::drop_na("elimination") |>
    dplyr::rename(code = elimination) |>
    dplyr::left_join(
      cells |> dplyr::select("variable-code", "code", "language", "value"),
      by = c("variable-code", "code", "language")
    ) |>
    dplyr::mutate(
      value = ifelse(is.na(.data$value), .data$code, .data$value)
    ) |>
    dplyr::mutate(keyword = "ELIMINATION") |>
    dplyr::select(
      "keyword", "language",
      "variable" = "variable-label", "value"
    ) |>
    wrap_varaible_in_list("value")

  valuenotes <-
    cells |>
    tidyr::pivot_longer(
      cols = starts_with("valuenote"),
      names_to = "keyword",
      values_to = "keyword_value"
    ) |>
    dplyr::mutate(keyword = toupper(.data$keyword)) |>
    tidyr::drop_na("keyword_value") |>
    wrap_varaible_in_list("keyword_value") |>
    dplyr::select("keyword",
      "language",
      "variable" = "variable-label",
      "cell" = "value",
      "value" = "keyword_value"
    )

  precision <-
    cells |>
    dplyr::mutate(keyword = "PRECISION") |>
    tidyr::drop_na("precision") |>
    wrap_varaible_in_list("precision") |>
    dplyr::select("keyword",
      "language",
      "variable" = "variable-label",
      "cell" = "value",
      "value" = "precision"
    )

  acrosscells <-
    x$acrosscells |>
    dplyr::mutate(
      across(c(px_stub(x), px_heading(x)), ~ ifelse(is.na(.), "*", .))
    ) |>
    tidyr::pivot_longer(
      cols = setdiff(get_base_acrosscells() |> names(), "language"),
      names_to = "keyword",
      values_to = "value"
    ) |>
    dplyr::mutate(keyword = toupper(.data$keyword)) |>
    tidyr::drop_na("value") |>
    tidyr::unite("variable", all_of(stub_heading_variables), sep = '","') |>
    wrap_varaible_in_list("value")

  metadata_df <-
    dplyr::bind_rows(
      metadata_template,
      languages,
      table,
      table_language_dependent,
      contvariable,
      variablecode,
      note_etc,
      head_stub,
      variable_type,
      timeval,
      elimination,
      code_value,
      valuenotes,
      precision,
      acrosscells
    ) |>
    replace_na_language_with_main_language() |>
    sort_metadata_df()

  metadata_df
}

# Functions to create roxygen2 documentation

keyword_to_function <- function(keyword) {
  paste0("px_", gsub("-", "_", tolower(keyword)))
}

function_to_keyword <- function(fnc_name) {
  toupper(substr(gsub("_", "-", fnc_name), 4, nchar(fnc_name)))
}

split_multiline_str_into_vector <- function(str) {
  str |>
    stringr::str_split(pattern = "\n") |>
    magrittr::extract2(1)
}

add_documentation_function <- function(fnc) {
  function(...) {
    fnc(...) |>
      split_multiline_str_into_vector()
  }
}

add_doc_keyword_function_intro <-
  add_documentation_function(doc_keyword_function_intro)

add_documentation_table1 <- function(keyword, example_value) {
  stringr::str_glue(
    doc_keyword_function_intro(keyword),
    "@param value {table1_param_value(keyword)}",
    param_validate(),
    return_px_or_char_str(),
    table1_example(keyword, example_value),
    "@export",
    .sep = "\n"
  ) |>
    split_multiline_str_into_vector()
}

add_documentation_table2 <- function(keyword, example_value1, example_value2) {
  stringr::str_glue(
    doc_keyword_function_intro(keyword),
    "@param value {table2_param_value(keyword)}",
    param_validate(),
    return_px_or_char_vector_or_df(),
    table2_example(keyword, example_value1, example_value2),
    "@export",
    .sep = "\n"
  ) |>
    split_multiline_str_into_vector()
}

add_documentation_head_stub <- function(keyword) {
  stringr::str_glue(
    doc_keyword_function_intro(keyword),
    "@param value {pivot_param_value(keyword)}",
    param_validate(),
    return_px_or_char_vector(),
    "@export",
    .sep = "\n"
  ) |>
    split_multiline_str_into_vector()
}

add_documentation_variables2 <- function(
  keyword, example_value1, example_value2, example_value3
) {
  stringr::str_glue(
    doc_keyword_function_intro(keyword),
    "@param value {variables2_param_value(keyword)}",
    param_validate(),
    return_px_or_char_vector_or_df(),
    variables2_example(keyword, example_value1, example_value2, example_value3),
    "@export",
    .sep = "\n"
  ) |>
    split_multiline_str_into_vector()
}

add_documentation_acrosscells <- function(keyword) {
  stringr::str_glue(
    doc_keyword_function_intro(keyword),
    "@param value {acrosscells_param_value(keyword)}",
    "@param na_to_star Optional. Convert all NAs to '*'.",
    param_validate(),
    return_px_or_df(),
    acrosscells_example(keyword),
    "@export",
    .sep = "\n"
  ) |>
    split_multiline_str_into_vector()
}

add_documentation_note <- function(keyword) {
  stringr::str_glue(
    doc_keyword_function_intro(keyword),
    "@param value {note_param_value(keyword)}",
    "@details {keyword} has a lot of possible ways to specify `value`, ",
    "because it can be set both for the entire PX-file and for ",
    "individual variables.",
    param_validate(),
    return_px_char_df_or_list(),
    note_example(keyword),
    "@export",
    .sep = "\n"
  ) |>
    split_multiline_str_into_vector()
}

doc_keyword_function_intro <- function(keyword) {
  str <-
    stringr::str_glue(
      "@title {keyword}",
      "@description Inspect or change {keyword}.",
      "@param x A px object",
      .sep = "\n"
    )

  url <-
    pxmake::px_keywords |>
    dplyr::filter(keyword == !!keyword) |>
    dplyr::pull(.data$documentation)

  if (!any(identical(url, character(0)), identical(url, ""))) {
    str <-
      stringr::str_glue("{str}",
        "@seealso [Statistics Sweden's documentation]({url})",
        .sep = "\n"
      )
  }

  str
}

description_start <- function(keyword) {
  stringr::str_glue("Inspect or change {keyword}.")
}

table1_example <- function(keyword, example_value) {
  px_function <- keyword_to_function(keyword)

  str <-
    stringr::str_glue(
      "@examples",
      "# Set {keyword}",
      "x1 <-",
      "   px(population_gl) |>",
      "   {px_function}('{example_value}')",
      "",
      "# Print {keyword}",
      "{px_function}(x1)",
      .sep = "\n"
    )

  if (!keyword %in% mandatory_keywords()) {
    str <-
      stringr::str_glue(
        "{str}",
        "",
        "# Remove {keyword}",
        "x2 <- {px_function}(x1, NULL)",
        "{px_function}(x2)",
        .sep = "\n"
      )
  }

  str
}

add_table1_example <- add_documentation_function(table1_example)

table2_example <- function(keyword, example_value1, example_value2) {
  px_function <- keyword_to_function(keyword)

  str <-
    stringr::str_glue(
      "@examples",
      "# Set {keyword} for all languages",
      "x1 <-",
      "  px(population_gl) |>",
      "  {px_function}('{example_value1}')",
      "",
      "# Print {keyword}",
      "{px_function}(x1)",
      "",
      "# Set {keyword} for individual languages",
      "library(tibble)",
      "x2 <-",
      "  x1 |>",
      "  px_languages(c('en', 'kl')) |>",
      "  {px_function}(tribble(~language, ~value,",
      "                      'en', '{example_value1}',",
      "                      'kl', '{example_value2}'))",
      "{px_function}(x2)",
      .sep = "\n"
    )

  if (!keyword %in% c(mandatory_keywords(), "TITLE")) {
    str <-
      stringr::str_glue(
        "{str}",
        "",
        "# Remove {keyword}",
        "x3 <- {px_function}(x2, NULL)",
        "{px_function}(x3)",
        .sep = "\n"
      )
  }

  str
}

variables2_example <- function(
  keyword, example_value1, example_value2, example_value3
) {
  if (is.na(example_value1)) {
    return("")
  }

  px_function <- keyword_to_function(keyword)

  stringr::str_glue(
    "@examples",
    "# Set {keyword} for all languages",
    "x1 <-",
    "  px(population_gl) |>",
    "  {px_function}('{example_value1}')",
    "",
    "# Print {keyword}",
    "{px_function}(x1)",
    "",
    "# Set {keyword} for individual variables",
    "library(tibble)",
    "x2 <-",
    "  x1 |>",
    "  {px_function}(tribble(~`variable-code`, ~{tolower(keyword)},",
    "                    'gender', '{example_value2}',",
    "                    'age',    '{example_value3}'))",
    "{px_function}(x2)",
    "",
    "# Set {keyword} for individual languages",
    "x3 <-",
    "  x2 |>",
    "  px_languages(c('en', 'kl')) |>",
    "  {px_function}(tribble(~`variable-code`, ~language, ~{tolower(keyword)},",
    "                    'gender',    'en',      '{example_value2}_en',",
    "                    'gender',    'kl',      '{example_value2}_kl',",
    "                    'age',       'en',      '{example_value3}_en'))",
    "{px_function}(x3)",
    "",
    "# Remove {keyword}",
    "x4 <- {px_function}(x3, NULL)",
    "{px_function}(x4)",
    .sep = "\n"
  )
}

cells1_example <- function(keyword, example_value1, example_value2) {
  px_function <- keyword_to_function(keyword)

  stringr::str_glue(
    "@examples",
    "# Set {keyword} for a variable",
    "library(tibble)",
    "x1 <-",
    "  population_gl |>",
    "  px() |>",
    "  {px_function}(tribble(~`variable-code`, ~{tolower(keyword)},",
    "                       'gender', {example_value1}))",
    "",
    "# Print {keyword}",
    "{px_function}(x1)",
    "",
    "# Set {keyword} for a value",
    "x2 <-",
    "  x1 |>",
    "  {px_function}(tribble(~`variable-code`, ~code, ~{tolower(keyword)},",
    "                       'age', '2004', {example_value2}))",
    "{px_function}(x2)",
    "",
    "# Remove {keyword}",
    "x3 <- {px_function}(x2, NULL)",
    "{px_function}(x3)",
    .sep = "\n"
  )
}

add_cells1_example <- add_documentation_function(cells1_example)

cells2_example <- function(
  keyword, example_value1, example_value2, example_value3
) {
  px_function <- keyword_to_function(keyword)

  stringr::str_glue(
    "@examples",
    "# Set {keyword} for a value",
    "library(tibble)",
    "x1 <-",
    "  population_gl |>",
    "  px() |>",
    "  {px_function}(",
    "    tribble(~`variable-code`, ~code,  ~{tolower(keyword)},",
    "            'year', '2004', '{example_value1}'))",
    "",
    "# Print {keyword}",
    "{px_function}(x1)",
    "",
    "# Set {keyword} for a value in specific language",
    "x2 <-",
    "  x1 |>",
    "  px_languages(c('en', 'kl')) |>",
    "  {px_function}(",
    "    tribble(~`variable-code`, ~code,  ~language, ~{tolower(keyword)},",
    "            'age', '0-6', 'en', '{example_value2}',",
    "            'age', '0-6', 'kl', '{example_value3}'))",
    "{px_function}(x2)",
    "",
    "# Remove {keyword}",
    "x3 <- {px_function}(x2, NULL)",
    "{px_function}(x3)",
    .sep = "\n"
  )
}

add_cells2_example <- add_documentation_function(cells2_example)

acrosscells_example <- function(keyword) {
  px_function <- keyword_to_function(keyword)

  stringr::str_glue(
    "@examples",
    "# Set {keyword} for a value",
    "library(tibble)",
    "x1 <-",
    "  population_gl |>",
    "  px() |>",
    "  {px_function}(",
    "    tribble(~gender,  ~age,  ~year, ~cellnote,",
    "             'male', '0-6', '2004', 'Approximation'))",
    "",
    "x2 <-",
    "  x1 |>",
    "  px_cellnote(",
    "    tribble(~gender,   ~age,  ~year, ~cellnote,",
    "           'female',    '*', '2014', 'Uncertainty in ages'))",
    "",
    "# Print {keyword}",
    "{px_function}(x2)",
    "",
    "# Set {keyword} in multiple languagese",
    "x3 <-",
    "  x1 |>",
    "  px_languages(c('en', 'kl')) |>",
    "  {px_function}(",
    "    tribble(~age, ~year, ~language, ~cellnote,",
    "             '*', '2003',  'en', 'Some of the figures are from 2003',",
    "             '*', '2003', 'kl', 'Kisitsisit ilaat 2003-imeersuupput'))",
    "{px_function}(x3)",
    "",
    "# Remove {keyword}",
    "x4 <- {px_function}(x3, NULL)",
    "{px_function}(x4)",
    .sep = "\n"
  )
}

add_acrosscells_example <- add_documentation_function(acrosscells_example)

note_example <- function(keyword) {
  px_function <- keyword_to_function(keyword)

  stringr::str_glue(
    "@examples",
    "library(tibble)",
    "",
    "# Set {keyword} for entire PX-file",
    "x1 <-",
    "  px(population_gl) |>",
    "  {px_function}('Note about PX-file')",
    "",
    "# Print {keyword}",
    "{px_function}(x1)",
    "",
    "# Set {keyword} for entire PX-file in multiple languages",
    "x2 <-",
    "  x1 |>",
    "  px_languages(c('en', 'kl')) |>",
    "  {px_function}(tribble(~language,            ~value,",
    "                       'en',    'English {tolower(keyword)}',",
    "                       'kl', 'Kalaallisut {tolower(keyword)}'",
    "                 )",
    "         )",
    "{px_function}(x2)",
    "",
    "# Set {keyword} for variables",
    "x3 <-",
    "  x1 |>",
    "  {px_function}(tribble(~`variable-code`, ~{tolower(keyword)},",
    "                 'year', 'Some data collected in following year',",
    "                 'age',  'Is rounded down'",
    "                 )",
    "        )",
    "{px_function}(x3)",
    "",
    "# Remove all {keyword}s",
    "x4 <- {px_function}(x3, NULL)",
    .sep = "\n"
  )
}

add_note_example <- add_documentation_function(note_example)

param_validate <- function() {
  stringr::str_glue(
    "@param validate Optional. If TRUE a number of validation checks ",
    "are performed on the px object, and an error is thrown if the ",
    "object is not valid. If FALSE, the checks are skipped, which ",
    "can be usefull for large px objects where the check can be time ",
    "consuming. Use [px_validate()] to manually preform the check."
  )
}

table_param_value_ending <- function(keyword) {
  if (keyword %in% mandatory_keywords()) {
    stringr::str_glue(
      "If NULL, an error is thrown because {keyword} cannot be removed."
    )
  } else {
    stringr::str_glue("If NULL, {keyword} is removed.")
  }
}

table1_param_value <- function(keyword) {
  start <- stringr::str_glue(
    "Optional. A character string. If missing, ",
    "the current {keyword} is returned."
  )

  stringr::str_glue("{start} {table_param_value_ending(keyword)}")
}

table2_param_value <- function(keyword) {
  start <- stringr::str_glue(
    "Optional. A character string to set the value for all ",
    "languages or a data frame with columns 'language' and ",
    "'value' to set it for specific languages. ",
    "If 'value' is missing, the current {keyword} is returned."
  )

  stringr::str_glue("{start} {table_param_value_ending(keyword)}")
}

cells_param_value <- function(keyword, number) {
  colname <- tolower(keyword)

  if (number == "1") {
    optional_columns <- "'variable-code', and 'code'"
  } else if (number == "2") {
    optional_columns <- "'variable-code', 'code', and 'language'"
  } else {
    unexpected_error()
  }

  stringr::str_glue(
    "Optional. A data frame with the columns '{colname}' and one ",
    "or more of the columns: {optional_columns}. If 'value' is ",
    "missing, the current {keyword} is returned. If NULL, {keyword} ",
    "is removed."
  )
}

variables2_param_value <- function(keyword) {
  colname <- tolower(keyword)

  stringr::str_glue(
    "Optional. A character string or data frame.
    \\itemize{{
      \\item Use character to set {keyword} for all languages and STUB/HEADING
      variables.
      \\item Use data frame with columns 'variable-code', 'language' and
      '{colname}' to set {keyword} for specific variables.
      \\item If missing, the current {keyword} is returned.
      \\item If NULL, {keyword} is removed for all variables.
    }}
    "
  )
}

acrosscells_param_value <- function(keyword) {
  colname <- tolower(keyword)

  stringr::str_glue(
    "Optional. A data frame with columns '{colname}' and one or more columns ",
    "with the names of the STUB and HEADING variables.",
    " The '{colname}' column is ",
    "the {colname} text, and the STUB/HEADING columns control which cells the ",
    "note applies to. Use star (*) if a note applies to all cells in a ",
    "variable. Use column 'language'  to set {keyword} for specific ",
    "languages. If 'value' is missing, the current {keyword} is returned. If ",
    "value is NULL, {keyword} is removed."
  )
}

note_param_value <- function(keyword) {
  colname <- tolower(keyword)
  stringr::str_glue(
    "
    Optional. A character string, a data frame, or a list.
     \\itemize{{
       \\item Use character, to set {keyword} for the entire table across
       all languages.
       \\item Use a data frame with columns 'language' and 'value' to set
       {keyword} for the entire table in a specific language.
       \\item Use a data frame with the columns 'variable-code' and '{colname}',
       to set {keyword} for a specific variable across all languages. Add the
       column 'language' to set {keyword} for specific language.
       \\item Use a list of the above elements to set {keyword} in muliple ways.
       This is the same as calling {keyword} multiple times with different
       values.
       \\item If missing, the current {keyword} is returned.
       \\item If NULL, {keyword} is removed for the table and all variables.
    }}
    "
  )
}

pivot_param_value <- function(keyword) {
  stringr::str_glue(
    "
    Optional. A character vector of variable names to change to
    STUB. This also changes the {keyword} order. With names in `variables`
    becoming 1, 2, ... If missing, the current {keyword} variables are returned.
    "
  )
}

return_px_or_df <- function(keyword = NULL) {
  "@returns A px object or data frame."
}

add_return_px_or_df <- add_documentation_function(return_px_or_df)

return_px_or_char_str <- function(keyword = NULL) {
  "@returns A px object or a character string."
}

add_return_px_or_char_str <- add_documentation_function(return_px_or_char_str)

return_px_or_char_vector <- function() {
  "@returns A px object or a character vector."
}

return_px_or_char_vector_or_df <- function() {
  "@returns A px object, a character string, or a data frame."
}

return_px_char_df_or_list <- function() {
  paste0(
    "@returns A px object, a character string, a data frame, or a list of ",
    "character strings and/or data frames."
  )
}

#' This test primarily exist to increase the test coverage

test_that("Documentation returns strings", {
  expect_string <- function(x) {
    expect_true(is.character(x))
  }

  expect_identical(keyword_to_function("AXIS-VERSION"), "px_axis_version")
  expect_identical(function_to_keyword("px_axis_version"), "AXIS-VERSION")
  expect_identical(split_multiline_str_into_vector("a\nb"), c("a", "b"))
  expect_string(add_documentation_table1("MATRIX", "B"))
  expect_string(add_documentation_table1("DECIMALS", "B"))
  expect_string(add_documentation_table2("A", "B", "C"))
  expect_string(add_documentation_head_stub("A"))
  expect_string(add_documentation_variables2("A", "B", "C", "D"))
  expect_string(add_documentation_note("A"))
  expect_string(cells1_example("A", "B", "C"))
  expect_string(cells2_example("A", "B", "C", "D"))
  expect_string(return_px_or_df())
  expect_string(doc_keyword_function_intro("ORDER"))
  expect_string(description_start("CONTENTS"))
  expect_string(table_param_value_ending("CONTENTS"))
  expect_string(table_param_value_ending("DESCRIPTION"))
  expect_string(table1_param_value("CONTENTS"))
  expect_string(table2_param_value("CONTENTS"))
  expect_string(cells_param_value("CONTENTS", "1"))
  expect_string(cells_param_value("CONTENTS", "2"))
  expect_string(variables2_param_value("CONTENTS"))
  expect_string(acrosscells_param_value("CONTENTS"))
  expect_string(note_param_value("CONTENTS"))
  expect_string(pivot_param_value("CONTENTS"))
})

handle_table2_keyword <- function(x, value, keyword, validate) {
  if (missing(value)) {
    return(get_table2_value(x, keyword))
  } else if (is.null(value)) {
    error_if_mandatory_keyword(x, keyword)

    x <- remove_keyword_table2(x, keyword)
  } else {
    x <- modify_table2(x, keyword, value)
  }

  return_px(x, validate)
}

#' @rdname px_baseperiod.px
#' @export
px_baseperiod <- function(x, value, validate) {
  UseMethod("px_baseperiod")
}

#' @eval add_documentation_table2("BASEPERIOD", "year", "ukioq")
px_baseperiod.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "BASEPERIOD", validate)
}


#' @rdname px_cfprices.px
#' @export
px_cfprices <- function(x, value, validate) {
  UseMethod("px_cfprices")
}

#' @eval add_documentation_table2("CFPRICES", "C", "F")
px_cfprices.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "CFPRICES", validate)
}


#' @rdname px_contact.px
#' @export
px_contact <- function(x, value, validate) {
  UseMethod("px_contact")
}

#' @eval add_documentation_table2("CONTACT", "Johan Ejstrud", "Lars Pedersen")
px_contact.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "CONTACT", validate)
}


#' @rdname px_contents.px
#' @export
px_contents <- function(x, value, validate) {
  UseMethod("px_contents")
}

#' @eval add_documentation_table2("CONTENTS", "Population", "Innuttaasut")
px_contents.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "CONTENTS", validate)
}


#' @rdname px_database.px
#' @export
px_database <- function(x, value, validate) {
  UseMethod("px_database")
}

#' @eval add_documentation_table2("DATABASE", "DB_NAME", "DB_NAME_KL")
px_database.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "DATABASE", validate)
}


#' @rdname px_datasymbol1.px
#' @export
px_datasymbol1 <- function(x, value, validate) {
  UseMethod("px_datasymbol1")
}

#' @eval add_documentation_table2("DATASYMBOL1", "missing", "amigaataapput")
px_datasymbol1.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "DATASYMBOL1", validate)
}


#' @rdname px_datasymbol2.px
#' @export
px_datasymbol2 <- function(x, value, validate) {
  UseMethod("px_datasymbol2")
}

#' @eval add_documentation_table2("DATASYMBOL2", "missing", "amigaataapput")
px_datasymbol2.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "DATASYMBOL2", validate)
}


#' @rdname px_datasymbol3.px
#' @export
px_datasymbol3 <- function(x, value, validate) {
  UseMethod("px_datasymbol3")
}

#' @eval add_documentation_table2("DATASYMBOL3", "missing", "amigaataapput")
px_datasymbol3.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "DATASYMBOL3", validate)
}


#' @rdname px_datasymbol4.px
#' @export
px_datasymbol4 <- function(x, value, validate) {
  UseMethod("px_datasymbol4")
}

#' @eval add_documentation_table2("DATASYMBOL4", "missing", "amigaataapput")
px_datasymbol4.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "DATASYMBOL4", validate)
}


#' @rdname px_datasymbol5.px
#' @export
px_datasymbol5 <- function(x, value, validate) {
  UseMethod("px_datasymbol5")
}

#' @eval add_documentation_table2("DATASYMBOL5", "missing", "amigaataapput")
px_datasymbol5.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "DATASYMBOL5", validate)
}


#' @rdname px_datasymbol6.px
#' @export
px_datasymbol6 <- function(x, value, validate) {
  UseMethod("px_datasymbol6")
}

#' @eval add_documentation_table2("DATASYMBOL6", "missing", "amigaataapput")
px_datasymbol6.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "DATASYMBOL6", validate)
}


#' @rdname px_datasymbolnil.px
#' @export
px_datasymbolnil <- function(x, value, validate) {
  UseMethod("px_datasymbolnil")
}

#' @eval add_documentation_table2("DATASYMBOLNIL", "missing", "amigaataapput")
px_datasymbolnil.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "DATASYMBOLNIL", validate)
}


#' @rdname px_description.px
#' @export
px_description <- function(x, value, validate) {
  UseMethod("px_description")
}

#' @eval add_documentation_table2("DESCRIPTION", "Population", "Innuttaasut")
#' @seealso [px_title()]
px_description.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "DESCRIPTION", validate)
}


#' @rdname px_infofile.px
#' @export
px_infofile <- function(x, value, validate) {
  UseMethod("px_infofile")
}

#' @eval add_documentation_table2("INFOFILE", "infofile_en", "infofile_kl")
px_infofile.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "INFOFILE", validate)
}


#' @rdname px_last_updated.px
#' @export
px_last_updated <- function(x, value, validate) {
  UseMethod("px_last_updated")
}

#' @eval add_documentation_table2(
#'  "LAST-UPDATED",
#'  "17070501 15:55",
#'  "20080621 15:55"
#' )
px_last_updated.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "LAST-UPDATED", validate)
}


#' @rdname px_link.px
#' @export
px_link <- function(x, value, validate) {
  UseMethod("px_link")
}

#' @eval add_documentation_table2(
#'  "LINK",
#'  "https://stat.gl/?lang=en",
#'  "https://stat.gl/"
#' )
px_link.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "LINK", validate)
}


#' @rdname px_refperiod.px
#' @export
px_refperiod <- function(x, value, validate) {
  UseMethod("px_refperiod")
}
#' @eval add_documentation_table2(
#'  "REFPERIOD",
#'  "20250311-20260311",
#'  "20250101-20260330"
#' )
px_refperiod.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "REFPERIOD", validate)
}


#' @rdname px_source.px
#' @export
px_source <- function(x, value, validate) {
  UseMethod("px_source")
}

#' @eval add_documentation_table2(
#'  "SOURCE",
#'  "Statistics Greenland",
#'  "Naatsorsueqqissaartarfik"
#' )
px_source.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "SOURCE", validate)
}


#' @rdname px_stockfa.px
#' @export
px_stockfa <- function(x, value, validate) {
  UseMethod("px_stockfa")
}

#' @eval add_documentation_table2("STOCKFA", "S", "F")
px_stockfa.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "STOCKFA", validate)
}


#' @rdname px_subject_area.px
#' @export
px_subject_area <- function(x, value, validate) {
  UseMethod("px_subject_area")
}

#' @eval add_documentation_table2("SUBJECT-AREA", "Population", "Innuttaasut")
px_subject_area.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "SUBJECT-AREA", validate)
}


#' @rdname px_title.px
#' @export
px_title <- function(x, value, validate) {
  UseMethod("px_title")
}

#' @eval add_documentation_table2("TITLE", "Population GR", "Innuttaasut KL")
#' @description TITLE can only be removed if DESCRIPTION is set.
#' @seealso [px_description()]
px_title.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "TITLE", validate)
}


#' @rdname px_units.px
#' @export
px_units <- function(x, value, validate) {
  UseMethod("px_units")
}

#' @eval add_documentation_table2("UNITS", "persons", "inuit amerlassusaat")
px_units.px <- function(x, value, validate = TRUE) {
  handle_table2_keyword(x, value, "UNITS", validate)
}

test_that("Table2 keyword is modified", {
  x <-
    "BEXSTA" |>
    get_data_path() |>
    readRDS() |>
    px()

  expect_identical(px_last_updated(x), NULL)

  datetime <- "2020-01-01 10:00"

  x2 <- px_last_updated(x, datetime)
  expect_identical(px_last_updated(x2), datetime)

  x3 <- px_languages(x, c("en", "dk"))
  expect_identical(px_last_updated(x3), NULL)

  x4 <- px_last_updated(x3, datetime)
  expect_identical(
    px_last_updated(x4),
    dplyr::tibble(
      language = c("en", "dk"),
      value = "2020-01-01 10:00"
    )
  )

  datetime_df <- dplyr::tibble(
    language = c("en", "dk"),
    value = c(
      "2020-01-01 10:00",
      "2022-01-01 10:00"
    )
  )

  x5 <- px_last_updated(x3, datetime_df)
  expect_identical(px_last_updated(x5), datetime_df)

  x6 <- px_last_updated(x5, NULL)
  expect_identical(px_last_updated(x6), NULL)

  expect_error(
    px_last_updated(x3, data.frame(
      language = c("sv", "kl"),
      value = datetime
    )),
    regex = "LANGUAGE"
  )
})

test_that("Other keywords are modified and removed", {
  x <-
    "BEXSTA" |>
    get_data_path() |>
    readRDS() |>
    px()

  x2 <-
    x |>
    px_contents("content") |>
    px_description("description") |>
    px_subject_area("subject area") |>
    px_title("title") |>
    px_units("units") |>
    px_contact("Johan Ejstrud") |>
    px_link("The Legend of Zelda") |>
    px_infofile("infofile") |>
    px_baseperiod("baseperiod") |>
    px_stockfa("S") |>
    px_cfprices("C") |>
    px_source("Statistic Greenland") |>
    px_database("dbname") |>
    px_refperiod("year") |>
    px_datasymbolnil("missing") |>
    px_datasymbol1("missingtype1") |>
    px_datasymbol2("missingtype2") |>
    px_datasymbol3("missingtype3") |>
    px_datasymbol4("missingtype4") |>
    px_datasymbol5("missingtype5") |>
    px_datasymbol6("missingtype6")

  expect_identical(px_contents(x2), "content")
  expect_identical(px_description(x2), "description")
  expect_identical(px_subject_area(x2), "subject area")
  expect_identical(px_title(x2), "title")
  expect_identical(px_units(x2), "units")
  expect_identical(px_contact(x2), "Johan Ejstrud")
  expect_identical(px_link(x2), "The Legend of Zelda")
  expect_identical(px_infofile(x2), "infofile")
  expect_identical(px_baseperiod(x2), "baseperiod")
  expect_identical(px_stockfa(x2), "S")
  expect_identical(px_cfprices(x2), "C")
  expect_identical(px_source(x2), "Statistic Greenland")
  expect_identical(px_database(x2), "dbname")
  expect_identical(px_refperiod(x2), "year")
  expect_identical(px_datasymbolnil(x2), "missing")
  expect_identical(px_datasymbol1(x2), "missingtype1")
  expect_identical(px_datasymbol2(x2), "missingtype2")
  expect_identical(px_datasymbol3(x2), "missingtype3")
  expect_identical(px_datasymbol4(x2), "missingtype4")
  expect_identical(px_datasymbol5(x2), "missingtype5")
  expect_identical(px_datasymbol6(x2), "missingtype6")

  x3 <-
    x2 |>
    px_description(NULL) |>
    px_contact(NULL) |>
    px_link(NULL) |>
    px_infofile(NULL) |>
    px_baseperiod(NULL) |>
    px_stockfa(NULL) |>
    px_cfprices(NULL) |>
    px_source(NULL) |>
    px_database(NULL) |>
    px_refperiod(NULL) |>
    px_datasymbolnil(NULL)

  expect_identical(px_description(x3), NULL)
  expect_identical(px_contact(x3), NULL)
  expect_identical(px_link(x3), NULL)
  expect_identical(px_infofile(x3), NULL)
  expect_identical(px_baseperiod(x3), NULL)
  expect_identical(px_stockfa(x3), NULL)
  expect_identical(px_cfprices(x3), NULL)
  expect_identical(px_source(x3), NULL)
  expect_identical(px_database(x3), NULL)
  expect_identical(px_refperiod(x3), NULL)
  expect_identical(px_datasymbolnil(x3), NULL)

  expect_error(px_contents(x3, NULL), regex = "mandatory")
  expect_error(px_units(x3, NULL), regex = "mandatory")
  expect_error(px_subject_area(x3, NULL), regex = "mandatory")
})

test_that("Either TITLE or DESCRIPTION should be defined", {
  x <-
    "BEXSTA" |>
    get_data_path() |>
    readRDS() |>
    px()

  expect_identical(px_title(x), "")

  expect_error(px_title(x, NULL), regex = "cannot be removed unless")

  x2 <-
    px_description(x, "description") |>
    px_title(NULL)

  expect_identical(px_description(x2), "description")

  expect_error(px_description(x2, NULL), regex = "cannot be removed unless")
})

# Helper functions for throwing errors, warnings and notes.

#' Throw error
#'
#' @param msg String, error message
#'
#' @returns Nothing
#' @keywords internal
error <- function(msg) {
  stop(msg, call. = FALSE)
}

error_if_excel_sheet_does_not_exist <- function(sheet_name, excel_path) {
  if (!excel_sheet_exists(sheet_name, excel_path)) {
    error(
      stringr::str_glue("The sheet '{sheet_name}' is missing in: {excel_path}.")
    )
  }
}

error_if_more_than_one_time_variable <- function(time_variable) {
  if (length(time_variable) > 1) {
    error(stringr::str_glue(
      "The metadata has more than one variable where ",
      "`timeval=TRUE`. Change the type of the variables ",
      "{time_variable}, so only one has `timeval=TRUE`."
    ))
  }
}

error_if_not_exactly_one_figures_variable <- function(figures_var) {
  if (length(figures_var) == 0) {
    error(stringr::str_glue(
      "There is no figures variable. Change the metadata ",
      "so one variable has `pivot=FIGURES`."
    ))
  } else if (length(figures_var) > 1) {
    error(stringr::str_glue(
      "There are more than one variable with figures. ",
      "Change the type of the variables {figures_var}, so ",
      "only one has `pivot=FIGURES`."
    ))
  }
}

error_if_not_exactly_one_data_line <- function(data_line_index) {
  if (length(data_line_index) != 1) {
    error(stringr::str_glue(
      "There are {length(data_line_index)} lines in the ",
      "PX-file like this: 'DATA='. There needs to be ",
      "exactly 1."
    ))
  }
}

excel_max_lines <- function() {
  1048576
}

error_if_too_many_rows_for_excel <- function(df) {
  data_lines <- nrow(df)
  excel_max_lines <- excel_max_lines()

  if (data_lines > excel_max_lines) {
    error(stringr::str_glue(
      "The data cube contains {data_lines} data lines ",
      "which is more than the {excel_max_lines} lines ",
      "supported by Excel. Use argument `rds_data_path=` ",
      "to store the data in an .rds file instead."
    ))
  }
}

unexpected_error <- function() {
  error(paste(
    "An unexpected error occurred. Please report the issue on GitHub,",
    "ideally with a minimal reproducible example.",
    "https://github.com/StatisticsGreenland/pxmake/issues "
  ))
}

#' A list of which variables should be in each sheet
#' @keywords internal
get_mandatory_variables <- function() {
  list(
    "Table" = c("keyword", "value"),
    "Table2" = c("keyword", "code"),
    "Variables" = c(
      "pivot", "order", "variable-code", "variable-label", "variable-type"
    ),
    "Cells" = c("variable-code", "sortorder", "code", "code-label", "precision")
  )
}

error_if_mandatory_keyword <- function(x, keyword) {
  if (keyword %in% mandatory_keywords()) {
    error(
      stringr::str_glue(
        "Keyword '{keyword}' is mandatory and cannot be removed."
      )
    )
  } else if (keyword %in% "TITLE") {
    if (is.null(px_description(x))) {
      error("Keyword TITLE cannot be removed unless DESCRIPTION is defined.")
    }
  } else if (keyword %in% "DESCRIPTION") {
    if (is.null(px_title(x))) {
      error("Keyword DESCRIPTION cannot be removed unless TITLE is defined.")
    }
  }
}

## validate_px checks
error_if_not_list <- function(x) {
  if (!is.list(x)) {
    error("px object must be a list")
  }
}

error_if_not_class_px <- function(x) {
  if (!inherits(x, "px")) {
    error("px object must have class 'px'")
  }
}

error_if_not_list_of_data_frames <- function(x) {
  if (!all(sapply(x, is.data.frame))) {
    error("px object must be a list of data frames")
  }
}

error_if_list_names_are_wrong <- function(x) {
  px_target <- get_base_px()

  input_names <- names(x)
  target_names <- names(px_target)

  missing_names <- setdiff(target_names, input_names)
  invalid_names <- setdiff(input_names, target_names)

  if (length(missing_names) > 0) {
    error(paste0(
      "px object is missing these names: ",
      paste0(missing_names, collapse = ", ")
    ))
  }

  if (length(invalid_names) > 0) {
    error(paste0(
      "px object contains these invalid names: ",
      paste0(invalid_names, collapse = ", ")
    ))
  }
}

error_if_data_frame_is_missing_column <- function(x) {
  px_target <- get_base_px()

  for (name in names(px_target)) {
    target_columns <- colnames(px_target[[name]])
    input_columns <- colnames(x[[name]])

    missing_columns <- setdiff(target_columns, input_columns)

    if (length(missing_columns) > 0) {
      error(paste0(
        "px object: '", name, "' is missing these columns: ",
        paste0(missing_columns, collapse = ", ")
      ))
    }
  }
}

error_if_multiple_time_variables <- function(x) {
  if (length(px_timeval(x)) > 1) {
    error(paste0(
      "px object has more than one time variable: ",
      paste0(px_timeval(x), collapse = ", ")
    ))
  }
}

error_if_timeval_isnot_heading_or_stub <- function(x) {
  if (!is.null(px_timeval(x))) {
    if (!px_timeval(x) %in% c(px_heading(x), px_stub(x))) {
      error("TIMEVAL has to be a STUB or HEADING variable.")
    }
  }
}

error_if_contvariable_isnot_heading_or_stub <- function(x) {
  if (!is.null(px_contvariable(x))) {
    if (!px_contvariable(x) %in% c(px_heading(x), px_stub(x))) {
      error("CONTVARIABLE has to be a STUB or HEADING variable.")
    }
  }
}

error_if_variable_label_is_na <- function(x) {
  if (any(is.na(x$variables2$`variable-label`))) {
    error("px object: in x$variables2 'variable-label' has NA values.")
  }
}

error_if_misplaced_keywords_in_table <- function(x, table_name) {
  if (table_name == "table1") {
    other_keywords <-
      pxmake::px_keywords |>
      dplyr::filter(!(.data$table_meta & !.data$language_dependent)) |>
      dplyr::pull(.data$keyword)
  } else if (table_name == "table2") {
    other_keywords <-
      pxmake::px_keywords |>
      dplyr::filter(!(.data$table_meta & .data$language_dependent)) |>
      dplyr::pull(.data$keyword)
  } else {
    unexpected_error()
  }

  misplaced_keywords <- intersect(x[[table_name]]$keyword, other_keywords)

  if (length(misplaced_keywords) > 0) {
    error(paste0(
      "px object: '", table_name, "' contains misplaced keywords: ",
      paste0(misplaced_keywords, collapse = ", ")
    ))
  }
}

error_if_variable_code_not_in_data <- function(x, table_name) {
  variables_not_in_data <- setdiff(
    x[[table_name]]$`variable-code`, names(x$data)
  )

  if (length(variables_not_in_data) > 0) {
    error(paste0(
      "px object: x$", table_name, " contains variable-codes not in x$data: ",
      paste0(variables_not_in_data, collapse = ", ")
    ))
  }
}

error_if_used_languages_are_not_defined <- function(x) {
  languages_in_tables <-
    unique(c(
      x$table2$language,
      x$variables2$language,
      x$cells2$language
    )) |>
    na.omit() |>
    as.character()

  undefined_languages <- setdiff(languages_in_tables, defined_languages(x))

  if (length(undefined_languages) > 0) {
    error(paste0(
      "px object contains languages that are not defined in 'x$languages' ",
      "or as keyword 'LANGUAGES' in x$table1: ",
      paste0(undefined_languages, collapse = ", ")
    ))
  }
}

error_if_language_not_in_languages <- function(x) {
  if (length(px_languages(x)) > 0) {
    if (!any(is.null(px_language(x)), px_language(x) %in% px_languages(x))) {
      error("px object: LANGUAGE is not in x$languages.")
    }
  }
}

error_if_language_is_undefined <- function(language, x) {
  if (!language %in% defined_languages(x)) {
    error(paste0("Language '", language, "' is not defined in the px object."))
  }
}

error_if_data_column_is_not_defined <- function(x, table_name) {
  undefined_variables <- setdiff(
    colnames(x$data), x[[table_name]]$`variable-code`
  )

  if (length(undefined_variables) > 0) {
    error(paste0(
      "px object: x$data contains columns that are not defined in x$",
      table_name, ": ", paste0(undefined_variables, collapse = ", ")
    ))
  }
}

error_if_value_contains_quotation_marks <- function(x) {
  contains_quote <- function(col) any(grepl('"', col, fixed = TRUE))
  metadata_frames <- x[names(x) != "data"]

  if (any(rapply(metadata_frames, contains_quote, classes = "character"))) {
    error("px object: values cannot contain quotation marks")
  }
}

error_if_data_table_contains_duplicates <- function(x) {
  key_data <- px_data(x) |> dplyr::select(-px_figures(x))

  has_duplicates <- ncol(key_data) > 0 && vctrs::vec_duplicate_any(key_data)

  if (has_duplicates) {
    duplicates <- key_data |>
      dplyr::group_by(across(everything())) |>
      dplyr::filter(dplyr::n() > 1)

    error(paste0(
      "px save: cannot save because data table contains duplicates:\n\n",
      paste(utils::capture.output(print(duplicates)),
        collapse = "\n"
      )
    ))
  }
}

is_url <- function(path) {
  if (!is.character(path) || length(path) != 1) {
    return(FALSE)
  }

  stringr::str_detect(path, "^https?://")
}

#' Check all arguments to px()
#'
#' @inheritParams px
#'
#' @returns Nothing
#' @keywords internal
validate_px_arguments <- function(input, data) {
  if (!any(
    is_px_file(input), is_xlsx_file(input), is.data.frame(input),
    is_rds_file(input), is_parquet_file(input), is.null(input),
    is_url(input)
  )) {
    error("Argument 'input' has wrong format. See ?px.")
  }

  if (is_rds_file(input)) {
    if (!"data.frame" %in% class(readRDS(input))) {
      error(paste0(
        "Argument 'input' has wrong format. The .rds file does not ",
        "contain a data frame. See ?px."
      ))
    }
  }

  if (!any(
    is.null(data), is.data.frame(data), is_rds_file(data), is_parquet_file(data)
  )) {
    error("Argument 'data' has wrong format. See ?px.")
  }

  if (!is.null(data) && !is_xlsx_file(input)) {
    error("Argument 'data' can only be used if 'input' is an .xlsx file.")
  }
}

#' Check all arguments to px_save()
#'
#' @inheritParams px_save
#'
#' @returns Nothing
#' @keywords internal
validate_px_save_arguments <- function(x, path, save_data, data_path) {
  if (!any(is_px_file(path), is_xlsx_file(path), is_r_file(path))) {
    error("Argument 'path' must be a path to an .px, .xlsx or .R file.")
  }

  if (!is.logical(save_data)) {
    error("Argument 'save_data' must be TRUE or FALSE.")
  }

  if (all(
    !is.null(data_path),
    all(
      !is_rds_file(data_path),
      !is_parquet_file(data_path)
    )
  )) {
    error("Argument 'data_path' must be a path to an .rds or .parquet file.")
  }

  if (all(
    !is.null(data_path), all(!is_xlsx_file(path), !is_r_file(path))
  )) {
    error(
      "Argument 'data_path' can only be used if 'path' is an .xlsx or .R file."
    )
  }

  if (all(!is.null(data_path), isFALSE(save_data))) {
    error("Argument 'data_path' can only be used if 'save_data' is TRUE.")
  }

  if (all(isFALSE(save_data), !is_xlsx_file(path))) {
    error("Argument 'save_data' can only be FALSE if 'path' is an .xlsx file.")
  }

  if (all(
    is_xlsx_file(path),
    save_data,
    is.null(data_path),
    nrow(x$data) > excel_max_lines()
  )
  ) {
    error(paste0(
      "x$data has too many rows to be saved as an .xlsx file. Use ",
      "'save_data = FALSE' or 'data_path' instead."
    ))
  }

  error_if_data_table_contains_duplicates(x)
}

#' Check all arguments to px_micro()
#'
#' @inheritParams px_micro
#'
#' @returns Nothing
#' @keywords internal
validate_px_micro_arguments <- function(x, out_dir) {
  if (!inherits(x, "px")) {
    error("Argument 'x' must be a px object.")
  }

  if (!is.null(out_dir)) {
    if (!is.character(out_dir)) {
      error("Argument 'out_dir' must be a character string.")
    }

    if (!dir.exists(out_dir)) {
      error("Argument 'out_dir' directory does not exist.")
    }
  }
}

#' Throw error if argument is not a character string
#' @keywords internal
error_if_agument_is_not_character_string <- function(name, value) {
  if (!all(is.character(value), length(value) == 1)) {
    error(paste0("Argument '", name, "' must be a character string."))
  }
}

#' Check all arguments to px_classification
#'
#' @inheritParams px_classification
#'
#' @returns Nothing
#' @keywords internal
validate_px_classification_arguments <- function(name,
                                                 prestext,
                                                 domain,
                                                 df,
                                                 vs_path,
                                                 agg_paths) {
  any_path_argument_is_defined <- (!missing(vs_path) | !missing(agg_paths))

  any_non_path_argument_is_defined <-
    (!missing(name) | !missing(prestext) | !missing(domain) | !missing(df))

  if (any_path_argument_is_defined && any_non_path_argument_is_defined) {
    error(paste0(
      "Cannot use path arguments and other aguments together. ",
      "Either define only name/prestext/domain/df, or only ",
      "vs_path/agg_paths."
    ))
  }

  if (any_path_argument_is_defined) {
    if (!missing(vs_path)) {
      if (!is.character(vs_path)) {
        error("Argument 'vs_path': must be a character string.")
      }

      if (!file.exists(vs_path)) {
        error("Argument 'vs_path': file does not exist.")
      }
    }

    if (!missing(agg_paths)) {
      if (!is.character(agg_paths)) {
        error("Argument 'agg_paths': must be a character vector.")
      }
    }
  }

  if (all(any_path_argument_is_defined, missing(vs_path))) {
    error(paste0(
      "Argument 'vs_path' is missing. If 'agg_paths' is defined, ",
      "'vs_path' must also be defined."
    ))
  }

  if (any_non_path_argument_is_defined) {
    error_if_agument_is_not_character_string("name", name)
    error_if_agument_is_not_character_string("prestext", prestext)
    error_if_agument_is_not_character_string("domain", domain)

    if (!is.data.frame(df)) {
      error("Argument 'df' must be a data frame.")
    }

    if (!"valuecode" %in% names(df)) {
      error("Argument 'df' is missing column 'valuecode'.")
    }
  }
}

#' Check all arguments to px_data
#'
#' @inheritParams px_data
#'
#' @returns Nothing
#' @keywords internal
validate_px_data_arguments <- function(x, value, labels, sort, validate) {
  if (!inherits(x, "px")) {
    error("Argument 'x' must be a px object.")
  }

  if (!missing(value)) {
    if (!(is.data.frame(value) || is.null(value))) {
      error("Argument 'value' must be a data frame or NULL.")
    }
  }

  if (!missing(labels)) {
    if (!is.logical(labels) && !is.character(labels)) {
      error("Argument 'labels' must be TRUE/FALSE or a character string.")
    }
  }

  # labels and sort can only be used if value is missing
  if (!missing(value) && !isFALSE(labels)) {
    error("Argument 'labels' can only be used if 'value' is missing.")
  }

  if (!missing(value) && !isFALSE(sort)) {
    error("Argument 'sort' can only be used if 'value' is missing.")
  }

  if (!missing(validate)) {
    if (!is.logical(validate)) {
      error("Argument 'validate' must be TRUE or FALSE.")
    }
  }

  if (!missing(sort)) {
    if (!is.logical(sort)) {
      error("Argument 'sort' must be TRUE or FALSE.")
    }
  }
}

error_if_vs_file_is_missing_mandatory_headings <- function(vs_lines) {
  headings <-
    stringr::str_extract_all(
      vs_lines,
      classification_file_section_heading_regexp()
    )

  missing_headings <- setdiff(
    c("[Descr]", "[Domain]", "[Valuecode]"),
    headings
  )

  if (length(missing_headings) > 0) {
    error(paste(
      ".vs file is missing mandatory sections: ",
      paste0(missing_headings, collapse = ", ")
    ))
  }
}

error_if_aggregation_file_is_missing_mandatory_headings <- function(agg_lines) {
  headings <-
    stringr::str_extract_all(
      agg_lines,
      classification_file_section_heading_regexp()
    )

  missing_headings <- setdiff(c("[Aggreg]", "[Aggtext]"), headings)

  if (length(missing_headings) > 0) {
    error(paste(
      ".agg file is missing mandatory sections: ",
      paste0(missing_headings, collapse = ", ")
    ))
  }
}

test_that("MAP is added in variables2", {
  x <- px(get_px_file_path("MAP"))

  map_rows_in_table_2 <-
    x$table2 |>
    dplyr::filter(keyword == "MAP")

  expect_identical(
    map_rows_in_table_2,
    dplyr::filter(map_rows_in_table_2, FALSE)
  )

  map_rows_in_variables2 <-
    x$variables2 |>
    tidyr::drop_na(map)

  expect_true(nrow(map_rows_in_variables2) > 0)
})

test_that("pxjob exists without errors (exit code 0)", {
  skip_if_not_installed("pxjob64Win", minimum_version = "1.1.0")

  expect_that_pxjob_runs_without_erros <- function(table_name) {
    px_path <- create_px_file(table_name)
    output <- get_pxjob_file_path(table_name)

    pxjob_exit_code <- pxjob64Win::pxjob(px_path, output)
    expect_equal(0, pxjob_exit_code)

    # File is removed manually because pxjob_clean() cannot be used because the
    # exit code from pxjob64Win::pxjob is needed.
    file.remove(output)
  }

  expect_that_pxjob_runs_without_erros("BEXLTALL")
  expect_that_pxjob_runs_without_erros("BEXSTA")
  expect_that_pxjob_runs_without_erros("FOTEST")
})

github_raw_url <-
  "https://raw.githubusercontent.com/StatisticsGreenland/pxmake/"

test_that("px runs without errors (URL)", {
  skip_if_offline(host = "raw.githubusercontent.com")

  px(input = paste0(github_raw_url, "main/tests/testthat/fixtures/px/TUX01.px"))

  expect_true(TRUE)
})

test_that("px can take a parquet url is a input", {
  skip_if_offline(host = "raw.githubusercontent.com")

  px(
    input = paste0(github_raw_url, "main/tests/testthat/fixtures/micro.parquet")
  )

  expect_true(TRUE)
})

test_that("CONTVARIABLE is handled correctly", {
  x <- px(get_px_file_path("CONTVARIABLE"))

  contvariable_name <- "ContentsCode"

  # CONTVARIABLE should be TRUE
  x$variables1 |>
    dplyr::filter(`variable-code` == contvariable_name) |>
    dplyr::pull(contvariable) |>
    expect_true()

  # CONTVARIABLE should be FALSE for all other variables
  x$variables1 |>
    dplyr::filter(`variable-code` != contvariable_name) |>
    dplyr::pull(contvariable) |>
    any() |>
    expect_false()

  # CONTVARIABLE should not be created as a keyword in table2
  x$table2 |>
    dplyr::filter(keyword == "CONTVARIABLE") |>
    nrow() |>
    expect_equal(0)
})

---
title: "Micro files"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Micro files}
  %\VignetteEncoding{UTF-8}
  %\VignetteEngine{knitr::rmarkdown}
editor_options: 
  chunk_output_type: console
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Use case

`px_micro()` exists to support a specific use case for [Statistics Greenland](https://stat.gl/default.asp?lang=en). 

They use it to create small PX-files to showcase and present metadata from a lager data set which cannot be made publicly available. See an example on [Statistic Greenland's microdata for Research and Analysis](https://bank.stat.gl/pxweb/en/GSmicro/).

## `px_micro()`

Apart from `px_save()`, `px_micro()` is the only other function that can save px objects as PX-files.

`px_micro()` turns a px object into many smaller PX-files, each containing a subset of the variables in the original px object.

## Input data

The basis of micro files are usually a data set which doesn't have a count variable (like most PX-files). `px_micro()` will instead create a count of each individual variable.

In this example we will use the built-in data  data set `greenlanders`.

```{r, include = FALSE}
set.seed(0)
micro_dir <- file.path("micro_files")
unlink(micro_dir, recursive = TRUE)
```

```{r}
library(pxmake)

greenlanders |>
  dplyr::sample_n(10) |>
  dplyr::arrange_all()
```

## How to create micro files

Create a px object with `px()`, and pass it to `px_micro()`.

```{r}
# Create px object
x <- px(greenlanders)

# Create folder for micro files
micro_dir <- file.path("micro_files")
dir.create(micro_dir)

# Write micro files to folder
px_micro(x, out_dir = micro_dir)
```
The folder now contains three PX-files, one for each variable except 'age'.
```{r}
list.files(micro_dir)
```

The reason 'age' didn't get a PX-file is because it is the HEADING variable in `x`, and `px_micro()` creates a file for each non-HEADING variable. Instead the HEADNING variable is used in all the created PX-files.

```{r}
# Print HEADING variables
px_heading(x)

# Print non-HEADING variables
c(px_stub(x), px_figures(x))
```

In this case, we want 'cohort' to be heading, and to create a PX-file for 'gender', 'age' and 'municipality'.

```{r}
x2 <-
  x |>
  px_stub("age") |> # Change age to STUB
  px_heading("cohort") # Change cohort to HEADING
```

```{r}
# Clear folder
unlink(file.path(micro_dir, "*.px"))

px_micro(x2, out_dir = micro_dir)
```

The folder now contains the files we wanted.

```{r}
list.files(micro_dir)
```

Each file contains one of the three variables as STUB, 'cohort' as HEADING, and a variable 'n' which is the count of each combination of the variables.

```{r}
px(file.path(micro_dir, "age.px"))$data

px(file.path(micro_dir, "gender.px"))$data

px(file.path(micro_dir, "municipality.px"))$data
```

## Keyword values

In general the keyword values from the px object are carried over to the micro files. This is the case for keywords like 'MATRIX', 'SUBJECT-CODE', 'CONTACT', 'LANGUAGE', 'CODEPAGE', etc.

To change keywords across all the micro files, the easiest is to change them in the px object before calling `px_micro()`.

```{r, eval = FALSE}
# Change CONTACT in all micro files
x2 |>
  px_contact("Johan Ejstrud") |>
  px_micro(out_dir = micro_dir)
```

However, some keywords need to be changed individually for each micro file. To do so, create a data frame with the column 'variable' and a column for each px keyword to change.

```{r}
individual_keywords <- tibble::tribble(
  ~variable, ~px_description,
  "age", "Age count 18-99",
  "gender", "Gender count",
  "municipality", "Municipality 2024"
)
```

Supply this dataframe to the `keyword_values` argument of `px_micro()`.

```{r}
px_micro(x2, out_dir = micro_dir, keyword_values = individual_keywords)
```

DESCRIPTION is changed in the micro files: 

```{r}
px(file.path(micro_dir, "age.px")) |> px_description()
px(file.path(micro_dir, "gender.px")) |> px_description()
px(file.path(micro_dir, "municipality.px")) |> px_description()
```


### Multilingual files

For multilingual files add a 'language' column to `keyword_values`.

```{r}
x3 <-
  x2 |>
  px_language("en") |>
  px_languages(c("en", "kl"))


individual_keywords_ml <-
  tibble::tribble(
    ~variable, ~language, ~px_description, ~px_matrix,
    "age", "en", "Age count 18-99", "AGE",
    "age", "kl", "Ukiut 18-99", NA,
    "gender", "en", "Gender count", "GEN",
    "gender", "kl", " Suiaassuseq", NA,
    "municipality", "en", "Municipality 2024", "MUN",
    "municipality", "kl", "Kommuni 2024", NA
  )

px_micro(x3, out_dir = micro_dir, keyword_values = individual_keywords_ml)
```

Here 'px_description' varies for each language, and 'px_matrix' is only set for one of the languages, since it is not a language dependent keywords. For language independant keywords it doesn't matter which language the value is set for.

### Filenames

The filenames of the micro files are by default the name of the variable, however these can also be changed by passing a 'filename' column to 'keyword_values'

```{r}
individual_keywords2 <-
  individual_keywords |>
  dplyr::mutate(filename = paste0(variable, "_2024", ".px"))

# Clear folder
unlink(file.path(micro_dir, "*.px"))

px_micro(x2, out_dir = micro_dir, keyword_values = individual_keywords2)

list.files(micro_dir)
```

---
title: "Languages"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Languages}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

The PX-file format supports having multiple langauges in one file. Let's start by creating a minimal example of a px object, based on the built in data set `population_gl`.

```{r setup}
library(pxmake)

x <- px(population_gl)
```
Currently, `x` has neither a main language nor any additional languages.

```{r}
px_language(x) # Main language
px_languages(x) # All languages
```

To set the main language, use `px_language()`, and add the [ISO 639](https://en.wikipedia.org/wiki/List_of_ISO_639_language_codes) language code.

```{r}
x1 <- px_language(x, "kl")
```

This adds the keyword 'LANGUAGE=kl' if the PX-file is saved, but otherwise there are no changes. 

To set multiple languages, use `px_languages()` with a list of language codes.

```{r}
x2 <- px_languages(x1, c("da", "kl"))
```

This doubles the number of values for all language dependent keywords. For example CONTACT, can be set simultaneously for both languages:

```{r}
x3 <- px_contact(x2, "Johan Ejstrud")
```

or it can be set with distinct values for the two languages:

```{r}
x4 <- px_contact(x3, dplyr::tribble(
  ~language, ~value,
  "kl", "Lars Pedersen",
  "da", "Johan Ejstrud"
))
```

In general, changing a keyword for multiple languages, requires a data frame where the column 'language' has been added. See the help pages for each functions to see the exact details on which columns are needed in the data frame.

---
title: "First PX-file"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{First PX-file}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

To create your first PX-file with pxmake, you can either start with an existing PX-file, or use a data set.

## How to create a PX-file from a data set

All workflows in pxmake, starts the same way: by using the `px()` to create a *px object*. Here we are using the built in dataset `population_gl`.

```{r}
library(pxmake)

population_gl |>
  head(10) |>
  print()
```

To create the px object, simply pass the data set to the `px()` function.

```{r}
x <- px(population_gl)
```

`x` is now a px object. The word *object* just means that the PX-file is stored in a specific format that all functions in the pxmake package uses.

Use `px_save()` to save the px object as a PX-file.

```{r}
px_save(x, "population_gl.px")
```

Since we provided no metadata, `px()` added a minimal set of metadata, which is necessary for creating a valid PX-file.

The resulting PX-file can be seen below.

```{r, comment="", echo = FALSE}
readLines("population_gl.px") |> cat(sep = "\n")
```

## How to create a PX-file from a PX-file

If you already have a PX-file, that you want to manipulate using pxmake, simple pass the path of the file, to the `px()` function.

```{r}
x2 <- px("population_gl.px")
```

Just as with the dataset, `x2` is now a px object and can be saved or modified using the other functions in pxmake.


## Modifying a px object

The real fun starts when you start modifying the px object, using one of pxmake many built in functions.

In general, each px keyword has a corresponding function in pxmake. For example, to change the title of the PX-file, you can use the `px_title()` function.

```{r}
x3 <- px_title(x, "Population in Greenland")
```

The `px_title()` function returns a new px object, with the title changed. The original px object is not modified.


```{r}
x3 |>
  px_codepage("UTF-8") |> # Change file encoding
  px_matrix("pop") |>
  px_contact("Johan Ejstrud") |>
  px_subject_code("GL") |>
  px_subject_area("Greenland") |>
  px_timeval("year") |>
  px_contents("Population in Greenland") |>
  px_units("People") |>
  px_note("See information about data: ?population_gl") |>
  px_last_updated(format(Sys.time(), "%Y%m%d %H:%M")) |>
  px_stub(c("age", "gender")) |> # Change order of STUB variables
  px_save("population_gl_modified.px")
```

The resulting PX-file can be seen below.

```{r, comment="", echo = FALSE}
readLines("population_gl_modified.px") |> cat(sep = "\n")
```

## Advanced use
See the other vignettes for more advanced use cases:
- `vignette("languages")`
- `vignette("micro-files")`

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{format_time_values}
\alias{format_time_values}
\title{Format time values for PX-file}
\usage{
format_time_values(values)
}
\arguments{
\item{values}{Time values}
}
\value{
A character vector
}
\description{
Format time values for PX-file
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_data.R
\name{update_data_table}
\alias{update_data_table}
\title{Add value as data table in x}
\usage{
update_data_table(x, value)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A data frame. If missing, the current
DATA is returned. If NULL, all data rows are removed.}
}
\description{
Add value as data table in x
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/px_file.R
\name{get_data_cube}
\alias{get_data_cube}
\title{Get data cube used in PX-file format}
\usage{
get_data_cube(metadata_df, data_df)
}
\arguments{
\item{metadata_df}{Data frame with metadata.}

\item{data_df}{A data frame with data.}
}
\value{
A data frame
}
\description{
Get data cube used in PX-file format
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{is_path_extension}
\alias{is_path_extension}
\title{Check if a path has a specific extension (function factory)}
\usage{
is_path_extension(extension)
}
\arguments{
\item{extension}{String, file name extension}
}
\value{
Logic
}
\description{
Check if a path has a specific extension (function factory)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classification.R
\name{new_classification}
\alias{new_classification}
\title{Create new classification object}
\usage{
new_classification(name, prestext, domain, df)
}
\arguments{
\item{name}{Optional. Name of the classification.}

\item{prestext}{Optional. Presentation text.}

\item{domain}{Optional. Character vector with domain names. Used to link to
PX-file.}

\item{df}{Optional. A data frame with required column 'valuecode' and
optional column 'valuetext', if the codes have texts. Each additional column
represents an aggregation. The column name is the name of the aggregation.
If the column type is character the aggregation levels will be sorted
alphabetically; use factors to control the ordering.}
}
\value{
A classification object
}
\description{
Constructor for internal functions
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/excel.R
\name{get_figures_variable_from_excel}
\alias{get_figures_variable_from_excel}
\title{Get figures variable from Excel workbook}
\usage{
get_figures_variable_from_excel(excel_path)
}
\arguments{
\item{excel_path}{Path to Excel metadata workbook.}
}
\value{
Character, name of figures variable.
}
\description{
Get figures variable from Excel workbook
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classification.R
\name{get_section}
\alias{get_section}
\title{Return lines in section}
\usage{
get_section(lines, head_line)
}
\arguments{
\item{lines}{A character vector}

\item{head_line}{A numeric value, line number of section heading}
}
\description{
Return lines in section
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_data.R
\name{px_data}
\alias{px_data}
\alias{px_data.px}
\title{DATA}
\usage{
px_data(x, value, labels, sort, validate)

\method{px_data}{px}(x, value, labels = FALSE, sort = FALSE, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A data frame. If missing, the current
DATA is returned. If NULL, all data rows are removed.}

\item{labels}{Optional. Logic or character vector. If TRUE, the data table
is returned with VALUES instead of CODES. By default the VALUES of the main
language are returned, use a character language code to return VALUES for a
specific language.}

\item{sort}{Optional. If TRUE, the data table is returned in the sort order
defined by \code{\link[=px_order]{px_order()}}. If FALSE, the data table is returned as is.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object or data frame.
}
\description{
Inspect or change DATA.
}
\details{
If adding a new data frame, metadata is generated for the new
columns and removed for columns that are no longer present.
}
\examples{
x1 <- px(population_gl)

# Print data table
px_data(x1)

# Change data table
population_gl_2024 <- subset(population_gl, year == 2024)

x2 <- px_data(x1, population_gl_2024)

# Return data table with VALUES instead of CODES
px_data(x1, labels = TRUE)

# Return VALUES for a specific language
x_mult <-
  x1 |>
  px_languages(c("en", "gl"))

px_data(x_mult, labels = "gl")

}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A32\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C502\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_languages.R
\name{px_languages}
\alias{px_languages}
\alias{px_languages.px}
\title{LANGUAGES}
\usage{
px_languages(x, value, validate)

\method{px_languages}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character vector. If missing, the current LANGUAGES
are returned. If NULL, LANGUAGES are removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object
}
\description{
Inspect or change LANGUAGES.
}
\details{
If LANGUAGE is defined it should be one of the values in LANGUAGES.

If LANGUAGE is set, it is considered the main language.
If LANGUAGE is not set, the first language in LANGUAGES is considered the
main language.
}
\examples{
# Set LANGUAGES to 'en' and 'kl', with 'en' as main language
x1 <-
  population_gl |>
  px() |>
  px_languages(c("en", "kl"))

# Print LANGUAGES
px_languages(x1)

# Remove LANGUAGES
x2 <- px_languages(x1, NULL)
px_languages(x2)

}
\seealso{
\code{\link{px_language}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/r_script.R
\name{convert_df_to_code}
\alias{convert_df_to_code}
\title{Create code to construct data frame}
\usage{
convert_df_to_code(df)
}
\arguments{
\item{df}{A data frame}
}
\value{
A character vector
}
\description{
Convert data frame to the code nessasary to construct it as a tibble.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata_df.R
\name{sort_metadata_df}
\alias{sort_metadata_df}
\title{Sort metadata data frame}
\usage{
sort_metadata_df(metadata_df)
}
\arguments{
\item{metadata_df}{Data frame with metadata.}
}
\value{
A data frame
}
\description{
The data frame is first sorted by the keyword order defined in the
px-specification and then by the language order.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_variable2.R
\name{px_domain}
\alias{px_domain}
\alias{px_domain.px}
\title{DOMAIN}
\usage{
px_domain(x, value, validate)

\method{px_domain}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string or data frame.
\itemize{
\item Use character to set DOMAIN for all languages and STUB/HEADING
variables.
\item Use data frame with columns 'variable-code', 'language' and
'domain' to set DOMAIN for specific variables.
\item If missing, the current DOMAIN is returned.
\item If NULL, DOMAIN is removed for all variables.
}}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object, a character string, or a data frame.
}
\description{
Inspect or change DOMAIN.
}
\examples{
# Set DOMAIN for all languages
x1 <-
  px(population_gl) |>
  px_domain('aggregation1')

# Print DOMAIN
px_domain(x1)

# Set DOMAIN for individual variables
library(tibble)
x2 <-
  x1 |>
  px_domain(tribble(~`variable-code`, ~domain,
                    'gender', 'aggregation2',
                    'age',    'aggregation3'))
px_domain(x2)

# Set DOMAIN for individual languages
x3 <-
  x2 |>
  px_languages(c('en', 'kl')) |>
  px_domain(tribble(~`variable-code`, ~language, ~domain,
                    'gender',    'en',      'aggregation2_en',
                    'gender',    'kl',      'aggregation2_kl',
                    'age',       'en',      'aggregation3_en'))
px_domain(x3)

# Remove DOMAIN
x4 <- px_domain(x3, NULL)
px_domain(x4)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A45\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C214\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/r_script.R
\name{save_px_as_r_script}
\alias{save_px_as_r_script}
\title{Save px object as an R script}
\usage{
save_px_as_r_script(x, path, data_path)
}
\arguments{
\item{x}{A px object}

\item{path}{Path to save R script at}

\item{data_path}{\code{.rds} or \code{.parquest} path to save data at}
}
\value{
Nothing
}
\description{
Running the R script creates the px object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{population_gl}
\alias{population_gl}
\title{Population Greenland}
\format{
An object of class \code{tbl_df} (inherits from \code{tbl}, \code{data.frame}) with 30 rows and 4 columns.
}
\source{
\url{https://bank.stat.gl/pxweb/en/Greenland/Greenland__BE__BE01__BE0120/BEXSTA.px/}
}
\usage{
population_gl
}
\description{
A subset of the population count data available in Statistic Greenland's
BEESTA table.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{read_lines_guess_encoding}
\alias{read_lines_guess_encoding}
\title{Read lines from file with guessed encoding}
\usage{
read_lines_guess_encoding(path)
}
\arguments{
\item{path}{Path to file}
}
\value{
A character vector
}
\description{
Wrapper around readLines to guess encoding before reading.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{quote_unless_yes_no}
\alias{quote_unless_yes_no}
\title{Add quotes around unless in some very specific cases required by
the px format}
\usage{
quote_unless_yes_no(str)
}
\arguments{
\item{str}{String to quote}
}
\value{
String
}
\description{
Add quotes around unless in some very specific cases required by
the px format
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata_df.R
\name{get_variable_label}
\alias{get_variable_label}
\title{Get variable code and label}
\usage{
get_variable_label(metadata_df)
}
\arguments{
\item{metadata_df}{Data frame with metadata.}
}
\value{
A data frame
}
\description{
Get a data frame with variable codes and label in all languages.
If VARIABLECODE is used, the relation is defined there, but otherwise the
variable code is set to be the main languages' label.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_variable2.R
\name{px_variable_label}
\alias{px_variable_label}
\alias{px_variable_label.px}
\title{Change VARIABLE-LABEL}
\usage{
px_variable_label(x, value, validate)

\method{px_variable_label}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string or data frame.
\itemize{
\item Use character to set VARIABLE-LABEL for all languages and STUB/HEADING
variables.
\item Use data frame with columns 'variable-code', 'language' and
'variable-label' to set VARIABLE-LABEL for specific variables.
\item If missing, the current VARIABLE-LABEL is returned.
\item If NULL, VARIABLE-LABEL is removed for all variables.
}}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object, a character string, or a data frame.
}
\description{
Inspect or change VARIABLE-LABEL.

The variable label is the name that is shown in the PX-file.
}
\examples{
# Set VARIABLE-LABEL for individual variables
library(tibble)
x1 <-
  px(population_gl) |>
  px_variable_label(tribble(
    ~`variable-code`, ~`variable-label`,
    "gender", "Gender",
    "age", "Age"
  ))
px_variable_label(x1)

# Set VARIABLE-LABEL for individual languages
x2 <-
  x1 \%>\%
  px_languages(c("en", "kl")) |>
  px_variable_label(tribble(
    ~`variable-code`, ~language, ~`variable-label`,
    "gender", "en", "Gender",
    "gender", "kl", "Suiaassuseq",
    "age", "en", "Age",
    "age", "kl", "Ukiut"
  ))
px_variable_label(x2)

# Remove VARIABLE-LABEL
x3 <- px_variable_label(x2, NULL)
px_variable_label(x3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pxmake-package.R
\docType{package}
\name{pxmake-package}
\alias{pxmake}
\alias{pxmake-package}
\alias{pxmake-package.R}
\title{pxmake: Make PX-Files in R}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

Create PX-files from scratch or read and modify existing ones. Includes a function for every PX keyword, making metadata manipulation simple and human-readable.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/StatisticsGreenland/pxmake}
  \item \url{https://statisticsgreenland.github.io/pxmake/}
  \item Report bugs at \url{https://github.com/StatisticsGreenland/pxmake/issues}
}

}
\author{
\strong{Maintainer}: Johan Ejstrud \email{johan@ejstrud.com}

Authors:
\itemize{
  \item Lars Pedersen \email{larp@stat.gl}
}

Other contributors:
\itemize{
  \item Statistics Greenland (https://stat.gl/) [copyright holder]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata_df.R
\name{replace_na_language_with_main_language}
\alias{replace_na_language_with_main_language}
\title{Impute missing language}
\usage{
replace_na_language_with_main_language(metadata_df)
}
\arguments{
\item{metadata_df}{Data frame with metadata.}
}
\value{
A data frame
}
\description{
Add main language as language if language is missing.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{add_sub_key_to_keyword}
\alias{add_sub_key_to_keyword}
\title{Add a sub key word (see px-specification for details on sub keys)}
\usage{
add_sub_key_to_keyword(keyword, name)
}
\arguments{
\item{keyword}{String, name of keyword}

\item{name}{String, name of sub key}
}
\value{
String
}
\description{
Add a sub key word (see px-specification for details on sub keys)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{age_classification}
\alias{age_classification}
\title{Age classification}
\format{
A data frame:
\describe{
\item{valuecode}{Value code}
\item{valuetext}{Value text}
\item{10-years_classes}{Aggregtation into 10 years classes}
\item{25-years_classes}{Aggregtation into 25 years classes}
}
}
\usage{
age_classification
}
\description{
Example data set to create age classification with aggreations form 10 and
25 years classes.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{get_values_from_time_format}
\alias{get_values_from_time_format}
\title{Get time values from TIMEVAL string}
\usage{
get_values_from_time_format(str)
}
\arguments{
\item{str}{String with time values}
}
\value{
A character vector
}
\description{
Get time values from TIMEVAL string
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_totals.R
\name{add_totals_to_df}
\alias{add_totals_to_df}
\title{Add total levels to multiple variables}
\usage{
add_totals_to_df(df, variables, level_names, sum_var = "value", na_rm = TRUE)
}
\arguments{
\item{df}{Data frame to add total levels to.}

\item{variables}{List of variables to add total levels to.}

\item{level_names}{Names of total levels. Should have length 1 or same length
as \code{variables}.}

\item{sum_var}{String, name of variable to sum over.}

\item{na_rm}{Optional. Logical. If TRUE, NAs are removed before summing.}
}
\description{
Wrapper around \link{add_total_level_to_var} to add levels to total level to
multiple variables.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/px.R
\name{px_save}
\alias{px_save}
\title{Save px object to file}
\usage{
px_save(x, path, save_data = TRUE, data_path = NULL)
}
\arguments{
\item{x}{A px object.}

\item{path}{Path to file. The file extension determines the format. Can be:
\itemize{
\item \code{.px} to save as a PX-file
\item \code{.xlsx} to save as an Excel workbook
\item \code{.R} to save an R-script that creates the px object
}}

\item{save_data}{If FALSE, no 'Data' sheet is created in the Excel workbook.
Can only be used if \code{path} is an \code{.xlsx} file.}

\item{data_path}{Path to an \code{.rds} or \code{.parquet} file to save data table at.
Can only be used if \code{path} is an \code{.xlsx} or \code{.R} file, and \code{save_data} is
\code{TRUE}.}
}
\value{
Nothing
}
\description{
Save px object to file
}
\details{
Use \code{px_codepage()} to change file encoding.
}
\examples{

# Save px object to PX-file
tmp_dir <- tempdir()

x <- px(population_gl)

px_save(x, file.path(tmp_dir, "population.px"))

# Save px object to Excel workbook
px_save(x, file.path(tmp_dir, "population.xlsx"))

# Save px object as R-script that creates the px object
px_save(x, file.path(tmp_dir, "population.R"))

}
\seealso{
\code{\link[=px_codepage]{px_codepage()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_table2.R
\name{px_datasymbol3}
\alias{px_datasymbol3}
\alias{px_datasymbol3.px}
\title{DATASYMBOL3}
\usage{
px_datasymbol3(x, value, validate)

\method{px_datasymbol3}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string to set the value for all languages or a data frame with columns 'language' and 'value' to set it for specific languages. If 'value' is missing, the current DATASYMBOL3 is returned. If NULL, DATASYMBOL3 is removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object, a character string, or a data frame.
}
\description{
Inspect or change DATASYMBOL3.
}
\examples{
# Set DATASYMBOL3 for all languages
x1 <-
  px(population_gl) |>
  px_datasymbol3('missing')

# Print DATASYMBOL3
px_datasymbol3(x1)

# Set DATASYMBOL3 for individual languages
library(tibble)
x2 <-
  x1 |>
  px_languages(c('en', 'kl')) |>
  px_datasymbol3(tribble(~language, ~value,
                      'en', 'missing',
                      'kl', 'amigaataapput'))
px_datasymbol3(x2)

# Remove DATASYMBOL3
x3 <- px_datasymbol3(x2, NULL)
px_datasymbol3(x3)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A44\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C422\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classification.R
\name{extract_section}
\alias{extract_section}
\title{Get section in .vs or .agg file}
\usage{
extract_section(lines, heading, key = NULL)
}
\arguments{
\item{lines}{A character vector}

\item{heading}{Character, the section heading}

\item{key}{Optional. Character, subkey in section}
}
\description{
Get section in .vs or .agg file
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{read_lines_with_encoding}
\alias{read_lines_with_encoding}
\title{Read lines with specific encoding}
\usage{
read_lines_with_encoding(path, encoding)
}
\arguments{
\item{path}{Path to file}

\item{encoding}{Encoding to use}
}
\value{
A character vector
}
\description{
Read lines with specific encoding
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{download_px_or_parquet_and_return_path}
\alias{download_px_or_parquet_and_return_path}
\title{Download parquet or PX-file and return path}
\usage{
download_px_or_parquet_and_return_path(url)
}
\arguments{
\item{url}{}
}
\value{
Character
}
\description{
Assumes the file is .px, or parquet.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_helpers.R
\name{modify_or_add_row}
\alias{modify_or_add_row}
\title{Add or modify value}
\usage{
modify_or_add_row(
  df,
  lookup_column,
  lookup_column_values,
  modify_column,
  new_value
)
}
\arguments{
\item{df}{Data frame}

\item{lookup_column}{Column to look up}

\item{lookup_column_values}{Values to look up}

\item{modify_column}{Column to modify}

\item{new_value}{New value to modify/add to modify_column}
}
\value{
A data frame
}
\description{
Modify a value in a row of a data frame based on its value in another column.
If the value is not found, a new row is added.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{break_long_lines}
\alias{break_long_lines}
\title{Split long strings at commas}
\usage{
break_long_lines(str, max_line_length = 256)
}
\arguments{
\item{str}{String}

\item{max_line_length}{Integer longest allowed line length}
}
\value{
A character vector
}
\description{
Long strings are split so they are no longer than 256 characters and end at
a comma.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_table1.R
\name{px_codepage}
\alias{px_codepage}
\alias{px_codepage.px}
\title{CODEPAGE}
\usage{
px_codepage(x, value, validate)

\method{px_codepage}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string. If missing, the current CODEPAGE is returned. If NULL, CODEPAGE is removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object or a character string.
}
\description{
Inspect or change CODEPAGE.
}
\details{
CODEPAGE controls which encoding PX-files are read and stored in.
Use \code{iconvlist()} to see available encodings on your system.
}
\examples{
# Set CODEPAGE
x1 <-
   px(population_gl) |>
   px_codepage('utf-8')

# Print CODEPAGE
px_codepage(x1)

# Remove CODEPAGE
x2 <- px_codepage(x1, NULL)
px_codepage(x2)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A41\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C497\%2C0\%5D}{Statistics Sweden's documentation}

\code{\link[=px_save]{px_save()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_languages.R
\name{modify_languages_in_px}
\alias{modify_languages_in_px}
\title{Change language in px object}
\usage{
modify_languages_in_px(x, new_languages)
}
\arguments{
\item{x}{A px object}

\item{new_languages}{A character vector}
}
\value{
A px object
}
\description{
Changes languages in 3 of the px tables
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_variable1.R
\name{px_timeval}
\alias{px_timeval}
\alias{px_timeval.px}
\title{TIMEVAL}
\usage{
px_timeval(x, value, validate)

\method{px_timeval}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string. If missing, the current TIMEVAL is returned. If NULL, TIMEVAL is removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object or a character string.
}
\description{
Inspect or change TIMEVAL.

There can only be one time variable.
}
\examples{
# Set TIMEVAL
x1 <-
   px(population_gl) |>
   px_timeval('year')

# Print TIMEVAL
px_timeval(x1)

# Remove TIMEVAL
x2 <- px_timeval(x1, NULL)
px_timeval(x2)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A54\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C181\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/px_file.R
\name{save_px_as_px_file}
\alias{save_px_as_px_file}
\title{Save px object to PX-file}
\usage{
save_px_as_px_file(x, path)
}
\arguments{
\item{x}{A px object}

\item{path}{Path to save PX-file at}
}
\value{
Nothing
}
\description{
Save px object to PX-file
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{add_cell_to_keyword}
\alias{add_cell_to_keyword}
\title{Add cell name to keywords that support it (see px-specification)}
\usage{
add_cell_to_keyword(keyword, name)
}
\arguments{
\item{keyword}{String, name of keyword}

\item{name}{String, cell name to add to keyword}
}
\value{
String
}
\description{
Add cell name to keywords that support it (see px-specification)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata_df.R
\name{get_metadata_df_from_px_lines}
\alias{get_metadata_df_from_px_lines}
\title{Get metadata df from px lines}
\usage{
get_metadata_df_from_px_lines(metadata_lines)
}
\arguments{
\item{metadata_lines}{A character vector with the header of a PX-file.}
}
\value{
A data frame
}
\description{
Get metadata df from px lines
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{lst_distinct_and_arrange}
\alias{lst_distinct_and_arrange}
\title{Get a sorted list of distinct values in list}
\usage{
lst_distinct_and_arrange(lst)
}
\arguments{
\item{lst}{List to sort}
}
\value{
List
}
\description{
Get a sorted list of distinct values in list
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/px_file.R
\name{px_from_px_file}
\alias{px_from_px_file}
\title{Create a px object form a PX-file}
\usage{
px_from_px_file(path)
}
\arguments{
\item{path}{Path to a PX-file}
}
\value{
A px object
}
\description{
Create a px object form a PX-file
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/excel.R
\name{get_excel_sheet}
\alias{get_excel_sheet}
\title{Get specific sheet from Excel workbook}
\usage{
get_excel_sheet(sheet, add_automatically = FALSE)
}
\arguments{
\item{sheet}{String. Sheet to read.}

\item{add_automatically}{Logical. If TRUE, return an empty data frame if the
sheet does not exist.}
}
\value{
A data frame.
}
\description{
Get specific sheet from Excel workbook
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/excel.R
\name{px_from_excel}
\alias{px_from_excel}
\title{Create a px object from an Excel workbook}
\usage{
px_from_excel(excel_path, data = NULL)
}
\arguments{
\item{excel_path}{Path to Excel metadata workbook.}

\item{data}{A data frame if data isen't stored in the Excel workbook.}
}
\value{
A px object.
}
\description{
Create a px object from an Excel workbook
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/note_warning_error.R
\name{error}
\alias{error}
\title{Throw error}
\usage{
error(msg)
}
\arguments{
\item{msg}{String, error message}
}
\value{
Nothing
}
\description{
Throw error
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classification.R
\name{px_classification_from_path}
\alias{px_classification_from_path}
\title{px classification from path}
\usage{
px_classification_from_path(vs_path, agg_paths)
}
\arguments{
\item{vs_path}{Optional. Path to a values set (.vs) file.}

\item{agg_paths}{Optional.
\itemize{
\item If NULL, aggregation paths are automatically taken from the
[Aggreg] field in the .vs file.
\item Use a vector of paths to one or more aggregation files (.agg) to
manually choose aggregations.
\item Use character(0) if aggregations from the .vs files should not be
added automatically.
}}
}
\value{
A classification object
}
\description{
Create px classification from .vs and .agg files
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/note_warning_error.R
\name{validate_px_save_arguments}
\alias{validate_px_save_arguments}
\title{Check all arguments to px_save()}
\usage{
validate_px_save_arguments(x, path, save_data, data_path)
}
\arguments{
\item{x}{A px object.}

\item{path}{Path to file. The file extension determines the format. Can be:
\itemize{
\item \code{.px} to save as a PX-file
\item \code{.xlsx} to save as an Excel workbook
\item \code{.R} to save an R-script that creates the px object
}}

\item{save_data}{If FALSE, no 'Data' sheet is created in the Excel workbook.
Can only be used if \code{path} is an \code{.xlsx} file.}

\item{data_path}{Path to an \code{.rds} or \code{.parquet} file to save data table at.
Can only be used if \code{path} is an \code{.xlsx} or \code{.R} file, and \code{save_data} is
\code{TRUE}.}
}
\value{
Nothing
}
\description{
Check all arguments to px_save()
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_note.R
\name{px_notex}
\alias{px_notex}
\alias{px_notex.px}
\title{NOTEX}
\usage{
px_notex(x, value, validate)

\method{px_notex}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string, a data frame, or a list.
\itemize{
\item Use character, to set NOTEX for the entire table across
all languages.
\item Use a data frame with columns 'language' and 'value' to set
NOTEX for the entire table in a specific language.
\item Use a data frame with the columns 'variable-code' and 'notex',
to set NOTEX for a specific variable across all languages. Add the
column 'language' to set NOTEX for specific language.
\item Use a list of the above elements to set NOTEX in muliple ways.
This is the same as calling NOTEX multiple times with different
values.
\item If missing, the current NOTEX is returned.
\item If NULL, NOTEX is removed for the table and all variables.
}}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object, a character string, a data frame, or a list of character strings and/or data frames.
}
\description{
Inspect or change NOTEX.
}
\details{
NOTEX has a lot of possible ways to specify \code{value},
because it can be set both for the entire PX-file and for
individual variables.
}
\examples{
library(tibble)

# Set NOTEX for entire PX-file
x1 <-
  px(population_gl) |>
  px_notex('Note about PX-file')

# Print NOTEX
px_notex(x1)

# Set NOTEX for entire PX-file in multiple languages
x2 <-
  x1 |>
  px_languages(c('en', 'kl')) |>
  px_notex(tribble(~language,            ~value,
                       'en',    'English notex',
                       'kl', 'Kalaallisut notex'
                 )
         )
px_notex(x2)

# Set NOTEX for variables
x3 <-
  x1 |>
  px_notex(tribble(~`variable-code`, ~notex,
                 'year', 'Some data collected in following year',
                 'age',  'Is rounded down'
                 )
        )
px_notex(x3)

# Remove all NOTEXs
x4 <- px_notex(x3, NULL)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A50\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C271\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_df.R
\name{format_data_df}
\alias{format_data_df}
\title{Format df for px format}
\usage{
format_data_df(data_df, figures_variable)
}
\arguments{
\item{data_df}{A data frame with data.}

\item{figures_variable}{Character. The name of the figures variable.}
}
\value{
A data frame
}
\description{
Turn all variables, except figures variable, into character and replace NA
with dash.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{align_data_frames}
\alias{align_data_frames}
\title{Align data frames}
\usage{
align_data_frames(df_a, df_b)
}
\arguments{
\item{df_a}{Data frame to align}

\item{df_b}{Data frame to align to}
}
\value{
A data frame
}
\description{
Align df_a to have the same columns and column types as df_b.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/px.R
\name{fix_px}
\alias{fix_px}
\title{Fix some common issues in px objects}
\usage{
fix_px(x)
}
\arguments{
\item{x}{A px object.}
}
\value{
A px object
}
\description{
Fix some common issues in px objects
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/px.R
\name{return_px}
\alias{return_px}
\title{Return px object}
\usage{
return_px(x, validate)
}
\arguments{
\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\description{
Wrapper potentially validate it and return px object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_table1.R
\name{px_official_statistics}
\alias{px_official_statistics}
\alias{px_official_statistics.px}
\title{OFFICIAL-STATISTICS}
\usage{
px_official_statistics(x, value, validate)

\method{px_official_statistics}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string. If missing, the current OFFICIAL-STATISTICS is returned. If NULL, OFFICIAL-STATISTICS is removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object or a character string.
}
\description{
Inspect or change OFFICIAL-STATISTICS.
}
\examples{
# Set OFFICIAL-STATISTICS
x1 <-
   px(population_gl) |>
   px_official_statistics('YES')

# Print OFFICIAL-STATISTICS
px_official_statistics(x1)

# Remove OFFICIAL-STATISTICS
x2 <- px_official_statistics(x1, NULL)
px_official_statistics(x2)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A51\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C699\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_table2.R
\name{px_last_updated}
\alias{px_last_updated}
\alias{px_last_updated.px}
\title{LAST-UPDATED}
\usage{
px_last_updated(x, value, validate)

\method{px_last_updated}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string to set the value for all languages or a data frame with columns 'language' and 'value' to set it for specific languages. If 'value' is missing, the current LAST-UPDATED is returned. If NULL, LAST-UPDATED is removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object, a character string, or a data frame.
}
\description{
Inspect or change LAST-UPDATED.
}
\examples{
# Set LAST-UPDATED for all languages
x1 <-
  px(population_gl) |>
  px_last_updated('17070501 15:55')

# Print LAST-UPDATED
px_last_updated(x1)

# Set LAST-UPDATED for individual languages
library(tibble)
x2 <-
  x1 |>
  px_languages(c('en', 'kl')) |>
  px_last_updated(tribble(~language, ~value,
                      'en', '17070501 15:55',
                      'kl', '20080621 15:55'))
px_last_updated(x2)

# Remove LAST-UPDATED
x3 <- px_last_updated(x2, NULL)
px_last_updated(x3)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A49\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C608\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/r_script.R
\name{convert_value_to_code}
\alias{convert_value_to_code}
\title{Create constructing code}
\usage{
convert_value_to_code(value)
}
\arguments{
\item{value}{Vector of values to create constructors for}
}
\value{
A character vector
}
\description{
Creates code that construct input value.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_df.R
\name{px_from_data_df}
\alias{px_from_data_df}
\title{Create a minimal px object from a data frame}
\usage{
px_from_data_df(df)
}
\arguments{
\item{df}{A data frame}
}
\value{
A px object
}
\description{
Create a minimal px object from a data frame
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_variable1.R
\name{px_figures}
\alias{px_figures}
\alias{px_figures.px}
\title{Change figures variable}
\usage{
px_figures(x, value, validate)

\method{px_figures}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. Name of variable to use as FIGRUES. If missing, the
current PX_FIGURES variable is returned.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object or a character string
}
\description{
Inspect or change which variable is used as figures. The previous figures
variable is changed to STUB. There can only be one figures variable.
}
\examples{
x1 <- px(population_gl)

# Print FIGURES
px_figures(x1)

# Change 'age' to FIGURES variable, 'n' i changed to STUB
x2 <- px_figures(x1, "age")
px_figures(x2)
px_stub(x2)
}
\seealso{
\code{\link{px_stub}} \code{\link{px_heading}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_note.R
\name{px_note}
\alias{px_note}
\alias{px_note.px}
\title{NOTE}
\usage{
px_note(x, value, validate)

\method{px_note}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string, a data frame, or a list.
\itemize{
\item Use character, to set NOTE for the entire table across
all languages.
\item Use a data frame with columns 'language' and 'value' to set
NOTE for the entire table in a specific language.
\item Use a data frame with the columns 'variable-code' and 'note',
to set NOTE for a specific variable across all languages. Add the
column 'language' to set NOTE for specific language.
\item Use a list of the above elements to set NOTE in muliple ways.
This is the same as calling NOTE multiple times with different
values.
\item If missing, the current NOTE is returned.
\item If NULL, NOTE is removed for the table and all variables.
}}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object, a character string, a data frame, or a list of character strings and/or data frames.
}
\description{
Inspect or change NOTE.
}
\details{
NOTE has a lot of possible ways to specify \code{value},
because it can be set both for the entire PX-file and for
individual variables.
}
\examples{
library(tibble)

# Set NOTE for entire PX-file
x1 <-
  px(population_gl) |>
  px_note('Note about PX-file')

# Print NOTE
px_note(x1)

# Set NOTE for entire PX-file in multiple languages
x2 <-
  x1 |>
  px_languages(c('en', 'kl')) |>
  px_note(tribble(~language,            ~value,
                       'en',    'English note',
                       'kl', 'Kalaallisut note'
                 )
         )
px_note(x2)

# Set NOTE for variables
x3 <-
  x1 |>
  px_note(tribble(~`variable-code`, ~note,
                 'year', 'Some data collected in following year',
                 'age',  'Is rounded down'
                 )
        )
px_note(x3)

# Remove all NOTEs
x4 <- px_note(x3, NULL)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A50\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C411\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{greenlanders}
\alias{greenlanders}
\title{Greenlanders}
\format{
An object of class \code{tbl_df} (inherits from \code{tbl}, \code{data.frame}) with 100 rows and 4 columns.
}
\usage{
greenlanders
}
\description{
A fictive data set with demographic data for Greenlanders split in two cohorts.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{wrap_varaible_in_list}
\alias{wrap_varaible_in_list}
\title{Convert a variable to a list}
\usage{
wrap_varaible_in_list(df, var)
}
\arguments{
\item{df}{Data frame}

\item{var}{Character. Variable to convert to list}
}
\value{
A data frame
}
\description{
Convert a variable to a list
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_table2.R
\name{px_title}
\alias{px_title}
\alias{px_title.px}
\title{TITLE}
\usage{
px_title(x, value, validate)

\method{px_title}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string to set the value for all languages or a data frame with columns 'language' and 'value' to set it for specific languages. If 'value' is missing, the current TITLE is returned. If NULL, TITLE is removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object, a character string, or a data frame.
}
\description{
Inspect or change TITLE.

TITLE can only be removed if DESCRIPTION is set.
}
\examples{
# Set TITLE for all languages
x1 <-
  px(population_gl) |>
  px_title('Population GR')

# Print TITLE
px_title(x1)

# Set TITLE for individual languages
library(tibble)
x2 <-
  x1 |>
  px_languages(c('en', 'kl')) |>
  px_title(tribble(~language, ~value,
                      'en', 'Population GR',
                      'kl', 'Innuttaasut KL'))
px_title(x2)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A36\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C533\%2C0\%5D}{Statistics Sweden's documentation}

\code{\link[=px_description]{px_description()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_cells2.R
\name{px_valuenote}
\alias{px_valuenote}
\alias{px_valuenote.px}
\title{VALUENOTE}
\usage{
px_valuenote(x, value, validate)

\method{px_valuenote}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A data frame with the columns 'valuenote' and one or more of the columns: 'variable-code', 'code', and 'language'. If 'value' is missing, the current VALUENOTE is returned. If NULL, VALUENOTE is removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object or data frame.
}
\description{
Inspect or change VALUENOTE.
}
\examples{
# Set VALUENOTE for a value
library(tibble)
x1 <-
  population_gl |>
  px() |>
  px_valuenote(
    tribble(~`variable-code`, ~code,  ~valuenote,
            'year', '2004', 'Counts are approximated'))

# Print VALUENOTE
px_valuenote(x1)

# Set VALUENOTE for a value in specific language
x2 <-
  x1 |>
  px_languages(c('en', 'kl')) |>
  px_valuenote(
    tribble(~`variable-code`, ~code,  ~language, ~valuenote,
            'age', '0-6', 'en', 'Some of the figures are from 2003',
            'age', '0-6', 'kl', 'Kisitsisit ilaat 2003-imeersuupput'))
px_valuenote(x2)

# Remove VALUENOTE
x3 <- px_valuenote(x2, NULL)
px_valuenote(x3)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A55\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C133\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{merge_named_lists}
\alias{merge_named_lists}
\title{Put two named lists together, remove duplicates and sort}
\usage{
merge_named_lists(lst1, lst2)
}
\arguments{
\item{lst1}{List to merge}

\item{lst2}{List to merge}
}
\value{
List
}
\description{
Put two named lists together, remove duplicates and sort
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_variable1.R
\name{px_variable_type}
\alias{px_variable_type}
\alias{px_variable_type.px}
\title{VARIABLE-TYPE}
\usage{
px_variable_type(x, value, validate)

\method{px_variable_type}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{A data frame with columns 'variable-code' and 'variable-type'.
If value is missing, the current VARIABLE-TYPE is returned. If NULL, all
VARIABLE-TYPE is removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object or data frame.
}
\description{
Inspect or change VARIABLE-TYPE.
}
\examples{
library(tibble)

# Set VARIABLE-TYPE
x1 <-
  px(population_gl) |>
  px_variable_type(
    tibble("variable-code" = "year", "variable-type" = "time")
  )

# Print VARIABLE-TYPE
px_variable_type(x1)

# Remove VARIABLE-TYPE
x2 <- px_variable_type(x1, NULL)
px_variable_type(x2)

}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A56\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C499\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classification.R
\name{px_classification}
\alias{px_classification}
\title{Create a classification object}
\usage{
px_classification(name, prestext, domain, df, vs_path, agg_paths)
}
\arguments{
\item{name}{Optional. Name of the classification.}

\item{prestext}{Optional. Presentation text.}

\item{domain}{Optional. Character vector with domain names. Used to link to
PX-file.}

\item{df}{Optional. A data frame with required column 'valuecode' and
optional column 'valuetext', if the codes have texts. Each additional column
represents an aggregation. The column name is the name of the aggregation.
If the column type is character the aggregation levels will be sorted
alphabetically; use factors to control the ordering.}

\item{vs_path}{Optional. Path to a values set (.vs) file.}

\item{agg_paths}{Optional.
\itemize{
\item If NULL, aggregation paths are automatically taken from the
[Aggreg] field in the .vs file.
\item Use a vector of paths to one or more aggregation files (.agg) to
manually choose aggregations.
\item Use character(0) if aggregations from the .vs files should not be
added automatically.
}}
}
\value{
A classification object
}
\description{
Create a classification object from a data frame or .vs and .agg files.
}
\details{
A classification is a combination of a value set and zero, one, or more
aggregations. The classification can be saved as .vs and .agg files
(see \code{\link[=px_save_classification]{px_save_classification()}}).

If a classification is created from a data frame, the arguments \code{name} and
\code{prestext} and \code{domain} are required. If a classification is created from .vs
and .agg files, all other arguments should be empty.

For aggregations, it's normally possible to have codes and values in the .agg
file under the sections '[Aggreg]' and '[Aggtext]' respectively. However,
in pxmake's implementation of classifications, it's not possible to
distinguish between these. When a new classification is created, the values
in the section '[Aggtext]' are used as aggregation values.

Only value sets of type 'V' are supported. Type values sets with type 'H' and
'N' are not supported.
}
\examples{
# Create classification from data frame
library(tibble)

c1 <- px_classification(
  name = "Age5",
  prestext = "Ages 0-9 - 60+",
  domain = "age",
  df = tribble(
    ~valuecode, ~valuetext, ~`25 years classes`,
    "0-4", "0-4 years", "0-24",
    "5-9", "5-9 years", "0-24",
    "10-14", "10-14 years", "0-24",
    "15-19", "15-19 years", "0-24",
    "20-24", "20-24 years", "0-24",
    "25-29", "25-29 years", "25-49",
    "30-34", "30-34 years", "25-49",
    "35-39", "35-39 years", "25-49",
    "40-44", "40-44 years", "25-49",
    "45-49", "45-49 years", "25-49",
    "50-54", "50-54 years", "50-74",
    "55-59", "55-59 years", "50-74",
    "60-64", "60-64 years", "50-74",
    "65-69", "65-69 years", "50-74",
    "70-74", "70-74 years", "50-74",
    "75+", "75+ years", "75+"
  )
)

# Create classifications from files

vs_file <- system.file("extdata", "Age5.vs", package = "pxmake")

agg_files <- c(
  system.file("extdata", "10-years_classes.agg", package = "pxmake"),
  system.file("extdata", "25-years_classes.agg", package = "pxmake")
)

if (vs_file != "" & all(agg_files != "")) {
  # Create classification from .vs file and use aggregations mentioned in .vs
  c2 <- px_classification(vs_path = vs_file)

  # Create classification from .vs file and manually specify aggregation
  # files
  c3 <- px_classification(
    vs_path = vs_file,
    agg_paths = agg_files
  )

  identical(c2, c3)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/px_file.R
\name{format_px_object_as_lines}
\alias{format_px_object_as_lines}
\title{Get lines for PX-file from px object}
\usage{
format_px_object_as_lines(x)
}
\arguments{
\item{x}{A px object}
}
\value{
A character vector
}
\description{
Get lines for PX-file from px object
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_cells1.R
\name{px_precision}
\alias{px_precision}
\alias{px_precision.px}
\title{PRECISION}
\usage{
px_precision(x, value, validate)

\method{px_precision}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A data frame with the columns 'precision' and one or more of the columns: 'variable-code', and 'code'. If 'value' is missing, the current PRECISION is returned. If NULL, PRECISION is removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object or data frame.
}
\description{
Inspect or change PRECISION.
}
\examples{
# Set PRECISION for a variable
library(tibble)
x1 <-
  population_gl |>
  px() |>
  px_precision(tribble(~`variable-code`, ~precision,
                       'gender', 2))

# Print PRECISION
px_precision(x1)

# Set PRECISION for a value
x2 <-
  x1 |>
  px_precision(tribble(~`variable-code`, ~code, ~precision,
                       'age', '2004', 3))
px_precision(x2)

# Remove PRECISION
x3 <- px_precision(x2, NULL)
px_precision(x3)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A51\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C385\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/micro.R
\name{px_micro}
\alias{px_micro}
\title{Create micro PX-files}
\usage{
px_micro(x, out_dir = NULL, keyword_values = NULL)
}
\arguments{
\item{x}{A px object.}

\item{out_dir}{Directory to save PX-files in.}

\item{keyword_values}{Optional. A data frame with column 'variable' and one
or more of: 'px_contents', 'px_title', 'px_description', and 'px_matrix'. The
columns will be added as keywords to the table for each non-HEADING variable
that match the 'variable' column. It probably work for other keywords as
well.

Use the column 'filename' to control the filename of each micro file. The
filename path is relative to 'out_dir'.

Use the column 'language' if the PX-file has multiple languages.}
}
\value{
Nothing
}
\description{
Split one px object into many small PX-files (micro files), with count of
the variables in it.
}
\details{
The HEADING variables are use in all the micro files, and a file is created
for each non-HEADING variable. The new PX-files are saved in a directory
specified by \code{out_dir}.

The main loop uses the furrr package for parallelisation. Use future::plan()
to choose how to parallelise.
}
\examples{
# Create px object with cohort as HEADING
x <-
  greenlanders |>
  px() |>
  px_stub(names(greenlanders)) |>
  px_heading("cohort")

# Create micro files, one for each of the non-HEADING variables (gender, age,
# municipality)
px_micro(x)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classification.R
\name{smallest_larger_value}
\alias{smallest_larger_value}
\title{Smallest larger value}
\usage{
smallest_larger_value(vec, value)
}
\arguments{
\item{vec}{A numeric vector}

\item{value}{A numeric value}
}
\value{
The smallest value in 'vec' that is larger than 'value'
}
\description{
Smallest larger value
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_cells2.R
\name{px_values}
\alias{px_values}
\alias{px_values.px}
\title{VALUES}
\usage{
px_values(x, value, validate)

\method{px_values}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A data frame with the columns 'values' and one or more of the columns: 'variable-code', 'code', and 'language'. If 'value' is missing, the current VALUES is returned. If NULL, VALUES is removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object or data frame.
}
\description{
Inspect or change VALUES.
}
\examples{
# Set VALUES for a value
library(tibble)
x1 <-
  population_gl |>
  px() |>
  px_values(
    tribble(~`variable-code`, ~code,  ~values,
            'year', '2004', 'Year 2024'))

# Print VALUES
px_values(x1)

# Set VALUES for a value in specific language
x2 <-
  x1 |>
  px_languages(c('en', 'kl')) |>
  px_values(
    tribble(~`variable-code`, ~code,  ~language, ~values,
            'age', '0-6', 'en', 'toddler',
            'age', '0-6', 'kl', 'meeraaqqap'))
px_values(x2)

# Remove VALUES
x3 <- px_values(x2, NULL)
px_values(x3)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A36\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C212\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_languages.R
\name{defined_languages}
\alias{defined_languages}
\title{Get languages used in px object}
\usage{
defined_languages(x)
}
\arguments{
\item{x}{A px object}
}
\value{
A character vector
}
\description{
Get languages used in px object
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/px.R
\name{px}
\alias{px}
\title{Create a px object}
\usage{
px(input = NULL, data = NULL, validate = TRUE)
}
\arguments{
\item{input}{Optional character string. Can be:
\itemize{
\item Path to a PX-file
\item Path to an Excel metadata workbook
\item A data frame
\item Path to an \code{.rds} or \code{.parquet} file with a data frame
\item Direct download URL of a PX-file
\item Direct download URL of a parquet file
}

If input is a data frame or NULL, a px object with minimal
metadata is created.}

\item{data}{Either a data frame or a path to an \code{.rds} or \code{.parquet} file
with a data frame. This can only be used if \code{input} is an Excel metadata
workbook.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object
}
\description{
Create a px object from a PX-file, an Excel metadata workbook, or a data
frame.
}
\examples{
# Create px object from dataset
x1 <- px(population_gl)

# Download PX-file for example
download_succeeded <- tryCatch(
  {
    px_path <- tempfile(fileext = ".px")
    url <- "https://bank.stat.gl:443/sq/0cf06962-19f1-4d5c-8d43-b7ed0009617d"
    download.file(url, px_path)
    TRUE
  },
  error = function(e) FALSE
)

# Run examples only if file was downloaded
if (download_succeeded) {
  # Create px object from PX-file
  x2 <- px(px_path)

  # Create px object from URL
  x3 <- px(url)
}

# Create minimal px object
x4 <- px()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_languages.R
\name{modify_languages_in_table}
\alias{modify_languages_in_table}
\title{Change language in table}
\usage{
modify_languages_in_table(df, new_languages, keep_vars, align_df)
}
\arguments{
\item{df}{A data frame to change languages in}

\item{new_languages}{A character vector}

\item{keep_vars}{A character vector}

\item{align_df}{A data frame to align with}
}
\value{
A data frame
}
\description{
Change language in table
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata_df.R
\name{get_px_metadata_regex}
\alias{get_px_metadata_regex}
\title{Regular expression to parse header in PX-file}
\usage{
get_px_metadata_regex()
}
\value{
A character vector
}
\description{
Regular expression to parse header in PX-file
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classification.R
\name{classification_file_section_heading_regexp}
\alias{classification_file_section_heading_regexp}
\title{Regexp that matches a classification file section heading}
\usage{
classification_file_section_heading_regexp()
}
\description{
Regexp that matches a classification file section heading
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/excel.R
\name{add_excel_sheet}
\alias{add_excel_sheet}
\title{Add a data frame as a sheet to an Excel workbook}
\usage{
add_excel_sheet(wb, df, sheet_name)
}
\arguments{
\item{wb}{An Excel workbook}

\item{df}{A data frame}

\item{sheet_name}{Name of the sheet}
}
\value{
Nothing
}
\description{
Add a data frame as a sheet to an Excel workbook
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_data.R
\name{get_data_table}
\alias{get_data_table}
\title{Return data table from px object}
\usage{
get_data_table(x, value, labels, sort)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A data frame. If missing, the current
DATA is returned. If NULL, all data rows are removed.}

\item{labels}{Optional. Logic or character vector. If TRUE, the data table
is returned with VALUES instead of CODES. By default the VALUES of the main
language are returned, use a character language code to return VALUES for a
specific language.}

\item{sort}{Optional. If TRUE, the data table is returned in the sort order
defined by \code{\link[=px_order]{px_order()}}. If FALSE, the data table is returned as is.}
}
\description{
Return data table from px object
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{temp_file_with_extension}
\alias{temp_file_with_extension}
\title{Create temporary file}
\usage{
temp_file_with_extension(extension)
}
\arguments{
\item{extension}{String, file name extension}
}
\value{
Path to temporary file
}
\description{
Get a temporary file path with a specific extension (function factory)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{add_language_to_keyword}
\alias{add_language_to_keyword}
\title{Add language, other than the main langue, to keyword}
\usage{
add_language_to_keyword(keyword, main_language, language)
}
\arguments{
\item{keyword}{String, name of keyword}

\item{main_language}{Main language of PX-file}

\item{language}{Lange to add to keyword}
}
\value{
String
}
\description{
Add language, other than the main langue, to keyword
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/note_warning_error.R
\name{validate_px_arguments}
\alias{validate_px_arguments}
\title{Check all arguments to px()}
\usage{
validate_px_arguments(input, data)
}
\arguments{
\item{input}{Optional character string. Can be:
\itemize{
\item Path to a PX-file
\item Path to an Excel metadata workbook
\item A data frame
\item Path to an \code{.rds} or \code{.parquet} file with a data frame
\item Direct download URL of a PX-file
\item Direct download URL of a parquet file
}

If input is a data frame or NULL, a px object with minimal
metadata is created.}

\item{data}{Either a data frame or a path to an \code{.rds} or \code{.parquet} file
with a data frame. This can only be used if \code{input} is an Excel metadata
workbook.}
}
\value{
Nothing
}
\description{
Check all arguments to px()
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_variable2.R
\name{px_map}
\alias{px_map}
\alias{px_map.px}
\title{MAP}
\usage{
px_map(x, value, validate)

\method{px_map}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string or data frame.
\itemize{
\item Use character to set MAP for all languages and STUB/HEADING
variables.
\item Use data frame with columns 'variable-code', 'language' and
'map' to set MAP for specific variables.
\item If missing, the current MAP is returned.
\item If NULL, MAP is removed for all variables.
}}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object, a character string, or a data frame.
}
\description{
Inspect or change MAP.
}
\examples{
# Set MAP for all languages
x1 <-
  px(population_gl) |>
  px_map('greenland')

# Print MAP
px_map(x1)

# Set MAP for individual variables
library(tibble)
x2 <-
  x1 |>
  px_map(tribble(~`variable-code`, ~map,
                    'gender', 'cities',
                    'age',    'municipalities'))
px_map(x2)

# Set MAP for individual languages
x3 <-
  x2 |>
  px_languages(c('en', 'kl')) |>
  px_map(tribble(~`variable-code`, ~language, ~map,
                    'gender',    'en',      'cities_en',
                    'gender',    'kl',      'cities_kl',
                    'age',       'en',      'municipalities_en'))
px_map(x3)

# Remove MAP
x4 <- px_map(x3, NULL)
px_map(x4)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A49\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C354\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{drop_blank_rows}
\alias{drop_blank_rows}
\title{Drop rows with only NA values}
\usage{
drop_blank_rows(df)
}
\arguments{
\item{df}{Data frame}
}
\value{
A data frame
}
\description{
Drop rows with only NA values
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classification.R
\name{aggregation_df}
\alias{aggregation_df}
\title{Get data set from aggregation file}
\usage{
aggregation_df(path)
}
\arguments{
\item{path}{to an aggregation file}
}
\value{
A data frame with columns 'valuecode' (character) and a
second column (ordered) named after the aggregation
}
\description{
Get data set from aggregation file
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_variable1.R
\name{get_pivot_variables}
\alias{get_pivot_variables}
\title{Get names of pivot variables}
\usage{
get_pivot_variables(x, pivot)
}
\arguments{
\item{x}{A px object}

\item{pivot}{A string, either "STUB", "HEADING" or "FIGURES"}
}
\value{
A character vector of variable codes
}
\description{
Get names of pivot variables
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_variable1.R
\name{px_elimination}
\alias{px_elimination}
\alias{px_elimination.px}
\title{ELIMINATION}
\usage{
px_elimination(x, value, validate)

\method{px_elimination}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. Use character to set ELIMINATION for all STUB and
HEADING variables. Use a data frame with columns 'variable-code' and
'elimination' to set ELIMINATION for individual variables. If value is
missing, the current ELIMINATION is returned. If NULL, ELIMINATION is removed
for all variables.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object or data frame.
}
\description{
Inspect or change ELIMINATION.
}
\examples{
library(tibble)

# Set ELIMINATION
x1 <-
  px(population_gl) |>
  px_elimination(tribble(
    ~`variable-code`, ~elimination,
    "gender", "T",
    "age", "YES"
  ))

# Print ELIMINATION
px_elimination(x1)

# Remove ELIMINATION
x2 <- px_elimination(x1, NULL)
px_elimination(x2)

}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A46\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C525\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_table1.R
\name{px_matrix}
\alias{px_matrix}
\alias{px_matrix.px}
\title{MATRIX}
\usage{
px_matrix(x, value, validate)

\method{px_matrix}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string. If missing, the current MATRIX is returned. If NULL, an error is thrown because MATRIX cannot be removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object or a character string.
}
\description{
Inspect or change MATRIX.
}
\examples{
# Set MATRIX
x1 <-
   px(population_gl) |>
   px_matrix('POPULATION')

# Print MATRIX
px_matrix(x1)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A35\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C442\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{zip_vectors}
\alias{zip_vectors}
\title{Zips list}
\usage{
zip_vectors(v1, v2)
}
\arguments{
\item{v1}{First vector Elements ends up on uneven indexes.}

\item{v2}{Second vector. Elements ends up on even indexes.}
}
\value{
List
}
\description{
Combine two list, by zipping them together in the order \code{v1[1]},
\code{v2[1]}, \code{v1[2]}, \code{v2[2]}, \code{v1[3]}, ...
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/excel.R
\name{save_px_as_xlsx}
\alias{save_px_as_xlsx}
\title{Save px object as an Excel workbook}
\usage{
save_px_as_xlsx(x, path, save_data, data_path)
}
\arguments{
\item{x}{A px object}

\item{path}{Path to save Excel workbook}

\item{save_data}{If FALSE, no 'Data' sheet is created in the Excel workbook.
Can only be used if \code{path} is an \code{.xlsx} file.}

\item{data_path}{Path to an \code{.rds} or \code{.parquet} file to save data table at.
Can only be used if \code{path} is an \code{.xlsx} or \code{.R} file, and \code{save_data} is
\code{TRUE}.}
}
\value{
Nothing
}
\description{
Save px object as an Excel workbook
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{temp_dir}
\alias{temp_dir}
\title{Create and return path to temporary directory}
\usage{
temp_dir()
}
\value{
Character
}
\description{
Create and return path to temporary directory
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/note_warning_error.R
\name{validate_px_classification_arguments}
\alias{validate_px_classification_arguments}
\title{Check all arguments to px_classification}
\usage{
validate_px_classification_arguments(
  name,
  prestext,
  domain,
  df,
  vs_path,
  agg_paths
)
}
\arguments{
\item{name}{Optional. Name of the classification.}

\item{prestext}{Optional. Presentation text.}

\item{domain}{Optional. Character vector with domain names. Used to link to
PX-file.}

\item{df}{Optional. A data frame with required column 'valuecode' and
optional column 'valuetext', if the codes have texts. Each additional column
represents an aggregation. The column name is the name of the aggregation.
If the column type is character the aggregation levels will be sorted
alphabetically; use factors to control the ordering.}

\item{vs_path}{Optional. Path to a values set (.vs) file.}

\item{agg_paths}{Optional.
\itemize{
\item If NULL, aggregation paths are automatically taken from the
[Aggreg] field in the .vs file.
\item Use a vector of paths to one or more aggregation files (.agg) to
manually choose aggregations.
\item Use character(0) if aggregations from the .vs files should not be
added automatically.
}}
}
\value{
Nothing
}
\description{
Check all arguments to px_classification
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_variable1.R
\name{change_pivot_variables}
\alias{change_pivot_variables}
\title{Change pivot variables}
\usage{
change_pivot_variables(x, value, pivot)
}
\arguments{
\item{x}{A px object}

\item{value}{A character vector of variable codes to change to the pivot
type}

\item{pivot}{Pivot type (STUB, HEADING, FIGURES)}
}
\value{
A px object
}
\description{
Change pivot variables
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{px_keywords}
\alias{px_keywords}
\title{px keywords}
\format{
A data frame:
\describe{
\item{keyword}{Name}
\item{px_function}{Function name}
\item{mandatory}{Is required in a PX-file}
\item{table_meta}{Is metadata for entire table; not individual variables or cells}
\item{language_dependent}{Is language dependent}
\item{indexed_by_contvariable}{Is indexed if CONTVARIABLE is set}
\item{quote_value}{Value should be quoted in PX-file}
\item{default_value}{Default value for mandatory keywords}
\item{documentation}{URL to Statistic Sweden's documentation}
\item{order}{Recommended order}
}
}
\source{
\url{https://www.scb.se/globalassets/vara-tjanster/px-programmen/PX-file_format_specification_2013.pdf}
}
\usage{
px_keywords
}
\description{
Properties of all px keywords. Used internally by the package.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/px.R
\name{px_validate}
\alias{px_validate}
\title{Check px object}
\usage{
px_validate(x)
}
\arguments{
\item{x}{A supposed px object.}
}
\value{
A valid px object.
}
\description{
Runs a number of checks on px object to see if it is valid.
}
\details{
This check is run by default by all \verb{px_*} functions, but can be skipped by
using \code{validate = FALSE}. This can be useful on large px objects where the
checks are time consuming. Instead of validating on every modifying function
\code{px_validate()} can be run as the final step to validate the object.
}
\examples{
# Turn off validation for modifying functions, and manually
# run validation as final step in creating px object.
x1 <-
  px(population_gl, validate = FALSE) |>
  px_title("Test", validate = FALSE) |>
  px_validate()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_acrosscells.R
\name{px_cellnotex}
\alias{px_cellnotex}
\alias{px_cellnotex.px}
\title{CELLNOTEX}
\usage{
px_cellnotex(x, value, na_to_star, validate)

\method{px_cellnotex}{px}(x, value, na_to_star = TRUE, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A data frame with columns 'cellnotex' and one or more columns with the names of the STUB and HEADING variables. The 'cellnotex' column is the cellnotex text, and the STUB/HEADING columns control which cells the note applies to. Use star (*) if a note applies to all cells in a variable. Use column 'language'  to set CELLNOTEX for specific languages. If 'value' is missing, the current CELLNOTEX is returned. If value is NULL, CELLNOTEX is removed.}

\item{na_to_star}{Optional. Convert all NAs to '*'.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object or data frame.
}
\description{
Inspect or change CELLNOTEX.
}
\examples{
# Set CELLNOTEX for a value
library(tibble)
x1 <-
  population_gl |>
  px() |>
  px_cellnotex(
    tribble(~gender,  ~age,  ~year, ~cellnote,
             'male', '0-6', '2004', 'Approximation'))

x2 <-
  x1 |>
  px_cellnote(
    tribble(~gender,   ~age,  ~year, ~cellnote,
           'female',    '*', '2014', 'Uncertainty in ages'))

# Print CELLNOTEX
px_cellnotex(x2)

# Set CELLNOTEX in multiple languagese
x3 <-
  x1 |>
  px_languages(c('en', 'kl')) |>
  px_cellnotex(
    tribble(~age, ~year, ~language, ~cellnote,
             '*', '2003',  'en', 'Some of the figures are from 2003',
             '*', '2003', 'kl', 'Kisitsisit ilaat 2003-imeersuupput'))
px_cellnotex(x3)

# Remove CELLNOTEX
x4 <- px_cellnotex(x3, NULL)
px_cellnotex(x4)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A40\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C277\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/note_warning_error.R
\name{error_if_agument_is_not_character_string}
\alias{error_if_agument_is_not_character_string}
\title{Throw error if argument is not a character string}
\usage{
error_if_agument_is_not_character_string(name, value)
}
\description{
Throw error if argument is not a character string
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/note_warning_error.R
\name{validate_px_micro_arguments}
\alias{validate_px_micro_arguments}
\title{Check all arguments to px_micro()}
\usage{
validate_px_micro_arguments(x, out_dir)
}
\arguments{
\item{x}{A px object.}

\item{out_dir}{Directory to save PX-files in.}
}
\value{
Nothing
}
\description{
Check all arguments to px_micro()
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_table1.R
\name{px_showdecimals}
\alias{px_showdecimals}
\alias{px_showdecimals.px}
\title{SHOWDECIMALS}
\usage{
px_showdecimals(x, value, validate)

\method{px_showdecimals}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string. If missing, the current SHOWDECIMALS is returned. If NULL, SHOWDECIMALS is removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object or a character string.
}
\description{
Inspect or change SHOWDECIMALS.
}
\examples{
# Set SHOWDECIMALS
x1 <-
   px(population_gl) |>
   px_showdecimals('2')

# Print SHOWDECIMALS
px_showdecimals(x1)

# Remove SHOWDECIMALS
x2 <- px_showdecimals(x1, NULL)
px_showdecimals(x2)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A53\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C305\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_totals.R
\name{px_add_totals}
\alias{px_add_totals}
\alias{px_add_totals.px}
\title{Add total levels to variables}
\usage{
px_add_totals(x, value, na_rm = TRUE, validate)

\method{px_add_totals}{px}(x, value, na_rm = TRUE, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{A character vector of variables to add total levels to.}

\item{na_rm}{Optional. Logical. If TRUE, NAs are removed before summing.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object
}
\description{
Adds a total level, which is the sum of the figures for all other levels of
the variable.

The default name of the total level is 'Total', unless \link{px_elimination}
is set, in which case the elimination code is used.
}
\examples{
# Create small px object example
x0 <- px(subset(population_gl, age == "65+"))
px_data(x0)

# Add total level to one variable
x1 <- px_add_totals(x0, "gender")
px_data(x1)

# Add total level to multiple variables
x2 <- px_add_totals(x0, c("gender", "age"))
px_data(x2)

# The name of the total level can be changed with px_elimination()
x3 <-
  x0 |>
  px_elimination("T") |>
  px_add_totals("gender")

px_data(x3)

}
\seealso{
\link{px_elimination}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{get_timeval_type_from_values}
\alias{get_timeval_type_from_values}
\title{Get time scale code from values (see TIMEVAL in px-specification)}
\usage{
get_timeval_type_from_values(values)
}
\arguments{
\item{values}{Values form PX-file}
}
\value{
A character vector
}
\description{
Get time scale code from values (see TIMEVAL in px-specification)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_totals.R
\name{add_total_level_to_var}
\alias{add_total_level_to_var}
\title{Add total level to variable}
\usage{
add_total_level_to_var(
  df,
  variable,
  level_name = "Total",
  sum_var = "value",
  na_rm = TRUE
)
}
\arguments{
\item{df}{Data frame to add total levels to.}

\item{variable}{Name of variable to add total level to.}

\item{level_name}{Total level name.}

\item{sum_var}{String, name of variable to sum over.}

\item{na_rm}{Optional. Logical. If TRUE, NAs are removed before summing.}
}
\value{
A data frame
}
\description{
Add a new level to a variables which is the sum of all other levels.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_variable1.R
\name{px_heading}
\alias{px_heading}
\alias{px_heading.px}
\title{HEADING}
\usage{
px_heading(x, value, validate)

\method{px_heading}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character vector of variable names to change to
STUB. This also changes the HEADING order. With names in \code{variables}
becoming 1, 2, ... If missing, the current HEADING variables are returned.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object or a character vector.
}
\description{
Inspect or change HEADING.
}
\examples{
x1 <- px(population_gl)

# Print HEADING
px_heading(x1)

# Add 'gender' to HEADING
x2 <- px_heading(x1, "gender")
px_heading(x2)

# Change order of HEADING
x3 <- px_heading(x2, "year")
px_heading(x3)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A35\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C610\%2C0\%5D}{Statistics Sweden's documentation}

\code{\link{px_stub}} \code{\link{px_figures}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_table2.R
\name{px_cfprices}
\alias{px_cfprices}
\alias{px_cfprices.px}
\title{CFPRICES}
\usage{
px_cfprices(x, value, validate)

\method{px_cfprices}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string to set the value for all languages or a data frame with columns 'language' and 'value' to set it for specific languages. If 'value' is missing, the current CFPRICES is returned. If NULL, CFPRICES is removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object, a character string, or a data frame.
}
\description{
Inspect or change CFPRICES.
}
\examples{
# Set CFPRICES for all languages
x1 <-
  px(population_gl) |>
  px_cfprices('C')

# Print CFPRICES
px_cfprices(x1)

# Set CFPRICES for individual languages
library(tibble)
x2 <-
  x1 |>
  px_languages(c('en', 'kl')) |>
  px_cfprices(tribble(~language, ~value,
                      'en', 'C',
                      'kl', 'F'))
px_cfprices(x2)

# Remove CFPRICES
x3 <- px_cfprices(x2, NULL)
px_cfprices(x3)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A40\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C178\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/px.R
\name{new_px}
\alias{new_px}
\title{Create new px object}
\usage{
new_px(
  languages,
  table1,
  table2,
  variables1,
  variables2,
  cells1,
  cells2,
  acrosscells,
  data
)
}
\arguments{
\item{languages}{A data frame with language metadata.}

\item{table1}{A data frame with language independent table metadata.}

\item{table2}{A data frame with language dependent table metadata.}

\item{variables1}{A data frame with language independent variable metadata.}

\item{variables2}{A data frame with language dependent variable metadata.}

\item{cells1}{A data frame with language independent cells metadata.}

\item{cells2}{A data frame with language dependent cells metadata.}

\item{acrosscells}{A data frame with metadata that spans multiple cells.}

\item{data}{A data frame with data.}
}
\value{
A px object
}
\description{
px constructor for internal functions
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/note_warning_error.R
\name{get_mandatory_variables}
\alias{get_mandatory_variables}
\title{A list of which variables should be in each sheet}
\usage{
get_mandatory_variables()
}
\description{
A list of which variables should be in each sheet
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_table1.R
\name{px_creation_date}
\alias{px_creation_date}
\alias{px_creation_date.px}
\title{CREATION-DATE}
\usage{
px_creation_date(x, value, validate)

\method{px_creation_date}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string. If missing, the current CREATION-DATE is returned. If NULL, CREATION-DATE is removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object or a character string.
}
\description{
Inspect or change CREATION-DATE.
}
\examples{
# Set CREATION-DATE
x1 <-
   px(population_gl) |>
   px_creation_date('19960612 14:20')

# Print CREATION-DATE
px_creation_date(x1)

# Remove CREATION-DATE
x2 <- px_creation_date(x1, NULL)
px_creation_date(x2)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A42\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C176\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_cells1.R
\name{px_order}
\alias{px_order}
\alias{px_order.px}
\title{Change value order}
\usage{
px_order(x, value, validate)

\method{px_order}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A data frame with the columns 'order' and one or more of the columns: 'variable-code', and 'code'. If 'value' is missing, the current ORDER is returned. If NULL, ORDER is removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object or data frame.
}
\description{
Inspect or change ORDER.
}
\examples{
# Set ORDER for a variable
library(tibble)
x1 <-
  population_gl |>
  px() |>
  px_order(tribble(~`variable-code`, ~order,
                       'gender', 8))

# Print ORDER
px_order(x1)

# Set ORDER for a value
x2 <-
  x1 |>
  px_order(tribble(~`variable-code`, ~code, ~order,
                       'age', '2004', 9))
px_order(x2)

# Remove ORDER
x3 <- px_order(x2, NULL)
px_order(x3)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_table2.R
\name{px_datasymbol4}
\alias{px_datasymbol4}
\alias{px_datasymbol4.px}
\title{DATASYMBOL4}
\usage{
px_datasymbol4(x, value, validate)

\method{px_datasymbol4}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string to set the value for all languages or a data frame with columns 'language' and 'value' to set it for specific languages. If 'value' is missing, the current DATASYMBOL4 is returned. If NULL, DATASYMBOL4 is removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object, a character string, or a data frame.
}
\description{
Inspect or change DATASYMBOL4.
}
\examples{
# Set DATASYMBOL4 for all languages
x1 <-
  px(population_gl) |>
  px_datasymbol4('missing')

# Print DATASYMBOL4
px_datasymbol4(x1)

# Set DATASYMBOL4 for individual languages
library(tibble)
x2 <-
  x1 |>
  px_languages(c('en', 'kl')) |>
  px_datasymbol4(tribble(~language, ~value,
                      'en', 'missing',
                      'kl', 'amigaataapput'))
px_datasymbol4(x2)

# Remove DATASYMBOL4
x3 <- px_datasymbol4(x2, NULL)
px_datasymbol4(x3)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A44\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C422\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata_df.R
\name{get_metadata_df_from_px}
\alias{get_metadata_df_from_px}
\title{Get metadata df from px object}
\usage{
get_metadata_df_from_px(x)
}
\arguments{
\item{x}{A px object}
}
\value{
A data frame
}
\description{
Get metadata df from px object
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_table1.R
\name{px_next_update}
\alias{px_next_update}
\alias{px_next_update.px}
\title{NEXT-UPDATE}
\usage{
px_next_update(x, value, validate)

\method{px_next_update}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string. If missing, the current NEXT-UPDATE is returned. If NULL, NEXT-UPDATE is removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object or a character string.
}
\description{
Inspect or change NEXT-UPDATE.
}
\examples{
# Set NEXT-UPDATE
x1 <-
   px(population_gl) |>
   px_next_update('20240621 15:55')

# Print NEXT-UPDATE
px_next_update(x1)

# Remove NEXT-UPDATE
x2 <- px_next_update(x1, NULL)
px_next_update(x2)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A49\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C144\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{str_quote}
\alias{str_quote}
\title{Add quotes around string}
\usage{
str_quote(str)
}
\arguments{
\item{str}{String to quote}
}
\value{
String
}
\description{
Add quotes around string
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classification.R
\name{write_aggregation}
\alias{write_aggregation}
\title{Save classification as .agg file}
\usage{
write_aggregation(aggregation, c, directory)
}
\arguments{
\item{aggregation}{Character, the column name of the aggregation in the
classification}

\item{c}{A classification object}

\item{directory}{Directory to save the file in}
}
\value{
Nothing
}
\description{
Write one aggregation column in a classification to an .agg file.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classification.R
\name{px_save_classification}
\alias{px_save_classification}
\title{Save classification as .vs and .agg files}
\usage{
px_save_classification(c, path)
}
\arguments{
\item{c}{A classification object}

\item{path}{Directory to save the files in}
}
\value{
Nothing.
}
\description{
Save a classification object as .vs and .agg files. The .vs file contains
the value set and the .agg files contain the aggregations.
}
\examples{
# Save classification as .vs as .agg files
c <- px_classification(
  name = "Age5",
  prestext = "Ages 0-9 - 60+",
  domain = "age",
  df = age_classification
)

px_save_classification(c, path = tempdir())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_variable1.R
\name{px_stub}
\alias{px_stub}
\alias{px_stub.px}
\title{STUB}
\usage{
px_stub(x, value, validate)

\method{px_stub}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character vector of variable names to change to
STUB. This also changes the STUB order. With names in \code{variables}
becoming 1, 2, ... If missing, the current STUB variables are returned.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object or a character vector.
}
\description{
Inspect or change STUB.
}
\examples{
x1 <- px(population_gl)
# Print STUB
px_stub(x1)
# Add 'year' to STUB
x2 <- px_stub(x1, "year")
px_stub(x2)

# Change order of STUB
x3 <- px_stub(x2, c("age", "gender"))
px_stub(x3)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A35\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C329\%2C0\%5D}{Statistics Sweden's documentation}

\code{\link{px_heading}} \code{\link{px_figures}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_data.R
\name{sort_data_table_by_order}
\alias{sort_data_table_by_order}
\title{Sort data table by order}
\usage{
sort_data_table_by_order(x)
}
\description{
Sort data table by order
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{read_px_file}
\alias{read_px_file}
\title{Get PX-file content as lines}
\usage{
read_px_file(px_path)
}
\arguments{
\item{px_path}{Path to a PX-file}
}
\value{
A character vector
}
\description{
Read file with correct encoding.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{create_dummy_tibbles}
\alias{create_dummy_tibbles}
\title{Create a tibble with dummy values}
\usage{
create_dummy_tibbles(dummy_value)
}
\arguments{
\item{dummy_value}{Value to set in all columns}
}
\description{
Create a tibble with dummy values
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classification.R
\name{write_value_set}
\alias{write_value_set}
\title{Save classification as .vs file}
\usage{
write_value_set(c, directory)
}
\arguments{
\item{c}{A classification object}

\item{directory}{Directory to save the file in}
}
\value{
Nothing
}
\description{
Write value set part of classification to a file.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{mutate_all_vars_to_character}
\alias{mutate_all_vars_to_character}
\title{Change all variables to character}
\usage{
mutate_all_vars_to_character(df)
}
\arguments{
\item{df}{Data frame}
}
\value{
A data frame
}
\description{
Change all variables to character
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{guess_file_encoding}
\alias{guess_file_encoding}
\title{Guess encoding of file}
\usage{
guess_file_encoding(path, prefer = list(`UTF-8` = 0.2, `ISO-8859-1` = 0.2))
}
\arguments{
\item{path}{Path to file}

\item{prefer}{Named list of encodings, used to give higher confidence to
specific encodings.}
}
\value{
Character
}
\description{
Guess encoding of file
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{get_default_encoding}
\alias{get_default_encoding}
\title{Default encoding to read and save PX-file in}
\usage{
get_default_encoding()
}
\value{
Character
}
\description{
Default encoding to read and save PX-file in
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_variable1.R
\name{px_contvariable}
\alias{px_contvariable}
\alias{px_contvariable.px}
\title{CONTVARIABLE}
\usage{
px_contvariable(x, value, validate)

\method{px_contvariable}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string. If missing, the current CONTVARIABLE is returned. If NULL, CONTVARIABLE is removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object or a character string.
}
\description{
Inspect or change CONTVARIABLE.

Setting CONTVARIABLE indexes several variables in table2.
Removing CONTVARIABLE removes the indexing from table2.
}
\examples{
# Set CONTVARIABLE
x1 <-
  px(population_gl) |>
  px_contvariable("gender")

# After setting CONTVARIABLE some variables are index by it, e.g. UNITS
px_units(x1)

# Remove CONTVARIABLE
x2 <- px_contvariable(x1, NULL)
px_contvariable(x2)

# Removing CONTVARIABLE also removes the index from UNITS
px_units(x2)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A42\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C484\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classification.R
\name{enumerate_lines}
\alias{enumerate_lines}
\title{Add lines numbers}
\usage{
enumerate_lines(lines)
}
\arguments{
\item{lines}{A character vector}
}
\value{
A character vector
}
\description{
Add '1=', '2=', etc. at start of each line.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_table2.R
\name{px_contact}
\alias{px_contact}
\alias{px_contact.px}
\title{CONTACT}
\usage{
px_contact(x, value, validate)

\method{px_contact}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string to set the value for all languages or a data frame with columns 'language' and 'value' to set it for specific languages. If 'value' is missing, the current CONTACT is returned. If NULL, CONTACT is removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object, a character string, or a data frame.
}
\description{
Inspect or change CONTACT.
}
\examples{
# Set CONTACT for all languages
x1 <-
  px(population_gl) |>
  px_contact('Johan Ejstrud')

# Print CONTACT
px_contact(x1)

# Set CONTACT for individual languages
library(tibble)
x2 <-
  x1 |>
  px_languages(c('en', 'kl')) |>
  px_contact(tribble(~language, ~value,
                      'en', 'Johan Ejstrud',
                      'kl', 'Lars Pedersen'))
px_contact(x2)

# Remove CONTACT
x3 <- px_contact(x2, NULL)
px_contact(x3)
}
\seealso{
\href{https://www.scb.se/globalassets/vara-tjanster/px-programmen/px-file_format_specification_2013.pdf#\%5B\%7B\%22num\%22\%3A42\%2C\%22gen\%22\%3A0\%7D\%2C\%7B\%22name\%22\%3A\%22XYZ\%22\%7D\%2C184\%2C665\%2C0\%5D}{Statistics Sweden's documentation}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/modify_languages.R
\name{px_language}
\alias{px_language}
\alias{px_language.px}
\title{LANGUAGE}
\usage{
px_language(x, value, validate)

\method{px_language}{px}(x, value, validate = TRUE)
}
\arguments{
\item{x}{A px object}

\item{value}{Optional. A character string. If missing, the current LANGUAGE
is returned. If NULL, LANGUAGE is removed.}

\item{validate}{Optional. If TRUE a number of validation checks are performed on the px object, and an error is thrown if the object is not valid. If FALSE, the checks are skipped, which can be usefull for large px objects where the check can be time consuming. Use \code{\link[=px_validate]{px_validate()}} to manually preform the check.}
}
\value{
A px object
}
\description{
Inspect or change LANGUAGE.
}
\details{
If LANGUAGES is defined, changing LANGUAGE will also add is to LANGUAGES.
}
\examples{
# Set LANGUAGE to 'en'
x1 <-
  population_gl |>
  px() |>
  px_language("en")

# Print LANGUAGE
px_language(x1)

# Remove LANGUAGE
x2 <- px_language(x1, NULL)
px_language(x2)

}
\seealso{
\code{\link{px_languages}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helper_functions.R
\name{get_encoding_from_px_file}
\alias{get_encoding_from_px_file}
\title{Get encoding listed in PX-file}
\usage{
get_encoding_from_px_file(px_path)
}
\arguments{
\item{px_path}{Path to a PX-file}
}
\value{
Character
}
\description{
Encoding is listed in CODEPAGE.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/classification.R
\name{section_interval}
\alias{section_interval}
\title{Get index of first and last line of section}
\usage{
section_interval(lines, head_line)
}
\arguments{
\item{lines}{A character vector}

\item{head_line}{A numeric value, line number of section heading}
}
\description{
Get index of first and last line of section
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/metadata_df.R
\name{get_main_language}
\alias{get_main_language}
\title{Get the main language from metadata}
\usage{
get_main_language(metadata_df)
}
\arguments{
\item{metadata_df}{Data frame with metadata.}
}
\value{
Character
}
\description{
Get the main language from metadata
}
\keyword{internal}
